#include "tcaps_ci.h"

TCAPS_Network_Task * start_comms_thread(unsigned short local_port);
TCAPS_Rule_Management_Task * start_rm_thread();

struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote);
void check_for_packets(TCAPS_Network_Task * comms_task);
void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote);
void handle_tp_packet(struct tcaps_sp_tp * packet, ACE_INET_Addr * remote);
void handle_pp_packet(struct tcaps_sp_pp * packet, ACE_INET_Addr * remote);
void handle_rmp_packet(struct tcaps_sp_rmp * packet, TCAPS_Rule_Management_Task * comms_task, ACE_INET_Addr * remote);

void send_packet(struct tcaps_sp_packet * packet, struct ip_address recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task);
void poll_server(TCAPS_Network_Task * tm_comms_task);

void perform_startup_functions(TCAPS_Network_Task * tm_comms_task);
void perform_shutdown_functions(TCAPS_Network_Task * tm_comms_task);

void cleanup();
void handler(int value);

void cleanup();
void handler(int value);
static sig_atomic_t finished = 0;

TCAPS_Network_Task * tm_comms_task;
TCAPS_Rule_Management_Task * rm_task;


int main (int argc,char *argv[])
{
	struct timespec sleep_timer;
	
	tm_comms_task = NULL;
	rm_task = NULL;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 10000000; //10ms
	
	
	ACE_Sig_Action sa ((ACE_SignalHandler) handler, SIGINT);
	
	
	if(tcaps_qosss_init(TCAPS_CI_INTERFACE))
	{
		printf("\nUnable to initialise QoS SS. Exiting\n");
		exit(1);
	}


	tm_comms_task = start_comms_thread(TCAPS_CI_LOCAL_PORT);
	rm_task = start_rm_thread();
	
	tm_comms_task->open();
	rm_task->open();
	
	perform_startup_functions(tm_comms_task);
	
	while(!finished)
	{
		check_for_packets(tm_comms_task);
		poll_server(tm_comms_task);
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Killing some time in the main thread...\n"));
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for a bit to be nice to other processes running on the system
	}
	
	cleanup();

	return 0;
}

void cleanup()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) We are in the cleanup function!\n"));

	/* Once we're done, we have to signal the Task objects to shut
    down. There are several choices including: - Send a message of
    zero length - Send a message with a special content I don't like
    these choices because they're likely to interfere with application
    logic.  Instead, I use the message type feature to send a message
    of type "hangup".  The default type is MB_DATA, so when the tasks
    get a MB_HANGUP type, they know to go away.  */

	perform_shutdown_functions(tm_comms_task);
	
	if(tm_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the tm_comms_task queue\n"));
		tm_comms_task->putq(message);
		tm_comms_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete tm_comms_task;
	}
	
	if(rm_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fm_task queue\n"));
		rm_task->putq(message);
		rm_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete rm_task;
	}
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to call tcaps_qosss_shutdown()\n"));
	tcaps_qosss_shutdown();
}

void handler(int value)
{
	finished = 1;
}

void perform_startup_functions(TCAPS_Network_Task * tm_comms_task)
{
	char macbuf[64];
	
	//send an initial registration poll packet to the server
	tcaps_id mac;
	
	tcaps_sp_get_interface_mac(&mac, TCAPS_CI_INTERFACE);
	
	printHexBytes(mac.octets, 6, macbuf, sizeof(macbuf), TRUE);
	
	struct tcaps_sp_packet packet;
	
	packet.packet_type = TCAPS_PP_PACKET_TYPE;
	packet.packet = tcaps_sp_pp_create(macbuf, TCAPS_PP_POLL_TYPE_REGISTER);

	struct ip_address tmpip;
	tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
	tmpip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
	tcaps_ip_aton(TCAPS_TM_IP, &tmpip);
	
	send_packet(&packet, tmpip, TCAPS_TM_REMOTE_PORT, tm_comms_task);
	
	tcaps_sp_pp_destroy((struct tcaps_sp_pp *)packet.packet);
}

void perform_shutdown_functions(TCAPS_Network_Task * tm_comms_task)
{
	char macbuf[64];
	
	//send an initial registration poll packet to the server
	tcaps_id mac;
	
	tcaps_sp_get_interface_mac(&mac, TCAPS_CI_INTERFACE);
	
	printHexBytes(mac.octets, 6, macbuf, sizeof(macbuf), TRUE);
	
	struct tcaps_sp_packet packet;
	
	packet.packet_type = TCAPS_PP_PACKET_TYPE;
	packet.packet = tcaps_sp_pp_create(macbuf, TCAPS_PP_POLL_TYPE_UNREGISTER);

	struct ip_address tmpip;
	tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
	tmpip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
	tcaps_ip_aton(TCAPS_TM_IP, &tmpip);
	
	send_packet(&packet, tmpip, TCAPS_TM_REMOTE_PORT, tm_comms_task);
	
	tcaps_sp_pp_destroy((struct tcaps_sp_pp *)packet.packet);
}

TCAPS_Network_Task * start_comms_thread(unsigned short local_port)
{
	TCAPS_Network_Task * comms_task = NULL;
	
	ACE_INET_Addr local(local_port);
	
	ACE_SOCK_Dgram * dgram = new ACE_SOCK_Dgram();
	
	if (dgram->open (local) != -1)
	{
		comms_task = new TCAPS_Network_Task(dgram);
	}
	
	return comms_task;
}

void poll_server(TCAPS_Network_Task * tm_comms_task)
{
	static unsigned int poll_timer = 0;
	char macbuf[64];
	
	if(poll_timer <= (unsigned int)ACE_OS::time())
	{
		//printf("poll_timer = %d, time = %d\n", poll_timer, (unsigned int)ACE_OS::time());
		
		//time to send another poll packet to the server
		tcaps_id mac;
		
		tcaps_sp_get_interface_mac(&mac, TCAPS_CI_INTERFACE);
		
		printHexBytes(mac.octets, 6, macbuf, sizeof(macbuf), TRUE);
		
		struct tcaps_sp_packet packet;
		
		packet.packet_type = TCAPS_PP_PACKET_TYPE;
		packet.packet = tcaps_sp_pp_create(macbuf, TCAPS_PP_POLL_TYPE_POLL);
		
		//struct tcaps_sp_packet * packet, struct ip_address recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task
		struct ip_address tmpip;
		tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
		tmpip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		tcaps_ip_aton(TCAPS_TM_IP, &tmpip);
		
		send_packet(&packet, tmpip, TCAPS_TM_REMOTE_PORT, tm_comms_task);
		
		tcaps_sp_pp_destroy((struct tcaps_sp_pp *)packet.packet);
		
		//update the poll timer value
		poll_timer = TCAPS_CI_POLL_INTERVAL + (unsigned int)ACE_OS::time();
	}
}

TCAPS_Rule_Management_Task * start_rm_thread()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Starting the tcaps_ci rule management thread\n"));
	
	TCAPS_Rule_Management_Task * rm_task = NULL;
	
	rm_task = new TCAPS_Rule_Management_Task();

	return rm_task;
}

void check_for_packets(TCAPS_Network_Task * comms_task)
{
	ACE_INET_Addr remote;
	struct tcaps_sp_packet * packet = receive_packet(comms_task, &remote);
    
	while(packet != NULL)
	{
		switch(packet->packet_type)
		{
			case TCAPS_AP_PACKET_TYPE:
				handle_ap_packet((struct tcaps_sp_ap * )packet->packet, &remote);
				break;
			case TCAPS_TP_PACKET_TYPE:
				handle_tp_packet((struct tcaps_sp_tp * )packet->packet, &remote);
				break;
			case TCAPS_PP_PACKET_TYPE:
				handle_pp_packet((struct tcaps_sp_pp * )packet->packet, &remote);
				break;
			case TCAPS_RMP_PACKET_TYPE:
				handle_rmp_packet((struct tcaps_sp_rmp * )packet->packet, rm_task, &remote);
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected tcaps_sp packet type\n"));
				break;
		}

		free(packet);
		
		packet = receive_packet(comms_task, &remote);
	}
}


void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_ap_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an AP that looks like this:\n%s\n", tmp));
	tcaps_sp_ap_destroy(packet);
}

void handle_pp_packet(struct tcaps_sp_pp * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_pp_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an PP that looks like this:\n%s\n", tmp));
	tcaps_sp_pp_destroy(packet);
}

void handle_tp_packet(struct tcaps_sp_tp * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_tp_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an TP that looks like this:\n%s\n", tmp));
	tcaps_sp_tp_destroy(packet);
}

void handle_rmp_packet(struct tcaps_sp_rmp * packet, TCAPS_Rule_Management_Task * rm_task, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_rmp_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an RMP that looks like this:\n%s\n", tmp));
	
//	char tmp[BUFFER_SIZE];
//	char rule_command[BUFFER_SIZE];
//	int rule_number;
//	
//	printf("\nAttempting to translate the RMP\n");
//	tcaps_translate_rmp2qosss(packet, TCAPS_CI_INTERFACE, rule_command, &rule_number);
//	printf("Translated RMP rule: %s\n", rule_command);
//	
//	if(packet->op_type == TCAPS_RMP_OP_TYPE_CREATE)
//	{
//		printf("Attempting to add the rule\n");
//		tcaps_qosss_add_rule(rule_command, TCAPS_CI_INTERFACE);
//	}
//	else if(packet->op_type == TCAPS_RMP_OP_TYPE_DELETE)
//	{
//		printf("Attempting to delete the rule\n");
//		tcaps_qosss_delete_rule(rule_command, TCAPS_CI_INTERFACE);
//	}
//	tcaps_sp_rmp_tostring(packet, tmp, sizeof(tmp));
//	tcaps_sp_rmp_destroy(packet);

        
    //printf("handle_rmp_packet() about to run..%d\n", packet->serialized_bytes);
  	Packet_Block * message;

  	message = new Packet_Block(sizeof(unsigned int) + packet->serialized_bytes);
	
	message->copy ((const char *) &packet->serialized_bytes, sizeof(unsigned int));
  	message->copy ((const char *) packet->serialized, packet->serialized_bytes);
      	

	ACE_DEBUG ((LM_DEBUG, "(%P|%t) adding a RMP to the rm_queue\n"));
	rm_task->putq(message);
	
	//tcaps_sp_rmp_destroy(packet);
}


struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote)
{
	ACE_Message_Block *message;
	struct tcaps_sp_packet * packet = NULL;
	ACE_Time_Value wait_for_message(1, 0);
	
	message = comms_task->receive_packet();
	
	if(message != NULL)
	{
		unsigned int num_serialized_bytes;
    		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to read a message from a comms_task!\n"));
		
		ACE_OS::memmove ((char *) remote, message->rd_ptr(), sizeof(ACE_INET_Addr));
		message->rd_ptr (sizeof(ACE_INET_Addr));  // Move the rd_ptr() beyond the data.
		
		ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
		message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
		
		u8 serialized_bytes[num_serialized_bytes];
		
		ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
		message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
		message->release();
		
		packet = (struct tcaps_sp_packet *)malloc(sizeof(struct tcaps_sp_packet));
		
		tcaps_sp_parse_packet((u8 *)serialized_bytes, num_serialized_bytes, packet);
	}
	
	return packet;

}


void send_packet(struct tcaps_sp_packet * packet, struct ip_address recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task)
{
	char iptmp[60], pktmp[1024];
	tcaps_ip_ntoa(&recipient_ip, iptmp, TRUE);	    
    ACE_INET_Addr remote(recipient_port, iptmp);
		
	/* Create a block large enough for our Data object as well as a
    text message.  */
  	Packet_Block *message;
  	u8 * serialized;
  	unsigned int serialized_bytes;
  	
  	switch(packet->packet_type)
  	{
//  		case TCAPS_CRMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_crmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_crmp *)packet->packet)->serialized;
//  			tcaps_sp_crmp_tostring(((struct tcaps_sp_crmp *)packet->packet), pktmp, sizeof(pktmp));
//  			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Adding a CRMP to the queue\n%s\n", pktmp));
//  			break;
//  		case TCAPS_FFP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_ffp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_ffp *)packet->packet)->serialized;
//  			break;
  		case TCAPS_RMP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_rmp *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_rmp *)packet->packet)->serialized;
  			break;
  		case TCAPS_PP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_pp *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_pp *)packet->packet)->serialized;
  			break;
  		case TCAPS_TP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_tp *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_tp *)packet->packet)->serialized;
  			break;
  		case TCAPS_AP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_ap *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_ap *)packet->packet)->serialized;
  			break;
  		default:
  			serialized_bytes = 0;
  			serialized = 0;
  			break;
  	}

  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
  	
  	
  	message->copy ((const char *) &remote, sizeof(remote));
  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
  	message->copy ((const char *) serialized, serialized_bytes);

	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) adding a packet to the queue\n"));
	comms_task->send_packet(message);

}







