#include "tcaps_ci.h"

///* Set our housekeeping pointer to NULL and tell the user we exist.  */
//TCAPS_Rule::TCAPS_Rule (unsigned int rn, char * rtext, unsigned int rtimeout)
//  : rule_number (rn), rule_timeout (rtimeout+ACE_OS::time())
//{
//  ACE_DEBUG ((LM_DEBUG, "(%P|%t) Task ctor 0x%x\n", (void *) this));
//  rule_text = (char *)malloc(strlen(rtext)+1);
//  strncpy(rule_text, rtext, strlen(rtext)+1);
//  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule ctor, rule_number: %d, rule_text: %s, rule_timeout: %d\n", rule_number, rule_text, rule_timeout));
//  
//}
//
///* Take care of cleanup & tell the user we're going away.  */
//TCAPS_Rule::~TCAPS_Rule (void)
//{
//  ACE_DEBUG ((LM_DEBUG, "(%P|%t) Task dtor 0x%x\n", (void *) this));
//  free(rule_text);
//}
//
//unsigned int TCAPS_Rule::get_timeout()
//{
//	return rule_timeout;
//}
//
//const char * TCAPS_Rule::get_rule_text()
//{
//	return rule_text;
//}













TCAPS_Rule::TCAPS_Rule ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule default ctor 0x%x\n", (void *) this));
	
	rule_text = NULL;
	expiry_time_secs = 0;
}

TCAPS_Rule::TCAPS_Rule (const struct tcaps_sp_rmp * rule_pack)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule(struct tcaps_sp_rmp * rule_pack) ctor 0x%x\n", (void *) this));
	
	u8 * serialized = (u8 *)malloc(rule_pack->serialized_bytes);
	memcpy(serialized, rule_pack->serialized, rule_pack->serialized_bytes);
	rule = tcaps_sp_rmp_parse_packet(serialized, rule_pack->serialized_bytes);
	expiry_time_secs = (unsigned int)ACE_OS::time() + rule->timeout;
	rule_text = NULL;
	free(serialized);
}

TCAPS_Rule::TCAPS_Rule (const TCAPS_Rule &copied_rule)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule copy ctor 0x%x\n", (void *) this));
	
	u8 * serialized = (u8 *)malloc(copied_rule.get_serialized_bytes());
	memcpy(serialized, copied_rule.get_serialized(), copied_rule.get_serialized_bytes());
	rule = tcaps_sp_rmp_parse_packet(serialized, copied_rule.get_serialized_bytes());
	expiry_time_secs = copied_rule.get_expiry_time();
	set_rule_text(copied_rule.get_rule_text());
	free(serialized);
}

TCAPS_Rule::~TCAPS_Rule (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule dtor 0x%x\n", (void *) this));
	
	if(rule != NULL)
		tcaps_sp_rmp_destroy(rule);
		
	if(rule_text != NULL)
		free(rule_text);
}
	
TCAPS_Rule & TCAPS_Rule::operator=(const TCAPS_Rule &copied_rule)
{
	if(this == &copied_rule)
		return *this;
	
	u8 * serialized = (u8 *)malloc(copied_rule.get_serialized_bytes());
	memcpy(serialized, copied_rule.get_serialized(), copied_rule.get_serialized_bytes());
	rule = tcaps_sp_rmp_parse_packet(serialized, copied_rule.get_serialized_bytes());
	expiry_time_secs = copied_rule.get_expiry_time();
	set_rule_text(copied_rule.get_rule_text());
	free(serialized);
	
	return *this;
}

const struct tcaps_sp_rmp * TCAPS_Rule::get_rmp() const
{
	return rule;
}

unsigned int TCAPS_Rule::get_expiry_time() const
{
	return expiry_time_secs;
}

void TCAPS_Rule::increment_expiry_time(u16 timeout)
{
	expiry_time_secs += timeout;
	tcaps_sp_rmp_set_timeout(rule, timeout);
}

u16 TCAPS_Rule::get_rule_num() const
{
	return rule->rule_number;
}

const u8 * TCAPS_Rule::get_serialized() const
{
	return rule->serialized;
}

unsigned int TCAPS_Rule::get_serialized_bytes() const
{
	return rule->serialized_bytes;
}


void TCAPS_Rule::to_string(char * buf, unsigned int bufsize) const
{
	tcaps_sp_rmp_tostring(rule, buf, bufsize);
}

const char * TCAPS_Rule::get_rule_text() const
{
	return rule_text;
}

void TCAPS_Rule::set_rule_text(const char * ruletext)
{
	if(rule_text != NULL)
		free(rule_text);
		
	rule_text = (char *)malloc(strlen(ruletext)+1);
	
	strncpy(rule_text, ruletext, strlen(ruletext)+1);
}

bool operator==(const TCAPS_Rule &r1, const TCAPS_Rule &r2)
{
	//if(r1.rule->rule_number == r2.rule->rule_number)
	
	if(r1.rule->serialized_bytes = r2.rule->serialized_bytes)
	{
		if(memcmp(r1.rule->serialized, r2.rule->serialized, r1.rule->serialized_bytes) == 0) 
		{
			return true;
		}
	}
	
	return false;
}
