#include "tcaps_ci.h"


TCAPS_Rule_List::TCAPS_Rule_List ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_List default ctor 0x%x\n", (void *) this));
}

TCAPS_Rule_List::TCAPS_Rule_List (const TCAPS_Rule_List &copied_rule_list)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_List copy ctor 0x%x\n", (void *) this));
	
	rule_list = copied_rule_list.get_rule_list();
}

TCAPS_Rule_List::~TCAPS_Rule_List (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_List dtor 0x%x\n", (void *) this)); 
}
	
TCAPS_Rule_List & TCAPS_Rule_List::operator=(const TCAPS_Rule_List &copied_rule_list)
{
	if(this == &copied_rule_list)
		return *this;
		
	rule_list = copied_rule_list.get_rule_list();
	
	return *this;
}


std::map<u16,TCAPS_Rule *> TCAPS_Rule_List::get_rule_list() const
{
	return rule_list;
}

void TCAPS_Rule_List::add_rule(TCAPS_Rule * therule)
{
	rule_list_mutex.acquire();
	
	//add rule to the list
	rule_list.insert(rule_entry(therule->get_rule_num(), therule));
	
	rule_list_mutex.release();
}

void TCAPS_Rule_List::acquire_mutex()
{
	rule_list_mutex.acquire();
}

void TCAPS_Rule_List::release_mutex()
{
	rule_list_mutex.release();
}

bool TCAPS_Rule_List::find_rule(u16 rulenum)
{
	std::map<u16,TCAPS_Rule *>::iterator rule_iter;
	
	rule_iter = rule_list.find(rulenum);
	   				 
	if (rule_iter != rule_list.end())
		return true;
		
	return false;
}

TCAPS_Rule * TCAPS_Rule_List::get_rule(u16 rulenum)
{
	std::map<u16,TCAPS_Rule *>::iterator rule_iter;
	
	rule_iter = rule_list.find(rulenum);
	   				 
	if (rule_iter != rule_list.end())
		return ((*rule_iter).second);
		
	return NULL;
}

TCAPS_Rule * TCAPS_Rule_List::get_rule(const struct tcaps_sp_rmp * rule_pack)
{
	std::map<u16,TCAPS_Rule *>::iterator rule_iter;
	   				 
	for(rule_iter = rule_list.begin(); rule_iter != rule_list.end(); rule_iter++)
	{
		if(tcaps_sp_rmp_compare_rules(rule_pack, ((*rule_iter).second)->get_rmp()))
			return ((*rule_iter).second);
	}
		
	return NULL;
}

void TCAPS_Rule_List::remove_rule(u16 rulenum)
{
	std::map<u16,TCAPS_Rule *>::iterator rule_iter;
	
	rule_iter = rule_list.find(rulenum);
	   				 
	if (rule_iter != rule_list.end())
	{
		delete ((*rule_iter).second);
		rule_list.erase(rule_iter);
	}
}

void TCAPS_Rule_List::accept(void (*f)(TCAPS_Rule *, void *), void * args)
{
	std::map<u16,TCAPS_Rule *>::iterator rule_iter;
	
	rule_list_mutex.acquire();
	
	for(rule_iter = rule_list.begin(); rule_iter != rule_list.end(); rule_iter++)
	{
		(*f)(rule_iter->second, args);
	}
	
	rule_list_mutex.release();
}


bool operator==(const TCAPS_Rule_List &f1, const TCAPS_Rule_List &f2)
{
//	if(//lists are the same size)
//	{
//		if(each flow in the list is same)
//		{
//			return true;
//		}
//	}
	
	return false;
}
