#include "tcaps_ci.h"

/* Set our housekeeping pointer to NULL and tell the user we exist.  */
TCAPS_Rule_Management_Task::TCAPS_Rule_Management_Task ()
 : n_threads_ (1)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_Management_Task ctor 0x%x\n", (void *) this));
	client_rules = new TCAPS_Rule_List();
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Rule_Management_Task::~TCAPS_Rule_Management_Task (void)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_Management_Task dtor 0x%x\n", (void *) this));

	ACE_Message_Block *message;
	this->getq (message);
	message->release ();
	
	//delete all QOS SS rules
	client_rules->accept(&TCAPS_Rule_Management_Task::delete_qosss_rules, NULL);
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int TCAPS_Rule_Management_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP, n_threads_);
}


int TCAPS_Rule_Management_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule_Management_Task close 0x%x\n", (void *) this));
  return inherited::close (flags);
}


int TCAPS_Rule_Management_Task::svc (void)
{
	ACE_Message_Block *message;
	TCAPS_Rule * tmp_rule;
	
	//this may be foobed as we are using relative time, and API says to use absolute time...
	ACE_Time_Value wait_for_message(0, 1000); //wait 1ms
	
	
	struct timespec sleep_timer;
	
	sleep_timer.tv_sec = 0;
	//sleep_timer.tv_nsec = 1000000; //1ms
	sleep_timer.tv_nsec = 500000000; //500ms
	
	
	
	
	while(1)
	{
		if(this->getq(message, &wait_for_message) == -1)
		{
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Nothing in the rule management queue...performing house keeping\n"));
			//no explicit rule addition/creation is needed, so lets do some housekeeping
			
			client_rules->accept(&TCAPS_Rule_Management_Task::perform_housekeeping, (void *)client_rules);
		}
		else
		{
			//a rule needs to be added or removed from the rule list
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Something exists in the rule management queue... should deal with it\n"));

			if (message->msg_type() == ACE_Message_Block::MB_HANGUP)
        	{
          		this->putq (message);
          		break;
        	}
        	else
        	{
        		unsigned int num_serialized_bytes = 0;
        		
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) about to read a rmp from the rm_queue!\n"));
      			
      			ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
      			message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
      			
      			u8 serialized_bytes[num_serialized_bytes];
      			
      			ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
      			message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
      			
      			
      			struct tcaps_sp_rmp * packet = tcaps_sp_rmp_parse_packet(serialized_bytes, num_serialized_bytes);
      			
      			char tmp[BUFFER_SIZE];
				tcaps_sp_rmp_tostring(packet, tmp, sizeof(tmp));
				//printf("The packet we just received in the rm_thread looks like this:\n%s\n", tmp);
      			
      			char rule_command[BUFFER_SIZE];
      			//printf("\nAttempting to translate the RMP\n");
				tcaps_translate_rmp2qosss(packet, TCAPS_CI_INTERFACE, rule_command, packet->rule_number);
				printf("Translated RMP rule: %s\n", rule_command);

				TCAPS_Rule * found_rule = client_rules->get_rule(packet->rule_number);
				
				if(found_rule == NULL)
				{
					if(packet->op_type == TCAPS_RMP_OP_TYPE_DELETE)
					{
						//we've been told to delete a rule we don't have
						//can either ignore this situation or send back an ack packet with appropriate error code
						//as you can see, we are going with the former for the moment
					}
					else
					{
						tmp_rule = new TCAPS_Rule(packet);
						
						tmp_rule->set_rule_text(rule_command);
						
						//printf("Attempting to add the rule\n");
						tcaps_qosss_add_rule(rule_command, TCAPS_CI_INTERFACE);
						client_rules->add_rule(tmp_rule);
					}
				}
				else
				{
					if(packet->op_type == TCAPS_RMP_OP_TYPE_DELETE)
					{
						ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to explicitly delete rule %d! Deleting it...\n", found_rule->get_rule_num()));
						tcaps_qosss_delete_rule(found_rule->get_rule_text(), TCAPS_CI_INTERFACE);
						client_rules->remove_rule(found_rule->get_rule_num());
						ACE_DEBUG ((LM_DEBUG, "(%P|%t) Rule deleted\n"));
					}
					else
					{
						//we receive a packet corresponding to a rule already in our list, and we've been instructed to create/update the rule
						//which we interpret as increasing the time to live field by the new RMP timeout amount
						found_rule->increment_expiry_time(packet->timeout);
					}
				}
				
				tcaps_sp_rmp_destroy(packet);
        	}
        	
        	 message->release();
      	}
			
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for sleep_timer.tv_nsec nanosecs to be nice to other processes running on the system
	}


  return 0;
}

void TCAPS_Rule_Management_Task::perform_housekeeping(TCAPS_Rule * therule, void * args)
{
	if(therule->get_expiry_time() <= (unsigned int)ACE_OS::time())
	{
		TCAPS_Rule_List * rule_list = (TCAPS_Rule_List *)args;
		//if the timeout for this rule has expired, remove it from the rule list
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Found a rule that has expired! Deleting it...\n"));
		tcaps_qosss_delete_rule(therule->get_rule_text(), TCAPS_CI_INTERFACE);
		rule_list->remove_rule(therule->get_rule_num());
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Rule deleted\n"));
	}
}

void TCAPS_Rule_Management_Task::delete_qosss_rules(TCAPS_Rule * therule, void * args)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Deleting QoS SS rule: %s\n", therule->get_rule_text()));
	tcaps_qosss_delete_rule(therule->get_rule_text(), TCAPS_CI_INTERFACE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Rule deleted\n"));
}
