#include "tcaps_fc.h"


TCAPS_Network_Task * start_comms_thread(unsigned short local_port);
TCAPS_Classifier_Task * start_classifier_thread();

struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote);
void check_for_packets(TCAPS_Network_Task * comms_task);
void handle_ffp_packet(struct tcaps_sp_ffp * packet, ACE_INET_Addr * remote);
void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote);

void send_changed_flow_states(TCAPS_Network_Task * comms_task);
void generate_crmp(TCAPS_Flow * tmpflow, struct ip_address clientip, struct tcaps_sp_packet * packet);
void send_packet(struct tcaps_sp_packet * packet, struct ip_address recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task);

void cleanup();
void handler(int value);
static sig_atomic_t finished = 0;

TCAPS_Network_Task * fs_comms_task;
TCAPS_Network_Task * tm_comms_task;
TCAPS_Classifier_Task * classifier_task;


int main (int argc,char *argv[])
{
	struct timespec sleep_timer;
	
	fs_comms_task = NULL;
	tm_comms_task = NULL;
	classifier_task = NULL;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 500000000; //500ms
	
	
	ACE_Sig_Action sa ((ACE_SignalHandler) handler, SIGINT);

	
	fs_comms_task = start_comms_thread(TCAPS_FF_LOCAL_PORT);
	tm_comms_task = start_comms_thread(TCAPS_CRM_LOCAL_PORT);
	
	classifier_task = start_classifier_thread();
	
	fs_comms_task->open();
	tm_comms_task->open();
	classifier_task->open();

	while(!finished)
	{
        
		check_for_packets(fs_comms_task);
		check_for_packets(tm_comms_task);
		send_changed_flow_states(tm_comms_task);
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Killing some time in the main thread...\n"));
		//ACE_OS::sleep(1);
		ACE_OS::nanosleep(&sleep_timer); //sleep for 1ms to be nice to other processes running on the system
	}
	
	cleanup();

  return 0;
}


void cleanup()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) We are in the cleanup function!\n"));

	/* Once we're done, we have to signal the Task objects to shut
    down. There are several choices including: - Send a message of
    zero length - Send a message with a special content I don't like
    these choices because they're likely to interfere with application
    logic.  Instead, I use the message type feature to send a message
    of type "hangup".  The default type is MB_DATA, so when the tasks
    get a MB_HANGUP type, they know to go away.  */
	
	if(fs_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fs_comms_task queue\n"));
		fs_comms_task->putq(message);
		fs_comms_task->wait();
	}
	
	if(tm_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the tm_comms_task queue\n"));
		tm_comms_task->putq(message);
		tm_comms_task->wait();
	}
	
	if(classifier_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the classifier_task queue\n"));
		classifier_task->putq(message);
		classifier_task->wait();
	}

}

void handler(int value)
{
	finished = 1;
}


TCAPS_Classifier_Task * start_classifier_thread()
{
	TCAPS_Classifier_Task * classifier_task = NULL;
	
	TCAPS_Classification_Algo * algo = new TCAPS_Simple_Algo();

   	classifier_task = new TCAPS_Classifier_Task(algo);

  //if (classifier_task->open () == -1)
  	//ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "open"), -1);

	
	return classifier_task;
}

TCAPS_Network_Task * start_comms_thread(unsigned short local_port)
{
	TCAPS_Network_Task * comms_task = NULL;
	
	ACE_INET_Addr local(local_port);
	
	ACE_SOCK_Dgram * dgram = new ACE_SOCK_Dgram();
	
	if (dgram->open (local) != -1)
	{
		comms_task = new TCAPS_Network_Task(dgram);
	}
	
	return comms_task;
}



void send_packet(struct tcaps_sp_packet * packet, struct ip_address recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task)
{
	char iptmp[60], pktmp[1024];
	tcaps_ip_ntoa(&recipient_ip, iptmp, TRUE);	    
    ACE_INET_Addr remote(recipient_port, iptmp);
		
	/* Create a block large enough for our Data object as well as a
    text message.  */
  	Packet_Block *message;
  	u8 * serialized;
  	unsigned int serialized_bytes;
  	
  	switch(packet->packet_type)
  	{
//  		case TCAPS_RMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_rmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_rmp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_PP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_pp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_pp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_TP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_tp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_tp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_FFP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_ffp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_ffp *)packet->packet)->serialized;
//  			break;
  		case TCAPS_AP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_ap *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_ap *)packet->packet)->serialized;
  			break;
  		case TCAPS_CRMP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_crmp *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_crmp *)packet->packet)->serialized;
  			tcaps_sp_crmp_tostring(((struct tcaps_sp_crmp *)packet->packet), pktmp, sizeof(pktmp));
  			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Adding a CRMP to the queue\n%s\n", pktmp));
  			//printf("Adding a CRMP to the queue\n%s\n", pktmp);
  			break;
  		default:
  			serialized_bytes = 0;
  			serialized = 0;
  			break;
  	}

  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
  	
  	
  	message->copy ((const char *) &remote, sizeof(remote));
  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
  	message->copy ((const char *) serialized, serialized_bytes);

	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) adding a packet to the queue\n"));
	comms_task->send_packet(message);

}



struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote)
{
	ACE_Message_Block *message;
	struct tcaps_sp_packet * packet = NULL;
	ACE_Time_Value wait_for_message(1, 0);
	
	message = comms_task->receive_packet();
	
	if(message != NULL)
	{
		unsigned int num_serialized_bytes;
    		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to read a message from a comms_task!\n"));
		
		ACE_OS::memmove ((char *) remote, message->rd_ptr(), sizeof(ACE_INET_Addr));
		message->rd_ptr (sizeof(ACE_INET_Addr));  // Move the rd_ptr() beyond the data.
		
		ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
		message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
		
		u8 serialized_bytes[num_serialized_bytes];
		
		ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
		message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
		message->release();
		
		packet = (struct tcaps_sp_packet *)malloc(sizeof(struct tcaps_sp_packet));
		
		tcaps_sp_parse_packet((u8 *)serialized_bytes, num_serialized_bytes, packet);
	}
	
	return packet;

}


void check_for_packets(TCAPS_Network_Task * comms_task)
{
	ACE_INET_Addr remote;
	struct tcaps_sp_packet * packet = receive_packet(comms_task, &remote);
    
	while(packet != NULL)
	{
		switch(packet->packet_type)
		{
	    	case TCAPS_FFP_PACKET_TYPE:
				handle_ffp_packet((struct tcaps_sp_ffp * )packet->packet, &remote);
				break;
			case TCAPS_AP_PACKET_TYPE:
				handle_ap_packet((struct tcaps_sp_ap * )packet->packet, &remote);
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected tcaps_sp packet type\n"));
				break;
		}

		free(packet);
		
		packet = receive_packet(comms_task, &remote);
	}
}

void handle_ffp_packet(struct tcaps_sp_ffp * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	//tcaps_sp_ffp_tostring(packet, tmp, sizeof(tmp));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received a flow feature packet that looks like this\n%s", tmp));


	TCAPS_Flow * flow = new TCAPS_Flow( packet->flow_id.flow_id_type,
										packet->flow_id.clientip_version,
										tcaps_sp_ffp_field_ptr(packet, TCAPS_SP_FFP_POINT_TO_FLOW_ID_SRC_OCTETS),
										packet->num_flow_feature_segments,
										tcaps_sp_ffp_field_ptr(packet, TCAPS_SP_FFP_POINT_TO_START_OF_FLOW_FEATURES)
										);

										
	//flow->get_flow_id()->get_ascii_flow_id(tmp);
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Our TCAPS_Flow_ID looks like this %s\n", tmp));
	
	
	struct ip_address temp;
	temp.ip_version = packet->flow_id.clientip_version;
	temp.octets = tcaps_sp_ffp_field_ptr(packet, TCAPS_SP_FFP_POINT_TO_START_OF_CLIENT_IP_OCTETS);
	
	TCAPS_Client * client;
	
	if(classifier_task->get_clients()->find_client(&temp))
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Found the client in the client list already!\n"));
		client = classifier_task->get_clients()->get_client(&temp);
	}
	else
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Did NOT find the client in the client list!\n"));
		
		client = new TCAPS_Client(packet->flow_id.clientip_version,
								  tcaps_sp_ffp_field_ptr(packet, TCAPS_SP_FFP_POINT_TO_START_OF_CLIENT_IP_OCTETS),
								  packet->tm_ip.ip_version,
								  packet->tm_ip.octets
								 );
		classifier_task->get_clients()->add_client(client);
	}
	
	//add/update the flow to the client
	
	client->update_client_flow(flow);
										
	
	

	
	tcaps_sp_ffp_destroy(packet);
}


void send_changed_flow_states(TCAPS_Network_Task * tm_comms_task)
{
	std::map<struct ip_address,TCAPS_Client *,ipcmp>::iterator c_itr;
	std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp>::iterator f_itr;
	std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> f_list;
	unsigned int client_count=0, flow_count=0;
	TCAPS_Flow_List * client_flows;
	TCAPS_Flow * tmpflow;
	TCAPS_Client * tmpclient;
	char ffbuf[4096];
	struct tcaps_sp_packet packet;
	
	
	std::map<struct ip_address,TCAPS_Client *,ipcmp> c_list = classifier_task->get_clients()->get_client_list();
	
	for(c_itr = c_list.begin(); c_itr != c_list.end(); c_itr++, client_count++)
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing client %d\n", client_count+1));
		
		tmpclient = ((*c_itr).second);
		//tmpclient->to_string(ffbuf, sizeof(ffbuf));
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", ffbuf));
		
		client_flows = tmpclient->get_client_flows();
		
		f_list = client_flows->get_flow_list();
		
		for(f_itr = f_list.begin(), flow_count = 0; f_itr != f_list.end(); f_itr++, flow_count++)
		{
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing flow %d\n", flow_count+1));
			
			tmpflow = ((*f_itr).second);
			//tmpflow->to_string(ffbuf, sizeof(ffbuf));
			
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", ffbuf));
			
			if(tmpflow->state_changed())
			{
				tmpflow->to_string(ffbuf, sizeof(ffbuf));
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Found flow with changed state\n%s\n", ffbuf));
				
				generate_crmp(tmpflow, tmpclient->get_client_ip(), &packet);
				send_packet(&packet, tmpclient->get_tm_ip(), TCAPS_CRM_REMOTE_PORT, tm_comms_task);
				
				tcaps_sp_crmp_destroy((struct tcaps_sp_crmp *)packet.packet);
				
				tmpflow->reset_state_changed();
			}

		}
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished processing client %d\n\n\n", client_count+1));
		
	}
}

void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_ap_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an acknowledgment packet from %s that looks like this\n%s", remote->get_host_addr(), tmp));
}

void generate_crmp(TCAPS_Flow * tmpflow, struct ip_address clientip, struct tcaps_sp_packet * packet)
{
	char clientipbuf[60], srcipbuf[60], dstipbuf[60];
	u8 optype;
	
	tcaps_ip_ntoa(&clientip, clientipbuf, TRUE);
	tmpflow->get_flow_id()->get_flow_id_srcip(srcipbuf);
	tmpflow->get_flow_id()->get_flow_id_dstip(dstipbuf);
	packet->packet_type = TCAPS_CRMP_PACKET_TYPE;
		
	if(tmpflow->get_flow_priority() == 255) //255 is the lowest priority that can be assigned to a flow i.e. the flow is unprioritised
	{
		optype = TCAPS_RMP_OP_TYPE_DELETE;
	}
	else
	{
		optype = TCAPS_RMP_OP_TYPE_CREATE;
	}
	
	//printf("client ip: %s\tsrcip: %s\tdstip: %s\n", clientipbuf, srcipbuf, dstipbuf);
	
	if(strncmp(srcipbuf, clientipbuf, sizeof(clientipbuf)) == 0)
	{
		switch(tmpflow->get_flow_id()->get_flow_id_type())
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				packet->packet = tcaps_sp_crmp_create_ip2ip(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, optype, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), srcipbuf, dstipbuf);
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				packet->packet = tcaps_sp_crmp_create_ipport2ipport(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, optype, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), srcipbuf, ((TCAPS_Flow_ID_Port *)(tmpflow->get_flow_id()))->get_flow_id_src_port(), dstipbuf, ((TCAPS_Flow_ID_Port *)(tmpflow->get_flow_id()))->get_flow_id_dst_port());
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				packet->packet = tcaps_sp_crmp_create_ipspi2ipspi(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, TCAPS_RMP_OP_TYPE_CREATE, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), srcipbuf, ((TCAPS_Flow_ID_Spi *)tmpflow->get_flow_id())->get_flow_id_src_spi(), dstipbuf, ((TCAPS_Flow_ID_Spi *)tmpflow->get_flow_id())->get_flow_id_dst_spi());
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "\n(%P|%t) BIG PROBLEMS in generate_crmp()\n\n"));
				break;
		}
	}
	else if(strncmp(dstipbuf, clientipbuf, sizeof(clientipbuf)) == 0)
	{
		switch(tmpflow->get_flow_id()->get_flow_id_type())
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				packet->packet = tcaps_sp_crmp_create_ip2ip(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, optype, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), dstipbuf, srcipbuf);
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				packet->packet = tcaps_sp_crmp_create_ipport2ipport(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, optype, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), dstipbuf, ((TCAPS_Flow_ID_Port *)(tmpflow->get_flow_id()))->get_flow_id_dst_port(), srcipbuf, ((TCAPS_Flow_ID_Port *)(tmpflow->get_flow_id()))->get_flow_id_src_port());
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				packet->packet = tcaps_sp_crmp_create_ipspi2ipspi(clientip.ip_version, clientipbuf, "00:00:00:00:00:00", 0, 0, 0, TCAPS_RMP_OP_TYPE_CREATE, tmpflow->get_flow_priority(), 1, tmpflow->get_flow_id()->get_flow_id_ip_version(), dstipbuf, ((TCAPS_Flow_ID_Spi *)tmpflow->get_flow_id())->get_flow_id_dst_spi(), srcipbuf, ((TCAPS_Flow_ID_Spi *)tmpflow->get_flow_id())->get_flow_id_src_spi());
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "\n(%P|%t) BIG PROBLEMS in generate_crmp()\n\n"));
				break;
		}
	}
	
}

