#include "tcaps_fc.h"

/* Set our housekeeping pointer to NULL and tell the user we exist.  */
TCAPS_Classifier_Task::TCAPS_Classifier_Task (TCAPS_Classification_Algo * a)
  : n_threads_ (1), algo (a)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Classifier_Task ctor 0x%x\n", (void *) this));
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Classifier_Task::~TCAPS_Classifier_Task (void)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Classifier_Task dtor 0x%x\n", (void *) this));

  /* Get our shutdown notification out of the queue and release it.  */
  ACE_Message_Block *message;

  /* Like the getq() in svc() below, this will block until a message
    arrives.  By blocking, we know that the destruction will be paused
    until the last thread is done with the message block.  */
  this->getq (message);
  message->release ();
  
  delete algo;
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int
TCAPS_Classifier_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP,
                         n_threads_);
}

/* Tell the user we're closing and invoke the baseclass' close() to
  take care of things.  */
int
TCAPS_Classifier_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG,
              "(%P|%t) TCAPS_Classifier_Task close 0x%x\n",
              (void *) this));
  return inherited::close (flags);
}

/* Our svc() method waits for work on the queue and then processes
  that work.  */
int
TCAPS_Classifier_Task::svc (void)
{
	ACE_Message_Block *message;
	//char packbuf[BUFFER_SIZE];
	//char buf[BUFFER_SIZE];
	//char tmp[BUFFER_SIZE];
	unsigned int ffbuf_size = 10000;
	char * ffbuf = (char *)malloc(ffbuf_size);
	//this may be foobed as we are using relative time, and API says to use absolute time...
	ACE_Time_Value wait_for_message(0, 1); //wait 1us
	
	
	struct timespec sleep_timer;
	
	sleep_timer.tv_sec = 0;
	//sleep_timer.tv_nsec = 1000; //1us
	sleep_timer.tv_nsec = 500000000; //500ms

	

	ACE_DEBUG ((LM_DEBUG, "(%P|%t) starting up Classifier_Task thread\n"));
		
		
	while(1)
	{
		if(this->getq(message, &wait_for_message) == -1)
		{
			//no new clients/client flows in the queue, keep processing existing client flows
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) No messages in queue. Performing flow classification\n"));
			
			std::map<struct ip_address,TCAPS_Client *,ipcmp>::iterator c_itr;
			std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp>::iterator f_itr;
			unsigned int client_count=0, flow_count=0;
			TCAPS_Flow_List * client_flows;
			TCAPS_Flow * tmpflow;
			TCAPS_Client * tmpclient;
			
			client_list.acquire_mutex();
			
			std::map<struct ip_address,TCAPS_Client *,ipcmp> c_list = client_list.get_client_list();
			
			for(c_itr = c_list.begin(); c_itr != c_list.end(); c_itr++, client_count++)
			{
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing client %d\n", client_count+1));
				
				tmpclient = ((*c_itr).second);
				tmpclient->to_string(ffbuf, ffbuf_size);
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", ffbuf));
				
				client_flows = tmpclient->get_client_flows();
				
				client_flows->acquire_mutex();
				
				std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> f_list = client_flows->get_flow_list();
				
				for(f_itr = f_list.begin(), flow_count = 0; f_itr != f_list.end(); f_itr++, flow_count++)
				{
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing flow %d\n", flow_count+1));
					
					tmpflow = ((*f_itr).second);
					tmpflow->to_string(ffbuf, ffbuf_size);
					
					//printf("\n%s\n\n", ffbuf);
					
					algo->classify_flow(tmpflow);
				}
				
				client_flows->release_mutex();
				
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished processing client %d\n\n\n", client_count+1));
				
			}
			
			client_list.release_mutex();
		}
		else
		{
			//We have something in the queue!
			
			//Check to see if it is an ACE hangup message
			if (message->msg_type() == ACE_Message_Block::MB_HANGUP)
	    	{
	    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to shutdown the Classifier_Task thread\n"));
	      		this->putq (message);
	      		break; //break out of the while loop - we are done
	    	}
	    	else
	    	{
	    		
	    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected message in the Classifier_Task message queue!\n"));
				message->release();

	    	}
		}
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for sleep_timer.tv_nsec nanosecs to be nice to other processes running on the system
		
	}


	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Exiting svc() in Classifier_Task thread\n"));
	
	return 0;
}


TCAPS_Client_List * TCAPS_Classifier_Task::get_clients()
{
	return &client_list;
}
