#ifndef _TCAPS_FC_CLIENT_LIST_H_
#define _TCAPS_FC_CLIENT_LIST_H_

#include <list>
#include <map>

#include "ace/Synch.h"

typedef ACE_Thread_Mutex mutex_t;
typedef std::pair<struct ip_address,TCAPS_Client *> client_entry;


struct ipcmp
{
	bool operator()(const struct ip_address &a, const struct ip_address &b) const
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_List.h in the compare function\n"));
		
		if(a.ip_version == b.ip_version)
		{
			if(TCAPS_SP_IP_VERSION_4 == a.ip_version)
    			return (memcmp(a.octets, b.octets, NUM_IPV4_OCTETS) < 0);
    		else if(TCAPS_SP_IP_VERSION_6 == a.ip_version)
    			return (memcmp(a.octets, b.octets, NUM_IPV6_OCTETS) < 0);
		}
		
		return (a.ip_version < b.ip_version); //IPv4 addresses will be "less" than IPv6 addresses

	}
}; 

class TCAPS_Client_List
{
public:
	TCAPS_Client_List ();
	TCAPS_Client_List (const TCAPS_Client_List &);
	virtual ~TCAPS_Client_List (void);

	//std::list<TCAPS_Client> get_client_list() const;
	std::map<struct ip_address,TCAPS_Client *,ipcmp> get_client_list() const;
	void add_client(TCAPS_Client * theclient);
	void acquire_mutex();
	void release_mutex();
	bool find_client(struct ip_address * clientip);
	TCAPS_Client * get_client(struct ip_address * clientip);
	
	TCAPS_Client_List & operator=(const TCAPS_Client_List &copied_client_list);
	
	friend bool operator==(const TCAPS_Client_List &f1, const TCAPS_Client_List &f2);

	
  
private:
	mutex_t client_list_mutex; //client_list_mutex.acquire(); client_list_mutex.release();
	//std::list<TCAPS_Client> client_list;
	std::map<struct ip_address,TCAPS_Client *,ipcmp> client_list;
};

#endif //_TCAPS_FC_CLIENT_LIST_H_
