#include "tcaps_fc.h"

static void pad(char * string, unsigned int len, unsigned int n);

TCAPS_Flow::TCAPS_Flow ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow default ctor 0x%x\n", (void *) this));
}

TCAPS_Flow::TCAPS_Flow (u8 fid_type, u8 ipversion, u8 * bytes, u16 num_serialized_features, u8 * serialized_features)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow ctor 0x%x\n", (void *) this));
	
	flow_id = NULL;
	
	switch(fid_type)
	{
		case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
			flow_id = new TCAPS_Flow_ID(TCAPS_FFP_FLOW_ID_TYPE_IPIP, ipversion, bytes);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
			flow_id = new TCAPS_Flow_ID_Port(TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT, ipversion, bytes);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
			flow_id = new TCAPS_Flow_ID_Spi(TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI, ipversion, bytes);
			break;
		default:
			//something is fooked
			break;
	}
	
	flow_priority = powf((float)2, (float)(8*sizeof(flow_priority)))-1;
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow priority set to: %d\n", flow_priority));
	
	flow_state_changed = false;
	
	u16 fbytes = sizeof(tcaps_flow_feature_type)+sizeof(tcaps_flow_feature);
	u16 * ftype;
	float * f;
	
	for(u16 i = 0; i < num_serialized_features; i++)
	{
		ftype = (u16 *)(serialized_features+(i*fbytes));
		f = (float *)(serialized_features+(i*fbytes)+sizeof(tcaps_flow_feature_type));
		flow_features.insert(feature_entry(*ftype, *f));
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added flow feature of type: %d\twith value: %0.2f\n", *ftype, *f));
	}
}

TCAPS_Flow::TCAPS_Flow (const TCAPS_Flow &copied_flow)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow copy ctor 0x%x\n", (void *) this));
	
	flow_id = copied_flow.get_flow_id();
	flow_features = copied_flow.get_flow_features();
}

TCAPS_Flow::~TCAPS_Flow (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow dtor 0x%x\n", (void *) this));
	
	if(flow_id != NULL)
		delete flow_id;
}
	
TCAPS_Flow & TCAPS_Flow::operator=(const TCAPS_Flow & copied_flow)
{
	if(this == &copied_flow)
		return *this;
		
	flow_id = copied_flow.get_flow_id();
	flow_features = copied_flow.get_flow_features();
	
	return *this;
}

tcaps_flow_feature TCAPS_Flow::get_flow_feature(tcaps_flow_feature_type type)
{
	std::map<tcaps_flow_feature_type,tcaps_flow_feature>::iterator ff_iter;
	
	ff_iter = flow_features.find(type);
	
	if(ff_iter != flow_features.end())
		return ((*ff_iter).second);
	else
		return -1;
}

TCAPS_Flow_ID * TCAPS_Flow::get_flow_id() const
{
	return flow_id;
}

std::map<tcaps_flow_feature_type,tcaps_flow_feature> TCAPS_Flow::get_flow_features() const
{
	return flow_features;
}

void TCAPS_Flow::set_flow_features(std::map<tcaps_flow_feature_type,tcaps_flow_feature> ff)
{
	flow_features = ff;
}

void TCAPS_Flow::set_flow_priority(u8 priority)
{
	if(flow_priority != priority)
	{
		flow_mutex.acquire();
		flow_state_changed = true;
		flow_mutex.release();
	}
		
	flow_priority = priority;
}

u8 TCAPS_Flow::get_flow_priority() const
{
	return flow_priority;
}

bool TCAPS_Flow::state_changed() const
{
	return flow_state_changed;
}

void TCAPS_Flow::reset_state_changed()
{
	flow_mutex.acquire();
	flow_state_changed = false;
	flow_mutex.release();
}


void TCAPS_Flow::to_string(char * buf, unsigned int bufsize)
{
	flow_id->get_ascii_flow_id(buf);
	
//	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\nflow_priority: %d\nflow_state_changed: %d\n", flow_priority, flow_state_changed);
//	
//	std::map<tcaps_flow_feature_type,tcaps_flow_feature>::iterator ff_itr;
//	
//	unsigned int padamt = 0;
//	
//	for(ff_itr = flow_features.begin(); ff_itr != flow_features.end(); ff_itr++)
//	{
//		padamt = strlen(flow_feature::get_flow_feature_description(ff_itr->first)) > 100 ? 1 : (100-strlen(flow_feature::get_flow_feature_description(ff_itr->first)));
//		//snprintf(buf+strlen(buf), bufsize-strlen(buf), "flow_feature_type: %d\tflow_feature_value: %0.2f\n", ff_itr->first, ff_itr->second);
//		snprintf(buf+strlen(buf), bufsize-strlen(buf), "\n%s:", flow_feature::get_flow_feature_description(ff_itr->first));
//		pad(buf+strlen(buf), bufsize-strlen(buf), padamt);
//		snprintf(buf+strlen(buf), bufsize-strlen(buf), "%0.2f", ff_itr->second);
//	}
}

static void pad(char * string, unsigned int len, unsigned int n)
{
	for(unsigned int i = 0; i < n; i++)
	{
		snprintf(string+strlen(string), len-strlen(string), " ");
	}
} 
	
bool operator<(const TCAPS_Flow &f1, const TCAPS_Flow &f2)
{
	return (f1.flow_id < f2.flow_id);
}

bool operator>(const TCAPS_Flow &f1, const TCAPS_Flow &f2)
{
	return (f1.flow_id > f2.flow_id);
}

bool operator==(const TCAPS_Flow &f1, const TCAPS_Flow &f2)
{
	return (f1.flow_id == f2.flow_id);
}
