#ifndef _TCAPS_FC_FLOW_H_
#define _TCAPS_FC_FLOW_H_

#include <map>
#include "ace/Synch.h"

typedef ACE_Thread_Mutex mutex_t;
typedef std::pair<tcaps_flow_feature_type,tcaps_flow_feature> feature_entry;

class TCAPS_Flow
{
public:
	TCAPS_Flow ();
	TCAPS_Flow (u8 fid_type, u8 ipversion, u8 * bytes, u16 num_serialized_features, u8 * serialized_features);
	TCAPS_Flow (const TCAPS_Flow &);
	virtual ~TCAPS_Flow (void);
	
	TCAPS_Flow_ID * get_flow_id() const;
	std::map<tcaps_flow_feature_type,tcaps_flow_feature> get_flow_features() const;
	void set_flow_features(std::map<tcaps_flow_feature_type,tcaps_flow_feature> ff);
	void set_flow_priority(u8 priority);
	tcaps_flow_feature get_flow_feature(tcaps_flow_feature_type type);
	u8 get_flow_priority() const;
	bool state_changed() const;
	void reset_state_changed();
	void to_string(char * buf, unsigned int bufsize);
	
	TCAPS_Flow & operator=(const TCAPS_Flow & copied_flow);
	
	friend bool operator<(const TCAPS_Flow &f1, const TCAPS_Flow &f2);
	friend bool operator>(const TCAPS_Flow &f1, const TCAPS_Flow &f2);
	friend bool operator==(const TCAPS_Flow &f1, const TCAPS_Flow &f2);

	
  
private:
	mutex_t flow_mutex;
	TCAPS_Flow_ID * flow_id;
	bool flow_state_changed;
	u8 flow_priority;
	std::map<tcaps_flow_feature_type,tcaps_flow_feature> flow_features;
	
  
};

#endif //_TCAPS_FC_FLOW_H_
