#ifndef TCAPS_FC_FLOW_FEATURES_H_
#define TCAPS_FC_FLOW_FEATURES_H_

//allows 2^16 features
typedef u16 tcaps_flow_feature_type;
typedef float tcaps_flow_feature;


//"since flow beginning" features
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS												1
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PACKETS									2
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PACKETS									3

#define TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_BIDIR_MS									4
#define TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS						5
#define TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS						6

#define TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS						7
#define TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS						8

#define TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS						9
#define TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS						10

#define TCAPS_FLOW_FEATURE_TOTAL_ACTIVE_TIME_SECS												11

#define TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PACKETS_PER_SECOND										12
#define TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PACKETS_PER_SECOND							13
#define TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PACKETS_PER_SECOND							14

#define TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PAYLOAD_BYTES										15
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PAYLOAD_BYTES								16
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PAYLOAD_BYTES								17

#define TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_SECOND								18
#define TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_SECOND						19
#define TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_SECOND						20

#define TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_PACKET								21
#define TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET						22
#define TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET						23

#define TCAPS_FLOW_FEATURE_MINIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET						24
#define TCAPS_FLOW_FEATURE_MINIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET						25

#define TCAPS_FLOW_FEATURE_MAXIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET						26
#define TCAPS_FLOW_FEATURE_MAXIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET						27

#define TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET				28
#define TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET				29
#define TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET				30
#define TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET				31
#define TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET				32
#define TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET				33
#define TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET				34
#define TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET				35
#define TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET				36
#define TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET				37

#define TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET		38
#define TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET		39
#define TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET		40
#define TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET		41
#define TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET		42
#define TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET		43
#define TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET		44
#define TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET		45
#define TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET		46
#define TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET		47


//sliding window based features
#define TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS_WINDOW										100



namespace flow_feature
{
	//char * get_flow_feature_description(u16 flow_feature_type);
	
	static char * get_flow_feature_description(u16 flow_feature_type)
	{
		switch(flow_feature_type)
		{
			//"since flow beginning" features
			
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS";
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PACKETS:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PACKETS";
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PACKETS:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PACKETS";
				
			case TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_BIDIR_MS:
				return "TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_BIDIR_MS";
			case TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS:
				return "TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS";
			case TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS:
				return "TCAPS_FLOW_FEATURE_AVERAGE_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS";
				
			case TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS:
				return "TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS";
			case TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS:
				return "TCAPS_FLOW_FEATURE_MINIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS";

			case TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS:
				return "TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_CLIENT_TO_OTHER_MS";
			case TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS:
				return "TCAPS_FLOW_FEATURE_MAXIMUM_INTER_ARRIVAL_TIME_OTHER_TO_CLIENT_MS";
				
			case TCAPS_FLOW_FEATURE_TOTAL_ACTIVE_TIME_SECS:
				return "TCAPS_FLOW_FEATURE_TOTAL_ACTIVE_TIME_SECS";
				
			case TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PACKETS_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PACKETS_PER_SECOND";
			case TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PACKETS_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PACKETS_PER_SECOND";
			case TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PACKETS_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PACKETS_PER_SECOND";
				
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PAYLOAD_BYTES:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PAYLOAD_BYTES";
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PAYLOAD_BYTES:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_CLIENT_TO_OTHER_PAYLOAD_BYTES";
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PAYLOAD_BYTES:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_OTHER_TO_CLIENT_PAYLOAD_BYTES";
				
			case TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_SECOND";
			case TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_SECOND";
			case TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_SECOND:
				return "TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_SECOND";
				
			case TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_AVERAGE_BIDIR_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_AVERAGE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_AVERAGE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
				
			case TCAPS_FLOW_FEATURE_MINIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_MINIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_MINIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_MINIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";

			case TCAPS_FLOW_FEATURE_MAXIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_MAXIMUM_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_MAXIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_MAXIMUM_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
				
			case TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
			case TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_PAYLOAD_BYTES_PER_PACKET";
			
			case TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_25TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_50TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_75TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_90TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_95TH_PERCENTILE_CLIENT_TO_OTHER_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_25TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_50TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_75TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_90TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET";
			case TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET:
				return "TCAPS_FLOW_FEATURE_95TH_PERCENTILE_OTHER_TO_CLIENT_INTER_ARRIVAL_TIME_MS_PER_PACKET";
				
				
				
				
				
			//sliding window based features
			case TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS_WINDOW:
				return "TCAPS_FLOW_FEATURE_TOTAL_NUM_BIDIR_PACKETS_WINDOW";
				
				
				
				
				
			default:
				return "UNKNOWN_FLOW_FEATURE_TYPE";
		}
	}
}

#endif /*TCAPS_FC_FLOW_FEATURES_H_*/
