#include "tcaps_fc.h"


TCAPS_Flow_List::TCAPS_Flow_List ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List default ctor 0x%x\n", (void *) this));
}

TCAPS_Flow_List::TCAPS_Flow_List (const TCAPS_Flow_List &copied_flow_list)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List copy ctor 0x%x\n", (void *) this));
	
	flow_list = copied_flow_list.get_flow_list();
}

TCAPS_Flow_List::~TCAPS_Flow_List (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List dtor 0x%x\n", (void *) this)); 
}
	
TCAPS_Flow_List & TCAPS_Flow_List::operator=(const TCAPS_Flow_List &copied_flow_list)
{
	if(this == &copied_flow_list)
		return *this;
		
	flow_list = copied_flow_list.get_flow_list();
	
	return *this;
}


//std::list<TCAPS_Flow> TCAPS_Flow_List::get_flow_list() const
//{
//	return flow_list;
//}

std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> TCAPS_Flow_List::get_flow_list() const
{
	return flow_list;
}

void TCAPS_Flow_List::add_flow(TCAPS_Flow * theflow)
{
	flow_list_mutex.acquire();
	
	//add flow to the list
	
	//flow_list.push_back(*theflow);
	
	//flow_list.insert(flow_entry(*(theflow->get_flow_id()), theflow));
	
	flow_list[*(theflow->get_flow_id())] = theflow;

	flow_list_mutex.release();
}

void TCAPS_Flow_List::acquire_mutex()
{
	flow_list_mutex.acquire();
}

void TCAPS_Flow_List::release_mutex()
{
	flow_list_mutex.release();
}

bool TCAPS_Flow_List::find_flow(TCAPS_Flow_ID * fid)
{
	std::map<TCAPS_Flow_ID,TCAPS_Flow *>::iterator flow_iter;
	
	flow_iter = flow_list.find(*fid);
	   				 
	if (flow_iter != flow_list.end())
	{
		return true;
	}
	
	TCAPS_Flow_ID * fid2 = get_flowid_reversed(fid);
	
	bool ret = false;
		
	flow_iter = flow_list.find(*fid2);
	   				 
	if(flow_iter != flow_list.end())
	{
		ret = true;
	}
	
	delete fid2;
	
	return ret;
}

TCAPS_Flow * TCAPS_Flow_List::get_flow(TCAPS_Flow_ID * fid)
{
	std::map<TCAPS_Flow_ID,TCAPS_Flow *>::iterator flow_iter;
	
	flow_iter = flow_list.find(*fid);
	   				 
	if (flow_iter != flow_list.end())
	{
		return ((*flow_iter).second);
	}
	
	TCAPS_Flow_ID * fid2 = get_flowid_reversed(fid);
	
	TCAPS_Flow * tmpflow = NULL;
		
	flow_iter = flow_list.find(*fid2);
	   				 
	if (flow_iter != flow_list.end())
	{
		tmpflow = ((*flow_iter).second);
	}

	delete fid2;
	
	return tmpflow;
}

TCAPS_Flow_ID * TCAPS_Flow_List::get_flowid_reversed(const TCAPS_Flow_ID * fid)
{
	unsigned int ip_size = 0, fid_bytes = fid->get_num_flow_id_bytes();
	u8 fid_ip_version = fid->get_flow_id_ip_version();
	u8 fid_type = fid->get_flow_id_type();
	TCAPS_Flow_ID * ret;
	
	const u8 * temp = fid->get_binary_flow_id();
	u8 * temp2 = (u8 *)malloc(fid_bytes);

	char fidbuf[50], fid2buf[50];
	
	if(fid_ip_version == TCAPS_SP_IP_VERSION_4)
		ip_size = NUM_IPV4_OCTETS;
	else
		ip_size = NUM_IPV6_OCTETS;

		
	u8 * temp3 = (u8 *)malloc(ip_size);
	
	//this set of memcpy's effectively creates a new binary flow id that has the src and dst IPs (& PORT/SPI) swapped
	
	//copy the src IP (& PORT/SPI) from fid to holding mem	
	memcpy(temp3, temp, ip_size);
	
	//copy the original binary flow id from fid to our new binary flow id memory location
	memcpy(temp2, temp, fid_bytes);
	
	//copy fid's dst IP (& PORT/SPI) over the top of the new flow id's src IP
	memcpy(temp2, temp+ip_size, ip_size);
	
	//copy fid's src IP (& PORT/SPI) over the top of the new flow id's dst IP
	memcpy(temp2+ip_size, temp3, ip_size);
	
	if(fid_type == TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT)
	{
		//copy the src PORT from fid to holding mem	
		memcpy(temp3, temp+(2*ip_size), sizeof(u16));
		
		//copy fid's dst PORT over the top of the new flow id's src PORT
		memcpy(temp2+(2*ip_size), temp+((2*ip_size)+sizeof(u16)), sizeof(u16));
		
		//copy fid's src PORT over the top of the new flow id's dst PORT
		memcpy(temp2+((2*ip_size)+sizeof(u16)), temp3, sizeof(u16));
		
		ret = new TCAPS_Flow_ID_Port(fid_type, fid_ip_version, temp2);
	}

	else if(fid_type == TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI)
	{
		//copy the src SPI from fid to holding mem	
		memcpy(temp3, temp+(2*ip_size), sizeof(u32));
		
		//copy fid's dst SPI over the top of the new flow id's src SPI
		memcpy(temp2+(2*ip_size), temp+((2*ip_size)+sizeof(u32)), sizeof(u32));
		
		//copy fid's src SPI over the top of the new flow id's dst SPI
		memcpy(temp2+((2*ip_size)+sizeof(u32)), temp3, sizeof(u32));
		
		ret = new TCAPS_Flow_ID_Spi(fid_type, fid_ip_version, temp2);
	}
	
	else
	{
		ret = new TCAPS_Flow_ID(fid_type, fid_ip_version, temp2);
	}
	
	printHexBytes(temp, fid_bytes, fidbuf, sizeof(fidbuf), TRUE);
	printHexBytes(temp2, fid_bytes, fid2buf, sizeof(fid2buf), TRUE);
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List fid_bytes: %s\n", fidbuf));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List fid2_bytes: %s\n", fid2buf));
	
	return ret;
}

bool operator==(const TCAPS_Flow_List &f1, const TCAPS_Flow_List &f2)
{
//	if(//lists are the same size)
//	{
//		if(each flow in the list is same)
//		{
//			return true;
//		}
//	}
	
	return false;
}
