#ifndef _TCAPS_FC_FLOW_LIST_H_
#define _TCAPS_FC_FLOW_LIST_H_


#include <map>

#include "ace/Synch.h"

typedef ACE_Thread_Mutex mutex_t;
typedef std::pair<TCAPS_Flow_ID,TCAPS_Flow *> flow_entry;

struct flowcmp
{
	bool operator()(const TCAPS_Flow_ID &a, const TCAPS_Flow_ID &b) const
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_List.h in the compare function\n"));
		
		if(a.get_flow_id_type() == b.get_flow_id_type())
		{
			return (a < b);
		}
		
		return (a.get_flow_id_type() < b.get_flow_id_type()); //IPIP flows < IPIPPORT flows < IPIPSPI flows

	}
}; 

class TCAPS_Flow_List
{
public:
	TCAPS_Flow_List ();
	TCAPS_Flow_List (const TCAPS_Flow_List &);
	virtual ~TCAPS_Flow_List (void);

	//std::list<TCAPS_Flow> get_flow_list() const;
	std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> get_flow_list() const;
	void add_flow(TCAPS_Flow * theflow);
	bool find_flow(TCAPS_Flow_ID * fid);
	TCAPS_Flow * get_flow(TCAPS_Flow_ID * fid);
	void acquire_mutex();
	void release_mutex();
	
	TCAPS_Flow_List & operator=(const TCAPS_Flow_List &copied_flow_list);
	
	friend bool operator==(const TCAPS_Flow_List &f1, const TCAPS_Flow_List &f2);

	
  
private:
	TCAPS_Flow_ID * get_flowid_reversed(const TCAPS_Flow_ID * fid);
	
	mutex_t flow_list_mutex; //flow_list_mutex.acquire(); flow_list_mutex.release();
	//std::list<TCAPS_Flow> flow_list;
	std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> flow_list;

};


#endif //_TCAPS_FC_FLOW_LIST_H_
