#include "tcaps_fs.h"

TCAPS_Network_Task * start_comms_thread(unsigned short local_port);

TCAPS_Sniff_Task * start_sniff_thread(pcap_t * sniff, TCAPS_Flow_Manager_Task * fm_task);
TCAPS_Flow_Manager_Task * start_flow_manager_thread();

void send_state_to_fc(TCAPS_Network_Task * tx_task);

void send_packet(struct tcaps_sp_packet * packet, struct ip_address * recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * tx_task);


struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote);
void check_for_packets(TCAPS_Network_Task * comms_task);
void handle_cmp_packet(struct tcaps_sp_cmp * packet, ACE_INET_Addr * remote);
void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote);


void cleanup();
void handler(int value);
static sig_atomic_t finished = 0;




TCAPS_Network_Task * tm_comms_task;
TCAPS_Network_Task * fc_comms_task;


TCAPS_Flow_Manager_Task * fm_task;
TCAPS_Sniff_Task * sniff_task;

	
	

int main (int argc,char *argv[])
{
	struct timespec sleep_timer;
	char pcap_ebuf[1024];
	
	tm_comms_task = NULL;
	fc_comms_task = NULL;
	fm_task = NULL;
	sniff_task = NULL;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 500000000; //.5s
	
	
	ACE_Sig_Action sa ((ACE_SignalHandler) handler, SIGINT);
	
	pcap_t * sniff = pcap_open_live(TCAPS_FS_SNIFF_INTERFACE, 0, 1, 1000, pcap_ebuf);
	
	if(sniff == NULL)
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Unable to start pcap sniffing session.\nReason: %s\nExiting...\n", pcap_ebuf));
		exit(1);
	}


	tm_comms_task = start_comms_thread(TCAPS_TM_LOCAL_PORT);
	fc_comms_task = start_comms_thread(TCAPS_FC_LOCAL_PORT);
	
	
	
	fm_task = start_flow_manager_thread();


	sniff_task = start_sniff_thread(sniff, fm_task);
	
	

	tm_comms_task->open();
	fc_comms_task->open();

	ACE_OS::nanosleep(&sleep_timer);
	fm_task->open();
	ACE_OS::nanosleep(&sleep_timer);
	sniff_task->open();
	ACE_OS::nanosleep(&sleep_timer);
	
	
	

	while(!finished)
	{
        
		send_state_to_fc(fc_comms_task);
		check_for_packets(tm_comms_task);
		check_for_packets(fc_comms_task);
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Killing some time in the main thread...\n"));

		ACE_OS::nanosleep(&sleep_timer); //sleep for 1ms to be nice to other processes running on the system
	}
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) We are out of the loop in the main thread...\n"));
	
	cleanup();

	return 0;
}


void cleanup()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) We are in the cleanup function!\n"));

	/* Once we're done, we have to signal the Task objects to shut
    down. There are several choices including: - Send a message of
    zero length - Send a message with a special content I don't like
    these choices because they're likely to interfere with application
    logic.  Instead, I use the message type feature to send a message
    of type "hangup".  The default type is MB_DATA, so when the tasks
    get a MB_HANGUP type, they know to go away.  */
	

	if(tm_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the tm_comms_task queue\n"));
		tm_comms_task->putq(message);
		tm_comms_task->wait();
	}
	
	if(fc_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fc_comms_task queue\n"));
		fc_comms_task->putq(message);
		fc_comms_task->wait();
	}
	
	if(sniff_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the sniff_task queue\n"));
		sniff_task->putq(message);
		pcap_breakloop(sniff_task->get_pcap_handle());
		sniff_task->wait();
	}
	
	if(fm_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type (ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fm_task queue\n"));
		fm_task->putq(message);
		fm_task->wait();
	}

	
}

void handler(int value)
{
	//cleanup();
  	finished = 1;
	
	//exit(value);
}

void send_state_to_fc(TCAPS_Network_Task * tx_task)
{
	static unsigned int send_timer = 0;
	
	if(send_timer <= (unsigned int)ACE_OS::time())
	{
		//printf("poll_timer = %d, time = %d\n", poll_timer, (unsigned int)ACE_OS::time());
		
		TCAPS_Client_List * client_list = fm_task->get_client_list();
		
		char macbuf[64], packet_temp[4096], fc_ip_buf[64], tm_ip_buf[64], client_ip_buf[64];
		struct tcaps_sp_packet packet, packet2;
		struct ip_address fc_ip, tm_ip;
		unsigned int num_features = 0;
		struct mac_address mac;
		std::map<struct ip_address,TCAPS_Client *,ipcmp>::iterator c_itr;
		std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp>::iterator f_itr;
		unsigned int client_count=0, flow_count=0;
		TCAPS_Flow_List * client_flows;
		TCAPS_Flow * tmpflow;
		TCAPS_Client * tmpclient;
		char ffbuf[10000];
		u8 whichisclient;
		
		tcaps_sp_get_interface_mac(&mac, TCAPS_FS_COMMS_INTERFACE);
		
		printHexBytes(mac.octets, 6, macbuf, sizeof(macbuf), TRUE);
		
		
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Setting fc_ip\n"));
//		//manually setting the FC IP for testing
//		fc_ip.ip_version = TCAPS_SP_IP_VERSION_4;
//		fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
//		tcaps_ip_aton("172.16.7.8", &fc_ip);
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) fc_ip set\n"));
		
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to perform a send_state_to_fc run\n"));
		
		//we only fiddle with the client list from the main (this) thread, therefore we don't need to acquire the
		//client list mutex. However, because we add flows to client's in the flow_manager_task, we need to lock the 
		//flows list for each client we process
		//client_list->acquire_mutex();
	    		
		std::map<struct ip_address,TCAPS_Client *,ipcmp> c_list = client_list->get_client_list();
		
		packet.packet_type = TCAPS_FFP_PACKET_TYPE;
		
		for(c_itr = c_list.begin(); c_itr != c_list.end(); c_itr++, client_count++)
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing client %d\n", client_count+1));
			
			
			
			tmpclient = ((*c_itr).second);
			
			struct ip_address tmpip = tmpclient->get_client_ip();
	   		tcaps_ip_ntoa(&tmpip, client_ip_buf, TRUE);
			
			tmpclient->to_string(ffbuf, sizeof(ffbuf));
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", ffbuf));
			
			client_flows = tmpclient->get_client_flows();
			
			tm_ip = tmpclient->get_tm_ip();
			tcaps_ip_ntoa(&tm_ip, tm_ip_buf, TRUE);
			fc_ip = tmpclient->get_fc_ip();
			
			#ifdef _MUTEX_DEBUG_
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) send_state_to_fc() acquiring flow list mutex for client %s\n", client_ip_buf));
			#endif
			
			client_flows->acquire_mutex();
			
			#ifdef _MUTEX_DEBUG_
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) send_state_to_fc() flow list mutex acquired for client %s\n", client_ip_buf));
			#endif
			
			std::map<TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> f_list = client_flows->get_flow_list();
			
			for(f_itr = f_list.begin(), flow_count = 0; f_itr != f_list.end(); f_itr++, flow_count++)
			{
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Processing flow %d\n", flow_count+1));

				tmpflow = ((*f_itr).second);
				
				if(tmpflow->state_changed())
				{
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to call calculate_window_flow_features()\n"));
					tmpflow->calculate_window_flow_features();
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished calculate_window_flow_features()\n"));
					
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to call calculate_histogram_flow_features()\n"));
					tmpflow->calculate_histogram_flow_features();
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished calculate_histogram_flow_features()\n"));
					
					tmpflow->to_string(ffbuf, sizeof(ffbuf));
					//printf("\n%s\n\n", ffbuf);
					
					u8 * serialized = tmpflow->serialize_flow_features(&num_features);
					
					struct ip_address * srcip = tmpflow->get_flow_id()->get_flow_id_srcip();
					
					if(tmpclient->get_client_ip().ip_version == TCAPS_SP_IP_VERSION_4)
					{
						if(memcmp(srcip->octets, tmpclient->get_client_ip().octets, NUM_IPV4_OCTETS) == 0)
							whichisclient = TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT;
						else
							whichisclient = TCAPS_FFP_FLOW_DST_IP_IS_CLIENT;
					}
					else if(tmpclient->get_client_ip().ip_version == TCAPS_SP_IP_VERSION_6)
					{
						if(memcmp(srcip->octets, tmpclient->get_client_ip().octets, NUM_IPV6_OCTETS) == 0)
							whichisclient = TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT;
						else
							whichisclient = TCAPS_FFP_FLOW_DST_IP_IS_CLIENT;
					}
					
					free(srcip->octets);
					free(srcip);
						
					//char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, u8 * flowid, u8 * serialized_features, u16 num_features
					packet.packet = tcaps_sp_ffp_create_withfeatures(macbuf,
																	 tmpclient->get_tm_ip().ip_version,
																	 tm_ip_buf,
																	 tmpflow->get_flow_id()->get_flow_id_type(),
																	 whichisclient,
																	 tmpclient->get_client_ip().ip_version,
																	 tmpflow->get_flow_id()->get_binary_flow_id(),
																	 serialized,
																	 num_features);
					tcaps_sp_ffp_tostring((struct tcaps_sp_ffp *)packet.packet, packet_temp, sizeof(packet_temp));												 
					//ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to send a state update that looks like this:\n%s", packet_temp));
					send_packet(&packet, &fc_ip, TCAPS_FC_LOCAL_PORT, fc_comms_task);
																	 
					tcaps_sp_ffp_destroy((struct tcaps_sp_ffp *)packet.packet);
					free(serialized);
					
					tmpflow->reset_state_changed();
				}
				else
					ACE_DEBUG ((LM_DEBUG, "(%P|%t) State unchanged for flow %d... skipping\n", flow_count+1));
			
			}
			
			client_flows->release_mutex();
			
			#ifdef _MUTEX_DEBUG_
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) send_state_to_fc() flow list mutex released for client %s\n", client_ip_buf));
			#endif
			
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Finished processing client %d\n\n\n", client_count+1));
			flow_count = 0;
			
		}
		
		//client_list->release_mutex();
		
		
		
		
		
//	if(client_count > 0)
//	{	
//		struct timespec sleep_timer;
//		
//		sleep_timer.tv_sec = 0;
//		sleep_timer.tv_nsec = 500000000; //.5s
//			
//			
//		ACE_OS::nanosleep(&sleep_timer);
//		
//		exit(0);
//	}
		
		
		
		//update the send timer value
		send_timer = TCAPS_FS_SEND_STATE_TO_FC_INTERVAL_SECS + (unsigned int)ACE_OS::time();
	}
}


TCAPS_Network_Task * start_comms_thread(unsigned short local_port)
{
	TCAPS_Network_Task * comms_task = NULL;
	
	ACE_INET_Addr local(local_port);
	
	ACE_SOCK_Dgram * dgram = new ACE_SOCK_Dgram();
	
	if (dgram->open (local) != -1)
	{
		comms_task = new TCAPS_Network_Task(dgram);
	}
	
	return comms_task;
}

TCAPS_Flow_Manager_Task * start_flow_manager_thread()
{
	TCAPS_Flow_Manager_Task * fm_task = NULL;
	
	fm_task = new TCAPS_Flow_Manager_Task (1);
	
	return fm_task;
}


TCAPS_Sniff_Task * start_sniff_thread(pcap_t * sniff, TCAPS_Flow_Manager_Task * fm_task)
{
	TCAPS_Sniff_Task * sniff_task;
	
	sniff_task = new TCAPS_Sniff_Task(1, sniff, fm_task);
	
	return sniff_task;
}





void send_packet(struct tcaps_sp_packet * packet, struct ip_address * recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task)
{
	char iptmp[60];
	tcaps_ip_ntoa(recipient_ip, iptmp, TRUE);	    
    ACE_INET_Addr remote(recipient_port, iptmp);
		
	/* Create a block large enough for our Data object as well as a
    text message.  */
  	Packet_Block *message;
  	u8 * serialized;
  	unsigned int serialized_bytes;
  	
  	switch(packet->packet_type)
  	{
//  		case TCAPS_RMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_rmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_rmp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_PP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_pp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_pp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_TP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_tp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_tp *)packet->packet)->serialized;
//  			break;
  		case TCAPS_AP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_ap *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_ap *)packet->packet)->serialized;
  			break;
  		case TCAPS_FFP_PACKET_TYPE:
  			serialized_bytes = ((struct tcaps_sp_ffp *)packet->packet)->serialized_bytes;
  			serialized = ((struct tcaps_sp_ffp *)packet->packet)->serialized;
  			break;
  	}

  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
  	
  	
  	message->copy ((const char *) &remote, sizeof(remote));
  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
  	message->copy ((const char *) serialized, serialized_bytes);

	ACE_DEBUG ((LM_DEBUG, "(%P|%t) adding a packet to the queue\n"));
	comms_task->send_packet(message);
	
      //if(tx_task->putq(message) == -1)
      	//something screwed up
}





struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote)
{
	ACE_Message_Block *message;
	struct tcaps_sp_packet * packet = NULL;
	
	message = comms_task->receive_packet();
	
	if(message != NULL)
	{
		unsigned int num_serialized_bytes;
		
		ACE_OS::memmove ((char *) remote, message->rd_ptr(), sizeof (ACE_INET_Addr));
		message->rd_ptr (sizeof (ACE_INET_Addr));  // Move the rd_ptr() beyond the data.
		
		ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
		message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
		
		u8 serialized_bytes[num_serialized_bytes];
		
		ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
		message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
		
		message->release();
		
		packet = (struct tcaps_sp_packet *)malloc(sizeof(struct tcaps_sp_packet));
		
		tcaps_sp_parse_packet((u8 *)serialized_bytes, num_serialized_bytes, packet);
	}
	
	return packet;
}


void check_for_packets(TCAPS_Network_Task * comms_task)
{
	ACE_INET_Addr remote;
	
	struct tcaps_sp_packet * packet = receive_packet(comms_task, &remote);
        
	while(packet != NULL)
	{
		switch(packet->packet_type)
		{
			case TCAPS_AP_PACKET_TYPE:
				handle_ap_packet((struct tcaps_sp_ap * )packet->packet, &remote);
				break;
	    	case TCAPS_CMP_PACKET_TYPE:
				handle_cmp_packet((struct tcaps_sp_cmp * )packet->packet, &remote);
				break;
			default:
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected tcaps_sp packet type\n"));
				break;
		}

		free(packet);
		
		packet = receive_packet(comms_task, &remote);
	}
}

void handle_ap_packet(struct tcaps_sp_ap * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_ap_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received an acknowledgment packet from %s that looks like this\n%s", remote->get_host_addr(), tmp));
}

void handle_cmp_packet(struct tcaps_sp_cmp * packet, ACE_INET_Addr * remote)
{
	char tmp[BUFFER_SIZE];
	tcaps_sp_cmp_tostring(packet, tmp, sizeof(tmp));
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just received a client management packet from %s that looks like this\n%s", remote->get_host_addr(), tmp));

	
	
	struct ip_address temp;
	temp.ip_version = packet->client_ip.ip_version;
	temp.octets = tcaps_sp_cmp_field_ptr(packet, TCAPS_SP_CMP_POINT_TO_START_OF_CLIENT_IP_OCTETS);
	
	TCAPS_Client * client;
	
	if(fm_task->get_client_list()->find_client(&temp))
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Found the client in the client list already!\n"));
		client = fm_task->get_client_list()->get_client(&temp);
		
		if(packet->op_type == TCAPS_SP_CMP_OP_TYPE_REMOVE)
		{
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) We have been instructed to delete this client\n"));
			
			fm_task->get_client_list()->remove_client(&temp);
		}
	}
	else
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Did NOT find the client in the client list!\n"));
		
		if(packet->op_type == TCAPS_SP_CMP_OP_TYPE_MONITOR)
		{
			struct ip_address tm_ip, fc_ip;
			tm_ip.octets = fc_ip.octets = NULL;
			
			if(packet->tm_ip_present == TRUE)
			{
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) We were given a TM IP address in the CMP\n"));
				
				tm_ip.ip_version = packet->tm_ip.ip_version;
				tm_ip.octets = tcaps_sp_cmp_field_ptr(packet, TCAPS_SP_CMP_POINT_TO_START_OF_TM_IP_OCTETS);
			}
			else
			{
				//we haven't been explicitly given a TM address for this client, so we can use the SRC_IP of the packet
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) We were NOT given a TM IP address in the CMP\n"));
				
				char ipbuf[60];
					
				if(remote->get_type() == PF_INET)
				{
					tm_ip.ip_version = TCAPS_SP_IP_VERSION_4;
					tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				}
				else if(remote->get_type() == PF_INET6)
				{
					tm_ip.ip_version = TCAPS_SP_IP_VERSION_6;
					tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				}
				
				tcaps_ip_aton(remote->get_host_addr((char *)&ipbuf, sizeof(ipbuf)), &tm_ip);
			}
			
			if(packet->fc_ip_present == TRUE)
			{
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) We were given a FC IP address in the CMP\n"));
				
				//pull the FC address from the packet and assign to the client
				
				fc_ip.ip_version = packet->fc_ip.ip_version;
				fc_ip.octets = tcaps_sp_cmp_field_ptr(packet, TCAPS_SP_CMP_POINT_TO_START_OF_FC_IP_OCTETS);
			}
			else
			{
				//we haven't been explicitly given an FC address for this client, so we can pick from our own list
				ACE_DEBUG ((LM_DEBUG, "(%P|%t) We were NOT given a FC IP address in the CMP\n"));

				fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				
				//hard coded at the moment... but eventually could assign an FC randomly from an available pool
				fc_ip.ip_version = TCAPS_SP_IP_VERSION_4;
				tcaps_ip_aton(TCAPS_FC_IP, &fc_ip);
			}
			
			client = new TCAPS_Client(packet->client_ip.ip_version,
									  tcaps_sp_cmp_field_ptr(packet, TCAPS_SP_CMP_POINT_TO_START_OF_CLIENT_IP_OCTETS),
									  tm_ip.ip_version,
									  tm_ip.octets,
									  fc_ip.ip_version,
									  fc_ip.octets
									  );
			
			if(packet->tm_ip_present == FALSE)
				free(tm_ip.octets);
											  
			if(packet->fc_ip_present == FALSE)
				free(fc_ip.octets);
			
			
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to add new client to list\n"));
			fm_task->get_client_list()->add_client(client);
			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Client added\n"));
		}
	}
										
	
	

	
	tcaps_sp_cmp_destroy(packet);
}




