#include "tcaps_fs.h"


TCAPS_Client::TCAPS_Client ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client ctor 0x%x\n", (void *) this));
}

TCAPS_Client::TCAPS_Client (u8 clientip_version, u8 * clientip_octets, u8 tmip_version, u8 * tmip_octets, u8 fcip_version, u8 * fcip_octets)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client (u8, u8*, u8, u8*, u8, u8*) ctor 0x%x\n", (void *) this));
	
	if(TCAPS_SP_IP_VERSION_4 == clientip_version)
	{
		client_ip.ip_version = TCAPS_SP_IP_VERSION_4;
		client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(client_ip.octets, clientip_octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == clientip_version)
	{
		client_ip.ip_version = TCAPS_SP_IP_VERSION_6;
		client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(client_ip.octets, clientip_octets, NUM_IPV6_OCTETS);
	}
	
	if(TCAPS_SP_IP_VERSION_4 == tmip_version)
	{
		tm_ip.ip_version = TCAPS_SP_IP_VERSION_4;
		tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(tm_ip.octets, tmip_octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == tmip_version)
	{
		tm_ip.ip_version = TCAPS_SP_IP_VERSION_6;
		tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(tm_ip.octets, tmip_octets, NUM_IPV6_OCTETS);
	}
	
	if(TCAPS_SP_IP_VERSION_4 == fcip_version)
	{
		fc_ip.ip_version = TCAPS_SP_IP_VERSION_4;
		fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(fc_ip.octets, fcip_octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == fcip_version)
	{
		fc_ip.ip_version = TCAPS_SP_IP_VERSION_6;
		fc_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(fc_ip.octets, fcip_octets, NUM_IPV6_OCTETS);
	}
	
	char buf[60];
	tcaps_ip_ntoa(&client_ip, buf, TRUE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added client address of: %s\n", buf));
	tcaps_ip_ntoa(&tm_ip, buf, TRUE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added tm address of: %s\n", buf));
	tcaps_ip_ntoa(&fc_ip, buf, TRUE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added fc address of: %s\n", buf));
	
}

TCAPS_Client::TCAPS_Client (const TCAPS_Client &copied_client)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client ctor 0x%x\n", (void *) this));
	
	client_ip = copied_client.get_client_ip();
	tm_ip = copied_client.get_tm_ip();
	client_flows = copied_client.get_client_flows();
}

TCAPS_Client::~TCAPS_Client (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client dtor 0x%x\n", (void *) this));
	
	if(client_ip.octets != NULL)
		free(client_ip.octets);
	
	if(tm_ip.octets != NULL)
		free(tm_ip.octets);
	
	if(fc_ip.octets != NULL)
		free(fc_ip.octets);
}
	
TCAPS_Client & TCAPS_Client::operator=(const TCAPS_Client &copied_client)
{
	if(this == &copied_client)
		return *this;
		
	client_ip = copied_client.get_client_ip();
	tm_ip = copied_client.get_tm_ip();
	client_flows = copied_client.get_client_flows();
	
	return *this;
}

struct ip_address TCAPS_Client::get_client_ip() const
{
	return client_ip;
}

struct ip_address TCAPS_Client::get_tm_ip() const
{
	return tm_ip;
}

struct ip_address TCAPS_Client::get_fc_ip() const
{
	return fc_ip;
}

TCAPS_Flow_List TCAPS_Client::get_client_flows() const
{
	return client_flows;
}

TCAPS_Flow_List * TCAPS_Client::get_client_flows()
{
	return &client_flows;
}

void TCAPS_Client::set_fc_ip(const struct ip_address * ip)
{
	if(ip->ip_version != fc_ip.ip_version)
	{
		free(fc_ip.octets);
		
		if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
		{
			fc_ip.ip_version = TCAPS_SP_IP_VERSION_4;
			fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		}
		else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
		{
			fc_ip.ip_version = TCAPS_SP_IP_VERSION_6;
			fc_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		}
	}
	
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		memcpy(fc_ip.octets, ip->octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		memcpy(fc_ip.octets, ip->octets, NUM_IPV6_OCTETS);
	}
}

//void TCAPS_Client::update_client_flow(TCAPS_Flow * flow)
//{
//	if(client_flows.find_flow(flow->get_flow_id()))
//	{
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Found the flow in the flow list already!\n"));
//		
//		update the existing flow's flow feature map
//		
//		TCAPS_Flow * found_flow = client_flows.get_flow(flow->get_flow_id());
//		found_flow->set_flow_features(flow->get_flow_features());
//		
//		delete flow;
//	}
//	else
//	{
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Did NOT find the flow in the flow list!\n"));
//		client_flows.add_flow(flow);
//	}
//}

void TCAPS_Client::to_string(char * buf, unsigned int bufsize)
{
	snprintf(buf, bufsize, "client_ip: ");
	
	tcaps_ip_ntoa(&client_ip, buf+strlen(buf), TRUE);
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\ttm_ip: ");
	
	tcaps_ip_ntoa(&tm_ip, buf+strlen(buf), TRUE);
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\tfc_ip: ");
	
	tcaps_ip_ntoa(&fc_ip, buf+strlen(buf), TRUE);
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\n");
}
	
bool operator<(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	if(TCAPS_SP_IP_VERSION_4 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV4_OCTETS) < 0);
	
	else if(TCAPS_SP_IP_VERSION_6 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV6_OCTETS) < 0);
	
	return false;
}

bool operator>(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	if(TCAPS_SP_IP_VERSION_4 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV4_OCTETS) > 0);
	
	else if(TCAPS_SP_IP_VERSION_6 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV6_OCTETS) > 0);
	
	return false;
}

bool operator==(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	if(TCAPS_SP_IP_VERSION_4 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV4_OCTETS) == 0);
	
	else if(TCAPS_SP_IP_VERSION_6 == f1.client_ip.ip_version)
		return (memcmp(f1.client_ip.octets, f2.client_ip.octets, NUM_IPV6_OCTETS) == 0);
	
	return false;
}
