#include "tcaps_fs.h"

Ethernet_Frame::Ethernet_Frame (u8 * frame) : Layer2_Frame(frame)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Ethernet_Frame ctor 0x%x\n", (void *) this));
	#endif
	
	layer2_protocol = LAYER2_PROTOCOL_ETHERNET;
	layer2_hdr_len = get_layer2_hdr_len();
	layer2_trailer_len = get_layer2_trailer_len();
	parse_ethernet_type();
	parse_src_mac();
	parse_dst_mac();
	
	//if(noerror)
		parse_layer3();
}

Ethernet_Frame::~Ethernet_Frame(void)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Ethernet_Frame dtor 0x%x\n", (void *) this));
	#endif
}

u16 Ethernet_Frame::get_ethernet_type() const
{
	return ethernet_type;
}

unsigned int Ethernet_Frame::get_layer2_hdr_len() const
{
	return (2*NUM_MAC_OCTETS)+sizeof(ethernet_type);
}

unsigned int Ethernet_Frame::get_layer2_trailer_len() const
{
	return 4;
}

const struct mac_address * Ethernet_Frame::get_src_mac_address() const
{
	return &mac_address_src;
}

const struct mac_address * Ethernet_Frame::get_dst_mac_address() const
{
	return &mac_address_dst;
}

void Ethernet_Frame::parse_layer3()
{
	if(ethernet_type == ETHERNET_FRAME_TYPE_IPV4)
	{
		layer3_packet = new IPv4_Packet(layer2_frame+layer2_hdr_len);
	}
	else if(ethernet_type == ETHERNET_FRAME_TYPE_IPV6)
	{
		layer3_packet = new IPv6_Packet(layer2_frame+layer2_hdr_len);
	}
}

void Ethernet_Frame::parse_ethernet_type()
{
	ethernet_type = ntohs(*((u16 *)(layer2_frame+(2*NUM_MAC_OCTETS))));
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) ethernet_type: 0x%x\n", ethernet_type));
}

void Ethernet_Frame::parse_src_mac()
{
	memcpy(mac_address_src.octets, layer2_frame, NUM_MAC_OCTETS);
}

void Ethernet_Frame::parse_dst_mac()
{
	memcpy(mac_address_dst.octets, layer2_frame+NUM_MAC_OCTETS, NUM_MAC_OCTETS);
}

void Ethernet_Frame::to_string(char * buf, unsigned int buflen) const
{
	char macbuf1[20], macbuf2[20];
	
	printHexBytes(mac_address_src.octets, sizeof(mac_address_src.octets), macbuf1, sizeof(macbuf1), TRUE);
	printHexBytes(mac_address_dst.octets, sizeof(mac_address_dst.octets), macbuf2, sizeof(macbuf2), TRUE);
	//tcaps_ether_ntoa(&mac_address_dst, macbuf2, FALSE);
	
	snprintf(buf,
			 buflen,
			 "layer2_protocol: ETHERNET\nlayer2_hdr_len: %d\nlayer2_trailer_len: %d\nsrc_mac_address: %s\ndst_mac_address: %s\n",
			 layer2_hdr_len,
			 layer2_trailer_len,
			 macbuf1,
			 macbuf2);
			 
}
