#ifndef _TCAPS_FS_FLOW_H_
#define _TCAPS_FS_FLOW_H_

#include <list>
#include <map>

#include "ace/Synch.h"

#define MAX_PCAP_PACKET_SIZE_BYTES	1600.0
#define MAX_NUM_FLOWS				10000.0
#define MAX_MEM_USAGE_MB			512.0
//									<-----------------Max bytes per flow-------------------->
#define MAX_PACKETS_PER_FLOW		((unsigned int)(((MAX_MEM_USAGE_MB/MAX_NUM_FLOWS)*1024*1024)/(MAX_PCAP_PACKET_SIZE_BYTES)))


typedef ACE_Thread_Mutex mutex_t;
typedef std::pair<tcaps_flow_feature_type,tcaps_flow_feature> feature_entry;


class TCAPS_Flow
{
public:
	TCAPS_Flow (TCAPS_Flow_ID * id);
	TCAPS_Flow (const TCAPS_Flow &);
	virtual ~TCAPS_Flow (void);
	TCAPS_Flow_ID * get_flow_id() const;
	unsigned int get_num_packets_in_flow() const;
	void add_packet(Pcap_Packet * packet, const struct ip_address * clientip);
	std::list<Pcap_Packet*> get_packet_list() const;
	
	const Pcap_Packet * get_last_rx_packet() const;
	
	void calculate_since_beginning_flow_features(Pcap_Packet * packet, bool is_client2other_packet);
	void calculate_window_flow_features();
	void calculate_histogram_flow_features();
	
	bool state_changed() const;
	void reset_state_changed();
	u8 * serialize_flow_features(unsigned int * num_serialized_features);
	void to_string(char * buf, unsigned int bufsize);
	
	TCAPS_Flow & operator=(const TCAPS_Flow & copied_flow);
	
	friend bool operator<(const TCAPS_Flow &f1, const TCAPS_Flow &f2);
	friend bool operator>(const TCAPS_Flow &f1, const TCAPS_Flow &f2);
	friend bool operator==(const TCAPS_Flow &f1, const TCAPS_Flow &f2);

private:
	bool is_client_to_other_packet(Pcap_Packet * packet, const struct ip_address * clientip);
	std::map <u8,float> TCAPS_Flow::calculate_histo_percentiles(std::map <unsigned int,unsigned int> & the_histo_map, unsigned int total_samples);

  
protected:
	TCAPS_Flow_ID * flow_id;
	std::list <Pcap_Packet*> flow_packets;
	std::map <tcaps_flow_feature_type,tcaps_flow_feature> flow_features;
	unsigned int num_packets_in_window;
	unsigned int flow_timeout;
	bool flow_state_changed;
	mutex_t flow_mutex;
	mutex_t flow_packets_mutex;
	mutex_t flow_features_mutex;
	Pcap_Packet * last_rx_packet;
	Pcap_Packet * last_rx_client2other_packet;
	Pcap_Packet * last_rx_other2client_packet;
	
	//variables used in "since flow beginning" calculations
	unsigned int flow_start_time_secs;
	unsigned int num_packets_bidir_since_beginning;
	unsigned int num_packets_client2other_since_beginning;
	unsigned int num_packets_other2client_since_beginning;
	u64 num_payload_bytes_bidir_since_beginning;
	u64 num_payload_bytes_client2other_since_beginning;
	u64 num_payload_bytes_other2client_since_beginning;
	float average_interarrival_time_bidir_ms;
	float average_interarrival_time_client2other_ms;
	float average_interarrival_time_other2client_ms;
	float min_interarrival_time_client2other_ms;
	float min_interarrival_time_other2client_ms;
	float max_interarrival_time_client2other_ms;
	float max_interarrival_time_other2client_ms;
	unsigned int min_client2other_payload_bytes;
	unsigned int min_other2client_payload_bytes;
	unsigned int max_client2other_payload_bytes;
	unsigned int max_other2client_payload_bytes;
	
	std::map <unsigned int,unsigned int> client2other_payload_bytes_histo;
	std::map <unsigned int,unsigned int> other2client_payload_bytes_histo;
	
	std::map <unsigned int,unsigned int> client2other_interarrival_time_ms_histo;
	std::map <unsigned int,unsigned int> other2client_interarrival_time_ms_histo;
  
};









#endif //_TCAPS_FS_FLOW_H_
