#include "tcaps_fs.h"

TCAPS_Flow_ID::TCAPS_Flow_ID ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID default ctor 0x%x\n", (void *) this));
}

TCAPS_Flow_ID::TCAPS_Flow_ID (const struct ip_address * src_ip, const struct ip_address * dst_ip)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID(u8 fid_type, u8 ipversion, u8 * bytes) ctor 0x%x\n", (void *) this));
	
	flow_id_type = TCAPS_FFP_FLOW_ID_TYPE_IPIP;
	ip_version = src_ip->ip_version;
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
	{
		num_bytes = 2*NUM_IPV4_OCTETS;
		flow_id = (u8 *)malloc(num_bytes);
		memcpy(flow_id, src_ip->octets, NUM_IPV4_OCTETS);
		memcpy(flow_id+NUM_IPV4_OCTETS, dst_ip->octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip_version)
	{
		num_bytes = 2*NUM_IPV6_OCTETS;
		flow_id = (u8 *)malloc(num_bytes);
		memcpy(flow_id, src_ip->octets, NUM_IPV6_OCTETS);
		memcpy(flow_id+NUM_IPV6_OCTETS, dst_ip->octets, NUM_IPV6_OCTETS);
	}
}

TCAPS_Flow_ID::TCAPS_Flow_ID (u8 fid_type, u8 ipversion, const u8 * bytes)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID ctor 0x%x\n", (void *) this));
	
	flow_id_type = fid_type;
	ip_version = ipversion;
	
	if(TCAPS_SP_IP_VERSION_4 == ipversion)
	{
		num_bytes = 2*NUM_IPV4_OCTETS;
	}
	else if(TCAPS_SP_IP_VERSION_6 == ipversion)
	{
		num_bytes = 2*NUM_IPV6_OCTETS;
	}
	
	flow_id = (u8 *)malloc(num_bytes);
		
	memcpy(flow_id, bytes, num_bytes);
}


TCAPS_Flow_ID::TCAPS_Flow_ID (const TCAPS_Flow_ID &copied_fid)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID copy ctor 0x%x\n", (void *) this));
	
	flow_id_type = copied_fid.get_flow_id_type();
	ip_version = copied_fid.get_flow_id_ip_version();
	num_bytes = copied_fid.get_num_flow_id_bytes();
	
	flow_id = (u8 *)malloc(num_bytes);
	
	memcpy(flow_id, copied_fid.get_binary_flow_id(), num_bytes);
}

TCAPS_Flow_ID::~TCAPS_Flow_ID (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID dtor 0x%x\n", (void *) this)); 

	free(flow_id);
}

	
	
bool operator<(const TCAPS_Flow_ID &f1, const TCAPS_Flow_ID &f2)
{
	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) < 0);
}

bool operator>(const TCAPS_Flow_ID &f1, const TCAPS_Flow_ID &f2)
{
	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) > 0);
}

bool operator==(const TCAPS_Flow_ID &f1, const TCAPS_Flow_ID &f2)
{
	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) == 0);
}


TCAPS_Flow_ID & TCAPS_Flow_ID::operator=(const TCAPS_Flow_ID &copied_fid)
{
	if(this == &copied_fid)
		return *this;
	
	if(flow_id != NULL)
		free(flow_id);
	
	flow_id_type = copied_fid.get_flow_id_type();
	ip_version = copied_fid.get_flow_id_ip_version();
	num_bytes = copied_fid.get_num_flow_id_bytes();
	
	flow_id = (u8 *)malloc(num_bytes);
	
	memcpy(flow_id, copied_fid.get_binary_flow_id(), num_bytes);
	
	return *this;
}

TCAPS_Flow_ID * TCAPS_Flow_ID::operator=(const TCAPS_Flow_ID *copied_fid)
{
	if(this == copied_fid)
		return this;
	
	if(flow_id != NULL)
		free(flow_id);
	
	flow_id_type = copied_fid->get_flow_id_type();
	ip_version = copied_fid->get_flow_id_ip_version();
	num_bytes = copied_fid->get_num_flow_id_bytes();
	
	flow_id = (u8 *)malloc(num_bytes);
	
	memcpy(flow_id, copied_fid->get_binary_flow_id(), num_bytes);
	
	return this;
}

unsigned int TCAPS_Flow_ID::get_num_flow_id_bytes() const
{
	return num_bytes;
}

u8 TCAPS_Flow_ID::get_flow_id_type() const
{
	return flow_id_type;
}

u8 TCAPS_Flow_ID::get_flow_id_ip_version() const
{
	return ip_version;
}

const u8 * TCAPS_Flow_ID::get_binary_flow_id() const
{
	return flow_id;
}

void TCAPS_Flow_ID::get_ascii_flow_id(char * buf) const
{
	ip_address temp;
	
	sprintf(buf, "src_ip: ");
	
	temp.octets = flow_id;
	temp.ip_version = ip_version;
	
	tcaps_ip_ntoa(&temp, buf+strlen(buf), TRUE);
	
	sprintf(buf+strlen(buf), "\tdst_ip: ");
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
		temp.octets = flow_id+NUM_IPV4_OCTETS;
	else
		temp.octets = flow_id+NUM_IPV6_OCTETS;
		
	tcaps_ip_ntoa(&temp, buf+strlen(buf), TRUE);
}

void TCAPS_Flow_ID::get_flow_id_srcip(char * buf)
{
	struct ip_address temp;
	temp.ip_version = ip_version;
	temp.octets = flow_id;
	tcaps_ip_ntoa(&temp, buf, TRUE);
}

void TCAPS_Flow_ID::get_flow_id_dstip(char * buf)
{
	struct ip_address temp;
	temp.ip_version = ip_version;
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
		temp.octets = flow_id+NUM_IPV4_OCTETS;
	else
		temp.octets = flow_id+NUM_IPV6_OCTETS;
	
	tcaps_ip_ntoa(&temp, buf, TRUE);
}

struct ip_address * TCAPS_Flow_ID::get_flow_id_srcip()
{
	struct ip_address * temp = (struct ip_address *)malloc(sizeof(struct ip_address));
	
	temp->ip_version = ip_version;
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
	{
		temp->octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(temp->octets, flow_id, NUM_IPV4_OCTETS);
	}
	else
	{
		temp->octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(temp->octets, flow_id, NUM_IPV6_OCTETS);
	}
	
	return temp;
	
}

struct ip_address * TCAPS_Flow_ID::get_flow_id_dstip()
{
	struct ip_address * temp = (struct ip_address *)malloc(sizeof(struct ip_address));
	
	temp->ip_version = ip_version;
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
	{
		temp->octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(temp->octets, flow_id+NUM_IPV4_OCTETS, NUM_IPV4_OCTETS);
		//temp->octets = flow_id+NUM_IPV4_OCTETS;
	}
	else
	{
		temp->octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(temp->octets, flow_id+NUM_IPV6_OCTETS, NUM_IPV6_OCTETS);
		//temp->octets = flow_id+NUM_IPV6_OCTETS;
	}
	
	return temp;
}
	

//
//
//
//
//
//
//
//
//
//
//#include "tcaps_fs.h"
//
//Flow_ID::Flow_ID ()
//{
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow_ID ctor 0x%x\n", (void *) this));
//}
//
//Flow_ID::Flow_ID (u8 * bytes, unsigned int num_b)
//{
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow_ID ctor 0x%x\n", (void *) this));
//	num_bytes = num_b;
//	
//	flow_id = (u8 *)malloc(num_bytes);
//	
//	memcpy(flow_id, bytes, num_bytes);
//}
//
//Flow_ID::Flow_ID (u8 * src_ip, u8 * dst_ip, u8 * dst_port, unsigned int ip_size_bytes, unsigned int port_size_bytes)
//{
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow_ID ctor 0x%x\n", (void *) this));
//	
//	num_bytes = (ip_size_bytes*2)+port_size_bytes;
//	
//	flow_id = (u8 *)malloc(num_bytes);
//	
//	memcpy(flow_id, src_ip, ip_size_bytes);
//	memcpy(flow_id+ip_size_bytes, dst_ip, ip_size_bytes);
//	memcpy(flow_id+(2*ip_size_bytes), dst_port, port_size_bytes);
//}
//
//Flow_ID::Flow_ID (const Flow_ID & copied_flow)
//{
//	num_bytes = copied_flow.get_num_flow_id_bytes();
//	
//	flow_id = (u8 *)malloc(num_bytes);
//	
//	memcpy(flow_id, copied_flow.get_binary_flow_id(), num_bytes);
//	
//}
//
//Flow_ID::~Flow_ID (void)
//{
//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Flow_ID dtor 0x%x\n", (void *) this)); 
//
//	free(flow_id);
//}
//
//bool Flow_ID::compare(const Flow_ID &f1, const Flow_ID &f2) const
//{
//	return f1 < f2;
//}
//
//	
//	
//bool operator<(const Flow_ID &f1, const Flow_ID & f2)
//{
//	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) < 0);
//}
//
//bool operator>(const Flow_ID &f1, const Flow_ID & f2)
//{
//	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) > 0);
//}
//
//bool operator==(const Flow_ID &f1, const Flow_ID & f2)
//{
//	return (memcmp((const void *)f1.flow_id, (const void *)f2.flow_id, f1.num_bytes) == 0);
//}
//
//
//Flow_ID & Flow_ID::operator=(const Flow_ID & copied_flow)
//{
//	if(this == &copied_flow)
//		return *this;
//		
//	free(flow_id);
//	
//	memcpy(flow_id, copied_flow.get_binary_flow_id(), copied_flow.get_num_flow_id_bytes());
//	num_bytes = copied_flow.get_num_flow_id_bytes();
//	
//	return *this;
//}
//
//unsigned int Flow_ID::get_num_flow_id_bytes() const
//{
//	return num_bytes;
//}
//
//
//const u8 * Flow_ID::get_binary_flow_id() const
//{
//	return flow_id;
//}
//
//void Flow_ID::get_ascii_flow_id(char * buf) const
//{
//	ip_address temp;
//	
//	temp.octets = flow_id;
//	temp.ip_version = TCAPS_SP_IP_VERSION_4;
//	
//	tcaps_ip_ntoa(&temp, buf, TRUE);
//	
//	sprintf(buf+strlen(buf), "->");
//	
//	temp.octets = flow_id+(TCAPS_SP_IPV4HDR_SRCIP_BITS/8);
//	tcaps_ip_ntoa(&temp, buf+strlen(buf), TRUE);
//	
//	sprintf(buf+strlen(buf), ":");
//	
//	sprintf(buf+strlen(buf), "%d", ntohs(*((u16*)(flow_id+((TCAPS_SP_IPV4HDR_SRCIP_BITS/8)+(TCAPS_SP_IPV4HDR_DSTIP_BITS/8))))));
//}
