#include "tcaps_fs.h"

TCAPS_Flow_ID_Port::TCAPS_Flow_ID_Port ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID_Port default ctor 0x%x\n", (void *) this));
}

TCAPS_Flow_ID_Port::TCAPS_Flow_ID_Port (const struct ip_address * src_ip, u16 src_port, const struct ip_address * dst_ip, u16 dst_port)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID_Port(src_ip, src_port, dst_ip, dst_port) ctor 0x%x\n", (void *) this));
	
	flow_id_type = TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT;
	ip_version = src_ip->ip_version;
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
	{
		num_bytes = (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
		flow_id = (u8 *)malloc(num_bytes);
		memcpy(flow_id, src_ip->octets, NUM_IPV4_OCTETS);
		memcpy(flow_id+NUM_IPV4_OCTETS, dst_ip->octets, NUM_IPV4_OCTETS);
		memcpy(flow_id+(2*NUM_IPV4_OCTETS), &src_port, sizeof(u16));
		memcpy(flow_id+(2*NUM_IPV4_OCTETS)+sizeof(u16), &dst_port, sizeof(u16));
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip_version)
	{
		num_bytes = (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
		flow_id = (u8 *)malloc(num_bytes);
		memcpy(flow_id, src_ip->octets, NUM_IPV6_OCTETS);
		memcpy(flow_id+NUM_IPV6_OCTETS, dst_ip->octets, NUM_IPV6_OCTETS);
		memcpy(flow_id+(2*NUM_IPV6_OCTETS), &src_port, sizeof(u16));
		memcpy(flow_id+(2*NUM_IPV6_OCTETS)+sizeof(u16), &dst_port, sizeof(u16));
	}
}

TCAPS_Flow_ID_Port::TCAPS_Flow_ID_Port (u8 fid_type, u8 ipversion, u8 * bytes)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID_Port ctor 0x%x\n", (void *) this));
	
	flow_id_type = fid_type;
	ip_version = ipversion;
	
	if(TCAPS_SP_IP_VERSION_4 == ipversion)
	{
		num_bytes = (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
	}
	else if(TCAPS_SP_IP_VERSION_6 == ipversion)
	{
		num_bytes = (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
	}
	
	flow_id = (u8 *)malloc(num_bytes);
		
	memcpy(flow_id, bytes, num_bytes);
}


TCAPS_Flow_ID_Port::TCAPS_Flow_ID_Port (const TCAPS_Flow_ID_Port &copied_fid)
{
	TCAPS_Flow_ID::TCAPS_Flow_ID((TCAPS_Flow_ID)copied_fid);

	src_port = copied_fid.get_flow_id_src_port();
	dst_port = copied_fid.get_flow_id_dst_port();
}

TCAPS_Flow_ID_Port::~TCAPS_Flow_ID_Port (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_ID_Port dtor 0x%x\n", (void *) this)); 

}


TCAPS_Flow_ID_Port & TCAPS_Flow_ID_Port::operator=(const TCAPS_Flow_ID_Port & copied_fid)
{
	TCAPS_Flow_ID::operator=((TCAPS_Flow_ID)copied_fid);

	src_port = copied_fid.get_flow_id_src_port();
	dst_port = copied_fid.get_flow_id_dst_port();
	
	return *this;
}

u16 TCAPS_Flow_ID_Port::get_flow_id_src_port() const
{
	return src_port;
}

u16 TCAPS_Flow_ID_Port::get_flow_id_dst_port() const
{
	return dst_port;
}
	

void TCAPS_Flow_ID_Port::get_ascii_flow_id(char * buf) const
{
	ip_address temp;
	
	temp.octets = flow_id;
	temp.ip_version = ip_version;
	
	tcaps_ip_ntoa(&temp, buf, TRUE);
	
	sprintf(buf+strlen(buf), ":%d\t", ntohs(*((u16*)(flow_id+(num_bytes-sizeof(u16)-sizeof(u16))))));
	
	if(TCAPS_SP_IP_VERSION_4 == ip_version)
		temp.octets = flow_id+NUM_IPV4_OCTETS;
	else
		temp.octets = flow_id+NUM_IPV6_OCTETS;
		
	tcaps_ip_ntoa(&temp, buf+strlen(buf), TRUE);

	sprintf(buf+strlen(buf), ":%d", ntohs(*((u16*)(flow_id+(num_bytes-sizeof(u16))))));
}
