#include "tcaps_fs.h"



/* Set our housekeeping pointer to NULL and tell the user we exist.  */
TCAPS_Flow_Manager_Task::TCAPS_Flow_Manager_Task (size_t n_threads)
 : n_threads_ (n_threads)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task ctor 0x%x\n", (void *) this));
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Flow_Manager_Task::~TCAPS_Flow_Manager_Task (void)
{
  ACE_DEBUG ((LM_DEBUG,
              "(%P|%t) TCAPS_Flow_Manager_Task dtor 0x%x\n",
              (void *) this));

  /* Get our shutdown notification out of the queue and release it.  */
  ACE_Message_Block *message;

  /* Like the getq() in svc() below, this will block until a message
    arrives.  By blocking, we know that the destruction will be paused
    until the last thread is done with the message block.  */
  this->getq (message);
  message->release ();
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int
TCAPS_Flow_Manager_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP,
                         n_threads_);
}

/* Tell the user we're closing and invoke the baseclass' close() to
  take care of things.  */
int
TCAPS_Flow_Manager_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG,
              "(%P|%t) TCAPS_Flow_Manager_Task close 0x%x\n",
              (void *) this));
  return inherited::close (flags);
}

/* Our svc() method waits for work on the queue and then processes
  that work.  */
int
TCAPS_Flow_Manager_Task::svc (void)
{
	ACE_Message_Block *message;
	boolean finished = FALSE;
	char pktbuf[BUFFER_SIZE];
	char ipbuf[64];
	
	//this may be foobed as we are using relative time, and API says to use absolute time...
	ACE_Time_Value wait_for_message(0, 1); //wait 1us
	
	unsigned int time_since_last_flow_kull = (unsigned int)ACE_OS::time();
	
	struct timespec sleep_timer;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 1000; //1us
	//sleep_timer.tv_nsec = 500000000; //500ms
	
	
	
	
	while(!finished)
	{
		if(this->getq(message, &wait_for_message) == -1)
		{
    		//Let's do some processing
    		
    		if(!unsifted_packets.empty())
    		{
	    		std::list<Pcap_Packet*>::iterator i;
	    		char buf[40];
	    		std::map<const TCAPS_Flow_ID,TCAPS_Flow *,flowcmp>::iterator flow_iter;
	    		TCAPS_Flow_ID *temp_fid;
	    		
	    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to sift %d packets\n", unsifted_packets.size()));
	
	   			for(i=unsifted_packets.begin(); i != unsifted_packets.end(); i++)
	   			{
	   				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to sift packet of length: %d bytes\n", (*i)->get_pcap_packet_header()->len));
	   				
	   				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Getting temp_fid\n"));
	   				temp_fid = (*i)->get_flow_id();
	   				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Got temp_fid\n"));
	   				
	   				if(temp_fid != NULL)
	   				{
	   					//the packet is a type able to be represented using a TCAPS_Flow_ID object
	   					
	   					//let's see if we have a client object the matches the src or dst IP of this packet
	   					struct ip_address * tempip;
	   					
	   					tempip = temp_fid->get_flow_id_srcip();
	   					
	   					client_list.acquire_mutex();
	   					
	   					TCAPS_Client * tmp_client = client_list.get_client(tempip);
	   					free(tempip->octets);
	   					free(tempip);
	   					
	   					if(tmp_client == NULL)
	   					{
	   						tempip = temp_fid->get_flow_id_dstip();
	   						tmp_client = client_list.get_client(tempip);
	   						free(tempip->octets);
	   						free(tempip);
	   					}
	   					
	   					if(tmp_client != NULL)
	   					{
	   						//the client this packet belongs to is in the list, so we should process this packet
	   						
	   						struct ip_address tmpclientip = tmp_client->get_client_ip();
	   						tcaps_ip_ntoa(&tmpclientip, ipbuf, TRUE);
	   						
	   						ACE_DEBUG ((LM_DEBUG, "(%P|%t) The client this packet belongs to is in the list\n"));
	   						
	   						TCAPS_Flow_List * f_list = tmp_client->get_client_flows();
	   						
	   						#ifdef _MUTEX_DEBUG_
	   						ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task::svc() acquiring flow list mutex for client %s\n", ipbuf));
	   						#endif
	   						
	   						f_list->acquire_mutex();
	   						
	   						#ifdef _MUTEX_DEBUG_
	   						ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task::svc() flow list mutex acquired for client %s\n", ipbuf));
	   						#endif

	   						TCAPS_Flow * tmp_flow = f_list->get_flow(temp_fid);
	   						
	   						if(tmp_flow == NULL)
	   						{
	   							//Flow does not exist, let's make a new one
	   							ACE_DEBUG ((LM_DEBUG, "(%P|%t) The flow this packet belongs to is NOT in the client's flow list\n"));
	   							tmp_flow = new TCAPS_Flow(temp_fid);
	   							f_list->add_flow(tmp_flow);
	   							
	   							char buf[1024];
	   							tmp_flow->to_string(buf, sizeof(buf));
	   							printf("Just found a new flow for client with IP %s:\n%s\n\n", ipbuf, buf);
	   						}
	   						else
	   						{
	   							delete temp_fid;
	   							ACE_DEBUG ((LM_DEBUG, "(%P|%t) The flow this packet belongs to IS in the client's flow list\n"));
	   						}
							
							#ifdef _MUTEX_DEBUG_
							ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task::svc() flow list mutex released for client %s\n", ipbuf));
							#endif
							
	   						tmp_flow->add_packet(*i, &tmpclientip);
	   						ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just finished adding the packet to the flow in TCAPS_Flow_Manager_Task::svc()\n"));
	   						
	   						f_list->release_mutex();
	   					}
	   					else
	   					{
	   						//we don't need to worry about this packet, free all resources
	   						//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Packet is useless... deleting resources\n"));
	   						delete temp_fid;
	   						delete (*i);
	   						//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Resources deleted\n"));
	   					}
	   					
	   					client_list.release_mutex();
	   				}
	   				
	   				
	   				
	   				
	   				
	   				
	   				
	   				
//	   				 flow_iter = flows.find(*temp);
//	   				 
//	   				 if (flow_iter != flows.end())
//	   				 {
//	   				 	//key exists, this is an existing flow
//	   				 	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Before insertion, key maps to %s\n", ));
//	   				 	flows[*temp]->add_packet(*i);
//	   				 }
//	   				 
//	   				 else
//	   				 {
//	   				 	//Key doesn't exist, this is a new flow
//	   				 	temp->get_ascii_flow_id(buf);
//	   				 	
////	   				 	if(num_flows < MAX_NUM_FLOWS)
////	   				 	{
//							TCAPS_Flow * newflow = new TCAPS_Flow(temp);
//							ACE_DEBUG ((LM_DEBUG, "(%P|%t) New flow id: %s\n", buf));
//							std::pair<const TCAPS_Flow_ID,TCAPS_Flow*> entry(*temp, newflow);
//							flows.insert(entry);
//							newflow->add_packet(*i);
//							//ACE_DEBUG ((LM_DEBUG, "(%P|%t) After insertion, key maps to %s\n", flows[*temp]));
//							//increase the flow counter
////	   				 	}
//	   				 }
	   				
	   				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Packet processed. Deleting unsifted_packets list entry\n"));
					unsifted_packets.erase(i);
					//ACE_DEBUG ((LM_DEBUG, "(%P|%t) unsifted_packets list entry deleted\n"));
	   			}
    		}
    		
    		if(((unsigned int)ACE_OS::time() - time_since_last_flow_kull) >= TCAPS_FLOW_TIMEOUT)
    		{
    			//we should check if there are any idle flows and delete them if there are
    			ACE_DEBUG ((LM_DEBUG, "(%P|%t) Performing housekeeping on client_list...\n"));
    			client_list.accept(&TCAPS_Flow_Manager_Task::client_list_housekeeping, NULL);
    			time_since_last_flow_kull = (unsigned int)ACE_OS::time();
    			
    		}
    			
    		

//    		std::pair<char*,char*> entry("192.168.0.1:5000", "IP is: 192.168.0.1, port is: 5000");
//    		
//    		flows.insert(entry);
//    		
//    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Key \"192.168.0.1:5000\" maps to %s\n", flows["192.168.0.1:5000"]));
    	}
    	
    	else
    	{
    		int ret = 0;
    		
    		while(!msg_queue_->is_empty() || ret != -1)
    		{
	    		if (message->msg_type() == ACE_Message_Block::MB_HANGUP)
		    	{
		    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to shutdown the flow_manager thread\n"));
		      		this->putq (message);
		      		finished = TRUE;
		      		break; //break out of the while loop - we are done
		    	}
		    	else
		    	{
		    		//we have received a packet from the sniff thread
		    		
		    		struct pcap_pkthdr * header = (struct pcap_pkthdr *)malloc(sizeof(struct pcap_pkthdr));
		    		
		    		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to receive a packet from the sniff thread!\n"));
			
					ACE_OS::memmove ((char *) header, message->rd_ptr(), sizeof (struct pcap_pkthdr));
					message->rd_ptr (sizeof (struct pcap_pkthdr));  // Move the rd_ptr() beyond the data.
					
					//ACE_DEBUG ((LM_DEBUG, "(%P|%t) We just received a packet of length: %d bytes\n", header->len));
					
					u_char * packet = (u_char *)malloc(header->len);
					
					ACE_OS::memmove ((char *) packet, message->rd_ptr(), header->len);
					message->rd_ptr (header->len);  // Move the rd_ptr() beyond the data.
					

					//if(packet.srcip == valid_clients || packet.dstip == valid_clients)
					//{
						Pcap_Packet * temp = new Pcap_Packet(header, packet);
						
						//temp->to_string(pktbuf, sizeof(pktbuf));
						//ACE_DEBUG ((LM_DEBUG, "(%P|%t) %s\n", pktbuf));
					
						unsifted_packets.push_back(temp);
					//}
					
					message->release();
					ret = this->getq(message, &wait_for_message);
		    		
		    	}
	    		
    		}
    	}
			
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for sleep_timer.tv_nsec nanosecs to be nice to other processes running on the system
	}

	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Exiting svc() in flow_manager thread\n"));
	
	return 0;
}

//std::map<const TCAPS_Flow_ID,TCAPS_Flow *,flowcmp> * TCAPS_Flow_Manager_Task::get_flows_map()
//{
//	return &flows;
//}

TCAPS_Client_List * TCAPS_Flow_Manager_Task::get_client_list()
{
	return &client_list;
}



void TCAPS_Flow_Manager_Task::client_list_housekeeping(TCAPS_Client * theclient, void * args)
{
	theclient->get_client_flows()->accept(&TCAPS_Flow_Manager_Task::flow_list_housekeeping, (void *)theclient->get_client_flows());
}

void TCAPS_Flow_Manager_Task::flow_list_housekeeping(TCAPS_Flow * theflow, void * args)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task::flow_list_housekeeping: time since last packet arrived for flow: %d\n", ((unsigned int)ACE_OS::time() - (theflow->get_last_rx_packet()->get_pcap_packet_header()->ts.tv_sec))));
	
	if(((unsigned int)ACE_OS::time() - (theflow->get_last_rx_packet()->get_pcap_packet_header()->ts.tv_sec)) >= TCAPS_FLOW_TIMEOUT)
	{
		//let's delete this flow
		TCAPS_Flow_List * theflowlist = (TCAPS_Flow_List *)args;
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Flow_Manager_Task::flow_list_housekeeping: Flow has expired... deleting it\n"));
		theflowlist->remove_flow(theflow->get_flow_id());
	}
}
