#include "tcaps_fs.h"

IPv4_Packet::IPv4_Packet (u8 * packet) : Layer3_Packet(packet)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) IPv4_Packet ctor 0x%x\n", (void *) this));
	#endif
	
	layer3_protocol = LAYER3_PROTOCOL_IPV4;
	layer3_hdr_len = get_layer3_hdr_len();
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) layer3_hdr_len: %d\n", layer3_hdr_len));
	parse_ipv4_protocol_field();
	parse_src_ip();
	parse_dst_ip();
	
	//if(noerror)
		parse_layer4();
}

IPv4_Packet::~IPv4_Packet(void)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) IPv4_Packet dtor 0x%x\n", (void *) this));
	#endif
}

unsigned int IPv4_Packet::get_layer3_hdr_len() const
{
	//IP hdr len (in 32 bit words) is the lower 4 bits of the first byte of the IP hdr. Multiply by 4 to get bytes.
	return ((*layer3_packet)&0x0f)*4;
}

u8 IPv4_Packet::get_protocol() const
{
	return protocol;
}

const struct ip_address * IPv4_Packet::get_src_ip() const
{
	return &ipv4_address_src;
}

const struct ip_address * IPv4_Packet::get_dst_ip() const
{
	return &ipv4_address_dst;
}
	
	
void IPv4_Packet::parse_layer4()
{
	if(protocol == IPV4_PACKET_PROTOCOL_TCP)
	{
		layer4_dgram = new TCP_Datagram(layer3_packet+layer3_hdr_len);
	}
	else if(protocol == IPV4_PACKET_PROTOCOL_UDP)
	{
		layer4_dgram = new UDP_Datagram(layer3_packet+layer3_hdr_len);
	}
}

void IPv4_Packet::parse_ipv4_protocol_field()
{
	protocol = *(layer3_packet+(TCAPS_SP_IPV4HDR_PROTOCOL_OFFSET/8));
}

void IPv4_Packet::parse_src_ip()
{
	ipv4_address_src.ip_version = TCAPS_SP_IP_VERSION_4;
	ipv4_address_src.octets = (layer3_packet+(TCAPS_SP_IPV4HDR_SRCIP_OFFSET/8));
}

void IPv4_Packet::parse_dst_ip()
{
	ipv4_address_dst.ip_version = TCAPS_SP_IP_VERSION_4;
	ipv4_address_dst.octets = (layer3_packet+(TCAPS_SP_IPV4HDR_DSTIP_OFFSET/8));
}

void IPv4_Packet::to_string(char * buf, unsigned int buflen) const
{
	char ipbuf1[20], ipbuf2[20];
	
	tcaps_ip_ntoa(&ipv4_address_src, ipbuf1, TRUE);
	tcaps_ip_ntoa(&ipv4_address_dst, ipbuf2, TRUE);
	
	snprintf(buf,
			 buflen,
			 "layer3_protocol: IPv4\nlayer3_hdr_len: %d\nsrc_ip: %s\ndst_ip: %s\nprotocol: %d\n",
			 layer3_hdr_len,
			 ipbuf1,
			 ipbuf2,
			 protocol);
			 
}
