#include "tcaps_fs.h"

IPv6_Packet::IPv6_Packet (u8 * packet) : Layer3_Packet(packet)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) IPv6_Packet ctor 0x%x\n", (void *) this));
	#endif
	
	layer3_protocol = LAYER3_PROTOCOL_IPV6;
	layer3_hdr_len = get_layer3_hdr_len();
	parse_ipv6_nexthdr_field();
	parse_src_ip();
	parse_dst_ip();
	
	//if(noerror)
		parse_layer4();
}

IPv6_Packet::~IPv6_Packet(void)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) IPv6_Packet dtor 0x%x\n", (void *) this));
	#endif
}

u8 IPv6_Packet::get_next_hdr() const
{
	return nexthdr;
}

unsigned int IPv6_Packet::get_layer3_hdr_len() const
{
	return 40;
}

const struct ip_address * IPv6_Packet::get_src_ip() const
{
	return &ipv6_address_src;
}

const struct ip_address * IPv6_Packet::get_dst_ip() const
{
	return &ipv6_address_dst;
}
	
	
void IPv6_Packet::parse_layer4()
{
	if(nexthdr == IPV6_PACKET_NEXTHDR_TCP)
	{
		layer4_dgram = new TCP_Datagram(layer3_packet+layer3_hdr_len);
	}
	else if(nexthdr == IPV6_PACKET_NEXTHDR_UDP)
	{
		layer4_dgram = new UDP_Datagram(layer3_packet+layer3_hdr_len);
	}
}

void IPv6_Packet::parse_ipv6_nexthdr_field()
{
	nexthdr = *(layer3_packet+(TCAPS_SP_IPV6HDR_NEXTHDR_OFFSET/8));
}

void IPv6_Packet::parse_src_ip()
{
	ipv6_address_src.ip_version = TCAPS_SP_IP_VERSION_6;
	ipv6_address_src.octets = (layer3_packet+(TCAPS_SP_IPV6HDR_SRCIP_OFFSET/8));
}

void IPv6_Packet::parse_dst_ip()
{
	ipv6_address_dst.ip_version = TCAPS_SP_IP_VERSION_6;
	ipv6_address_dst.octets = (layer3_packet+(TCAPS_SP_IPV6HDR_DSTIP_OFFSET/8));
}

void IPv6_Packet::to_string(char * buf, unsigned int buflen) const
{
	char ipbuf1[50], ipbuf2[50];
	
	tcaps_ip_ntoa(&ipv6_address_src, ipbuf1, TRUE);
	tcaps_ip_ntoa(&ipv6_address_dst, ipbuf2, TRUE);
	
	snprintf(buf,
			 buflen,
			 "layer3_protocol: IPv6\nlayer3_hdr_len: %d\nsrc_ip: %s\ndst_ip: %s\n",
			 layer3_hdr_len,
			 ipbuf1,
			 ipbuf2);
			 
}
