#include "tcaps_fs.h"


Pcap_Packet::Pcap_Packet (struct pcap_pkthdr * h, u_char * p)
 : header (h), packet_size (h->caplen), packet (p)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet ctor 0x%x\n", (void *) this));
	#endif

	layer2_frame = NULL;
	parse_packet();
}

Pcap_Packet::Pcap_Packet (const Pcap_Packet & copied_packet)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet copy ctor 0x%x\n", (void *) this));
	#endif
	
	layer2_frame = NULL;
	
	packet_size = copied_packet.get_packet_size();
	header = (struct pcap_pkthdr *)malloc(sizeof(struct pcap_pkthdr));
	memcpy(header, copied_packet.get_pcap_packet_header(), sizeof(struct pcap_pkthdr));
	
	packet = (u8 *)malloc(header->caplen);
	memcpy(packet, copied_packet.get_pcap_packet_bytes(), header->caplen);
		
	parse_packet();
}

Pcap_Packet::~Pcap_Packet (void)
{
	#ifdef _PACKET_PARSE_DEBUG_
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet dtor 0x%x\n", (void *) this)); 
	#endif
	
	if(layer2_frame != NULL)
		delete layer2_frame;
		
	free(header);
	free(packet);
}

unsigned int Pcap_Packet::get_packet_size() const
{
	return packet_size;
}

struct pcap_pkthdr * Pcap_Packet::get_pcap_packet_header() const
{
	return header;
}

u8 * Pcap_Packet::get_pcap_packet_bytes() const
{
	return packet;
}

TCAPS_Flow_ID * Pcap_Packet::get_flow_id()
{
	TCAPS_Flow_ID * flow_id = NULL;
	
	if(layer2_frame->get_layer3_packet() != NULL)
	{
		if(layer2_frame->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV4
			|| layer2_frame->get_layer3_packet()->get_layer3_protocol() == LAYER3_PROTOCOL_IPV6)
		{
			IPv4_Packet * tempip = (IPv4_Packet *)layer2_frame->get_layer3_packet();
			
			if(layer2_frame->get_layer3_packet()->get_layer4_dgram() != NULL)
			{
				if(layer2_frame->get_layer3_packet()->get_layer4_dgram()->get_layer4_protocol() == LAYER4_PROTOCOL_TCP
					|| layer2_frame->get_layer3_packet()->get_layer4_dgram()->get_layer4_protocol() == LAYER4_PROTOCOL_UDP)
				{
					UDP_Datagram * tempudp = (UDP_Datagram *)layer2_frame->get_layer3_packet()->get_layer4_dgram();
					
					flow_id =  new TCAPS_Flow_ID_Port(tempip->get_src_ip(),
													 tempudp->get_src_port(),
													 tempip->get_dst_ip(),
													 tempudp->get_dst_port());
				}
			}
			else
			{
				flow_id = new TCAPS_Flow_ID(tempip->get_src_ip(), tempip->get_dst_ip());
			}
		}
	}
	
	return flow_id;
}

unsigned int Pcap_Packet::get_payload_len() const
{
	unsigned int header_lengths = 0;
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet packet_len: %d\n", header->len)); 
	
	if(layer2_frame != NULL)
	{
		header_lengths += layer2_frame->get_layer2_hdr_len();
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet layer2_len: %d\n", layer2_frame->get_layer2_hdr_len())); 
		
		if(layer2_frame->get_layer3_packet() != NULL)
		{
			header_lengths += layer2_frame->get_layer3_packet()->get_layer3_hdr_len();
			
			//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet layer3_len: %d\n", layer2_frame->get_layer3_packet()->get_layer3_hdr_len())); 
			
			if(layer2_frame->get_layer3_packet()->get_layer4_dgram() != NULL)
			{
				header_lengths += layer2_frame->get_layer3_packet()->get_layer4_dgram()->get_layer4_hdr_len();
				//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet layer4_len: %d\n", layer2_frame->get_layer3_packet()->get_layer4_dgram()->get_layer4_hdr_len())); 
			}
		}
	}
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Pcap_Packet (header->len)-header_lengths: %d\n", (header->len)-header_lengths)); 
	
	return (header->len)-header_lengths;
}

void Pcap_Packet::parse_packet()
{
	parse_layer2();
		
}

void Pcap_Packet::parse_layer2()
{
	//we'll assume we're sniffing ethernet for the moment...
	if(1)
	{
		layer2_frame = new Ethernet_Frame(packet);
	}
}

void Pcap_Packet::to_string(char * buf, unsigned int buflen)
{
	snprintf(buf, buflen, "PCAP_PACKET\ntotal_length: %d\n", packet_size);
	if(layer2_frame != NULL)
	{
		layer2_frame->to_string(buf+strlen(buf), buflen-strlen(buf));
		
		if(layer2_frame->get_layer3_packet() != NULL)
		{
			layer2_frame->get_layer3_packet()->to_string(buf+strlen(buf), buflen-strlen(buf));
			
			if(layer2_frame->get_layer3_packet()->get_layer4_dgram() != NULL)
			{
				layer2_frame->get_layer3_packet()->get_layer4_dgram()->to_string(buf+strlen(buf), buflen-strlen(buf));
			}
		}
	}
}

Layer2_Frame * Pcap_Packet::get_layer2_frame() const
{
	return layer2_frame;
}
	
