#include "tcaps_fs.h"

static void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet);


/* Set our housekeeping pointer to NULL and tell the user we exist.  */
TCAPS_Sniff_Task::TCAPS_Sniff_Task (size_t n_threads, pcap_t * s, TCAPS_Flow_Manager_Task * fm_task)
 : n_threads_ (n_threads), sniff (s), flow_manager (fm_task)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Sniff_Task ctor 0x%x\n", (void *) this));
}

/* Take care of cleanup & tell the user we're going away.  */
TCAPS_Sniff_Task::~TCAPS_Sniff_Task (void)
{
  ACE_DEBUG ((LM_DEBUG,
              "(%P|%t) TCAPS_Sniff_Task dtor 0x%x\n",
              (void *) this));

  /* Get our shutdown notification out of the queue and release it.  */
  ACE_Message_Block *message;

  /* Like the getq() in svc() below, this will block until a message
    arrives.  By blocking, we know that the destruction will be paused
    until the last thread is done with the message block.  */
  this->getq (message);
  message->release ();
}

/* Open the object to do work.  Next, we activate the Task into the
  number of requested threads.  */
int
TCAPS_Sniff_Task::open (void *unused)
{
  ACE_UNUSED_ARG (unused);

  return this->activate (THR_NEW_LWP,
                         n_threads_);
}

/* Tell the user we're closing and invoke the baseclass' close() to
  take care of things.  */
int
TCAPS_Sniff_Task::close (u_long flags)
{
  ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Sniff_Task close 0x%x\n", (void *) this));
  pcap_close(sniff);
  return inherited::close (flags);
}

/* Our svc() method waits for work on the queue and then processes
  that work.  */
int
TCAPS_Sniff_Task::svc (void)
{
	ACE_Message_Block *message;
	
	//this may be foobed as we are using relative time, and API says to use absolute time...
	ACE_Time_Value wait_for_message(0, 1); //wait 1us
	
	
	struct timespec sleep_timer;
	
	sleep_timer.tv_sec = 0;
	//sleep_timer.tv_nsec = 1000000; //1ms
	sleep_timer.tv_nsec = 500000000; //500ms
	bool finished = false;
	
	
	
	while(!finished)
	{
		if(this->getq(message, &wait_for_message) == -1)
		{
    		//Let's do some sniffing!
    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to enter pcap_loop\n"));
    		pcap_loop(sniff, -1, (pcap_handler)got_packet, (u_char*)this);
    	}
    	
    	else
    	{
    		if (message->msg_type() == ACE_Message_Block::MB_HANGUP)
	    	{
	    		ACE_DEBUG ((LM_DEBUG, "(%P|%t) We've been told to shutdown the sniff thread\n"));
	      		this->putq (message);
	      		finished = true;
	    	}
    		else
    			message->release();
    	}
			
		
		ACE_OS::nanosleep(&sleep_timer); //sleep for sleep_timer.tv_nsec nanosecs to be nice to other processes running on the system
	}
	
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Exiting svc() in sniff thread\n"));


  return 0;
}

pcap_t * TCAPS_Sniff_Task::get_pcap_handle()
{
	return sniff;
}

TCAPS_Flow_Manager_Task * TCAPS_Sniff_Task::get_flow_manager()
{
	return flow_manager;
}


void got_packet(u_char *args, const struct pcap_pkthdr *header, const u_char *packet)
{
	//if there is a message in the task's message queue, we need to break loop so we can have a look at it
	if(!(((TCAPS_Sniff_Task*)args)->msg_queue_->is_empty()))
	{
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to call pcap_breakloop()\n"));
		pcap_breakloop(((TCAPS_Sniff_Task*)args)->get_pcap_handle());
	}
	
	//	char buf[1024];
	//	printHexBytes(packet+36, 2, buf, sizeof(buf), FALSE);
	//	ACE_DEBUG ((LM_DEBUG, "(%P|%t) First 4 bytes of tcp hdr in hex: %s\n", buf));
	
		
	//let's process the packet
	//stuff the packet details we need into a queue for another thread ot handle
	
	//ACE_DEBUG ((LM_DEBUG, "(%P|%t) We just received a packet in callback function of length: %d bytes\n", header->len));
	
    Packet_Block *message;

  	message = new Packet_Block(sizeof(struct pcap_pkthdr) + header->len);
	
	message->copy ((const char *) header, sizeof(struct pcap_pkthdr));
  	message->copy ((const char *) packet, header->len);
  	
  	((TCAPS_Sniff_Task*)args)->get_flow_manager()->putq(message);
  	
}
