#include "tcaps_manager.h"

TCAPS_Network_Task * start_comms_thread(unsigned short local_port);
TCAPS_Client_Manager_Task * start_cm_thread(TCAPS_Network_Task * ci_comms_task, TCAPS_Network_Task * fs_comms_task);

void send_packet(struct tcaps_sp_packet * packet, struct ip_address * recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task);

void check_for_packets(TCAPS_Network_Task * comms_task);
struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote);

void handle_pp_packet(struct tcaps_sp_pp * packet, ACE_INET_Addr * remote);

void testsend();



void cleanup();
void handler(int value);
static sig_atomic_t finished = 0;


TCAPS_Network_Task * fs_comms_task;
TCAPS_Network_Task * fc_comms_task;
TCAPS_Network_Task * ci_comms_task;
TCAPS_Client_Manager_Task * cm_task;




int main (int argc,char *argv[])
{
	struct timespec sleep_timer;
	
	fs_comms_task = NULL;
	fc_comms_task = NULL;
	ci_comms_task = NULL;
	cm_task = NULL;
	
	sleep_timer.tv_sec = 0;
	sleep_timer.tv_nsec = 500000000; //500ms
	
	
	ACE_Sig_Action sa ((ACE_SignalHandler) handler, SIGINT);

	fs_comms_task = start_comms_thread(TCAPS_FS_LOCAL_PORT);
	fc_comms_task = start_comms_thread(TCAPS_FC_LOCAL_PORT);
	ci_comms_task = start_comms_thread(TCAPS_CI_LOCAL_PORT);
	cm_task = start_cm_thread(ci_comms_task, fs_comms_task);


	fs_comms_task->open();
	fc_comms_task->open();
	ci_comms_task->open();
	cm_task->open();

	while(!finished)
	{
        
//		check_for_packets(fs_comms_task);
//		check_for_packets(fc_comms_task);
//		check_for_packets(ci_comms_task);
//		
//		testsend();
		
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) Killing some time in the main thread...\n"));
		//ACE_OS::sleep(1);
		ACE_OS::nanosleep(&sleep_timer); //sleep for a bit to be nice to other processes running on the system
	}
	
	cleanup();

  return 0;
}

TCAPS_Client_Manager_Task * start_cm_thread(TCAPS_Network_Task * ci_comms_task, TCAPS_Network_Task * fs_comms_task)
{
	TCAPS_Client_Manager_Task * cm_task = NULL;
	
	cm_task = new TCAPS_Client_Manager_Task(ci_comms_task, fs_comms_task);
	
	return cm_task;
}


//void testsend()
//{
//	struct tcaps_sp_packet packet;
//	struct ip_address tmpip;
//	
//	packet.packet_type = TCAPS_CMP_PACKET_TYPE;
//	
//	packet.packet = tcaps_sp_cmp_create(TCAPS_SP_CMP_OP_TYPE_MONITOR, TCAPS_SP_IP_VERSION_4, "172.16.7.8");
//	
//	tmpip.ip_version = TCAPS_SP_IP_VERSION_4;
//	tmpip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
//	tcaps_ip_aton("172.16.7.8", &tmpip);
//	
//	send_packet(&packet, &tmpip, 10003, fs_comms_task);
//	
//	tcaps_sp_cmp_destroy((struct tcaps_sp_cmp *)packet.packet);
//	free(tmpip.octets);
//}



void cleanup()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) We are in the cleanup function!\n"));

	/* Once we're done, we have to signal the Task objects to shut
    down. There are several choices including: - Send a message of
    zero length - Send a message with a special content I don't like
    these choices because they're likely to interfere with application
    logic.  Instead, I use the message type feature to send a message
    of type "hangup".  The default type is MB_DATA, so when the tasks
    get a MB_HANGUP type, they know to go away.  */

	if(cm_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the cm_task queue\n"));
		cm_task->putq(message);
		cm_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete cm_task;
	}
	
	if(fs_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fs_comms_task queue\n"));
		fs_comms_task->putq(message);
		fs_comms_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete fs_comms_task;
	}
	
	if(fc_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the fc_comms_task queue\n"));
		fc_comms_task->putq(message);
		fc_comms_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete fc_comms_task;
	}
	
	if(ci_comms_task != NULL)
	{
		Packet_Block *message = new Packet_Block();
		message->msg_type(ACE_Message_Block::MB_HANGUP);
		ACE_DEBUG ((LM_DEBUG, "(%P|%t) About to place the hangup message in the ci_comms_task queue\n"));
		ci_comms_task->putq(message);
		ci_comms_task->wait(); /* Wait for the threads in our task objects to go away.  */
		delete ci_comms_task;
	}

}

void handler(int value)
{
	finished = 1;
}









TCAPS_Network_Task * start_comms_thread(unsigned short local_port)
{
	TCAPS_Network_Task * comms_task = NULL;
	
	ACE_INET_Addr local(local_port);
	
	ACE_SOCK_Dgram * dgram = new ACE_SOCK_Dgram();
	
	if (dgram->open (local) != -1)
	{
		comms_task = new TCAPS_Network_Task(dgram);
	}
	
	return comms_task;
}


//void send_packet(struct tcaps_sp_packet * packet, struct ip_address * recipient_ip, unsigned short recipient_port, TCAPS_Network_Task * comms_task)
//{
//	char iptmp[60];
//	tcaps_ip_ntoa(recipient_ip, iptmp, TRUE);	    
//    ACE_INET_Addr remote(recipient_port, iptmp);
//		
//  	Packet_Block *message;
//  	u8 * serialized;
//  	unsigned int serialized_bytes;
//  	
//  	switch(packet->packet_type)
//  	{
//  		case TCAPS_RMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_rmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_rmp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_PP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_pp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_pp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_TP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_tp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_tp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_AP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_ap *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_ap *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_FFP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_ffp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_ffp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_CRMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_crmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_crmp *)packet->packet)->serialized;
//  			break;
//  		case TCAPS_CMP_PACKET_TYPE:
//  			serialized_bytes = ((struct tcaps_sp_cmp *)packet->packet)->serialized_bytes;
//  			serialized = ((struct tcaps_sp_cmp *)packet->packet)->serialized;
//  			break;
//  		default:
//  			serialized_bytes = 0;
//  			serialized = 0;
//  			break;
//  	}
//
//  	message = new Packet_Block(sizeof(remote) + sizeof(unsigned int) + serialized_bytes);
//  	
//  	message->copy ((const char *) &remote, sizeof(remote));
//  	message->copy ((const char *) &serialized_bytes, sizeof(unsigned int));
//  	message->copy ((const char *) serialized, serialized_bytes);
//
//
//	comms_task->send_packet(message);
//}
//
//
//
//
//struct tcaps_sp_packet * receive_packet(TCAPS_Network_Task * comms_task, ACE_INET_Addr * remote)
//{
//	ACE_Message_Block *message;
//	struct tcaps_sp_packet * packet = NULL;
//	
//	message = comms_task->receive_packet();
//	
//	if(message != NULL)
//	{
//		unsigned int num_serialized_bytes;
//		
//		ACE_OS::memmove ((char *) remote, message->rd_ptr(), sizeof (ACE_INET_Addr));
//		message->rd_ptr (sizeof (ACE_INET_Addr));  // Move ther rd_ptr() beyond the data.
//		
//		ACE_OS::memmove ((char *) &num_serialized_bytes, message->rd_ptr(), sizeof (unsigned int));
//		message->rd_ptr (sizeof (unsigned int));  // Move the rd_ptr() beyond the data.
//		
//		u8 serialized_bytes[num_serialized_bytes];
//		
//		ACE_OS::memmove ((char *) &serialized_bytes, message->rd_ptr(), num_serialized_bytes);
//		message->rd_ptr (num_serialized_bytes);  // Move the rd_ptr() beyond the data.
//		
//		message->release();
//		
//		packet = (struct tcaps_sp_packet *)malloc(sizeof(struct tcaps_sp_packet));
//		
//		tcaps_sp_parse_packet((u8 *)serialized_bytes, num_serialized_bytes, packet);
//	}
//	
//	return packet;
//}
//
//
//void check_for_packets(TCAPS_Network_Task * comms_task)
//{
//	ACE_INET_Addr remote;
//	char buf[BUFFER_SIZE];
//	
//	struct tcaps_sp_packet * packet = receive_packet(comms_task, &remote);
//        
//	while(packet != NULL)
//	{
//		switch(packet->packet_type)
//		{
//			case TCAPS_RMP_PACKET_TYPE:
//					tcaps_sp_rmp_tostring((struct tcaps_sp_rmp *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_rmp_destroy((struct tcaps_sp_rmp *)packet->packet);
//				break;
//			case TCAPS_AP_PACKET_TYPE:
//					tcaps_sp_ap_tostring((struct tcaps_sp_ap *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_ap_destroy((struct tcaps_sp_ap *)packet->packet);
//				break;
//			case TCAPS_TP_PACKET_TYPE:
//					tcaps_sp_tp_tostring((struct tcaps_sp_tp *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_tp_destroy((struct tcaps_sp_tp *)packet->packet);
//				break;
//			case TCAPS_PP_PACKET_TYPE:
//					tcaps_sp_pp_tostring((struct tcaps_sp_pp *)packet->packet, buf, sizeof(buf));
//					handle_pp_packet((struct tcaps_sp_pp * )packet->packet, &remote);
//					tcaps_sp_pp_destroy((struct tcaps_sp_pp *)packet->packet);
//				break;
//			case TCAPS_FFP_PACKET_TYPE:
//					tcaps_sp_ffp_tostring((struct tcaps_sp_ffp *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_ffp_destroy((struct tcaps_sp_ffp *)packet->packet);
//				break;
//			case TCAPS_CRMP_PACKET_TYPE:
//					tcaps_sp_crmp_tostring((struct tcaps_sp_crmp *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_crmp_destroy((struct tcaps_sp_crmp *)packet->packet);
//				break;
//			case TCAPS_CMP_PACKET_TYPE:
//					tcaps_sp_cmp_tostring((struct tcaps_sp_cmp *)packet->packet, buf, sizeof(buf));
//					tcaps_sp_cmp_destroy((struct tcaps_sp_cmp *)packet->packet);
//				break;
//			defaultr:
//				ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received an unexpected tcaps_sp packet type\n"));
//				break;
//		}
//		
//		ACE_DEBUG ((LM_DEBUG, "(%P|%t) Received a tcaps_sp packet that looks like this:\n%s\n", buf));
//
//		free(packet);
//		
//		packet = receive_packet(comms_task, &remote);
//	}
//}
//
//void handle_pp_packet(struct tcaps_sp_pp * packet, ACE_INET_Addr * remote)
//{
//	switch(packet->poll_type)
//	{
//		case TCAPS_PP_POLL_TYPE_REGISTER:
//			break;
//	}
//}

