#include "tcaps_manager.h"


TCAPS_Client::TCAPS_Client()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client ctor 0x%x\n", (void *) this));
	
	client_rules = new TCAPS_Rule_List();
	last_poll_time = (unsigned int)ACE_OS::time();
}

TCAPS_Client::TCAPS_Client (struct mac_address * mac)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client(struct mac_address * mac) ctor 0x%x\n", (void *) this));	

	memcpy(client_mac.octets, mac->octets, NUM_MAC_OCTETS);
	
	char buf[60];
	tcaps_ether_ntoa(&client_mac, buf, TRUE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added client with mac address of: %s\n", buf));
	
	client_rules = new TCAPS_Rule_List();
	last_poll_time = (unsigned int)ACE_OS::time();
}

TCAPS_Client::TCAPS_Client (struct mac_address * mac, struct ip_address * ip, u16 port)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client(struct mac_address * mac, struct ip_address * ip) ctor 0x%x\n", (void *) this));	

	memcpy(client_mac.octets, mac->octets, NUM_MAC_OCTETS);
	
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		client_ip.ip_version = TCAPS_SP_IP_VERSION_4;
		client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		memcpy(client_ip.octets, ip->octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		client_ip.ip_version = TCAPS_SP_IP_VERSION_6;
		client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		memcpy(client_ip.octets, ip->octets, NUM_IPV6_OCTETS);
	}
	
	client_port = port;
	
	char macbuf[60], ipbuf[60];
	tcaps_ether_ntoa(&client_mac, macbuf, TRUE);
	tcaps_ip_ntoa(&client_ip, ipbuf, TRUE);
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) Just added client with mac address of: %s\tip address/port of: %s:%d\n", macbuf, ipbuf, port));
	
	client_rules = new TCAPS_Rule_List();
	last_poll_time = (unsigned int)ACE_OS::time();
}

TCAPS_Client::TCAPS_Client (const TCAPS_Client &copied_client)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client ctor 0x%x\n", (void *) this));
	
	client_port = copied_client.get_client_port();
	client_ip = copied_client.get_client_ip();
	fs_ip = copied_client.get_fs_ip();
	client_mac = copied_client.get_client_mac();
	client_rules = new TCAPS_Rule_List(*copied_client.get_client_rules());
	last_poll_time = copied_client.get_last_poll_time();
}

TCAPS_Client::~TCAPS_Client (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client dtor 0x%x\n", (void *) this));
	
	delete client_rules;
	
	free(client_ip.octets);
	free(fs_ip.octets);
}
	
TCAPS_Client & TCAPS_Client::operator=(const TCAPS_Client &copied_client)
{
	if(this == &copied_client)
		return *this;
	
	client_port = copied_client.get_client_port();
	client_ip = copied_client.get_client_ip();
	fs_ip = copied_client.get_fs_ip();
	client_mac = copied_client.get_client_mac();
	client_rules = new TCAPS_Rule_List(*copied_client.get_client_rules());
	last_poll_time = copied_client.get_last_poll_time();
	
	return *this;
}

unsigned int TCAPS_Client::get_last_poll_time() const
{
	return last_poll_time;
}

void TCAPS_Client::set_last_poll_time(unsigned int time)
{
	last_poll_time = time;
}

u16 TCAPS_Client::get_client_port() const
{
	return client_port;
}

struct ip_address TCAPS_Client::get_client_ip() const
{
	return client_ip;
}

struct ip_address TCAPS_Client::get_fs_ip() const
{
	return fs_ip;
}


struct mac_address TCAPS_Client::get_client_mac() const
{
	return client_mac;
}

TCAPS_Rule_List * TCAPS_Client::get_client_rules() const
{
	return client_rules;
}

void TCAPS_Client::set_client_port(u16 port)
{
	client_port = port;
}

void TCAPS_Client::set_client_ip(const struct ip_address * ip)
{
	if(ip->ip_version != client_ip.ip_version)
	{
		free(client_ip.octets);
		
		if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
		{
			client_ip.ip_version = TCAPS_SP_IP_VERSION_4;
			client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		}
		else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
		{
			client_ip.ip_version = TCAPS_SP_IP_VERSION_6;
			client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		}
	}
	
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		memcpy(client_ip.octets, ip->octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		memcpy(client_ip.octets, ip->octets, NUM_IPV6_OCTETS);
	}
}

void TCAPS_Client::set_fs_ip(const struct ip_address * ip)
{
	if(ip->ip_version != fs_ip.ip_version)
	{
		free(fs_ip.octets);
		
		if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
		{
			fs_ip.ip_version = TCAPS_SP_IP_VERSION_4;
			fs_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		}
		else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
		{
			fs_ip.ip_version = TCAPS_SP_IP_VERSION_6;
			fs_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		}
	}
	
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		memcpy(fs_ip.octets, ip->octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		memcpy(fs_ip.octets, ip->octets, NUM_IPV6_OCTETS);
	}
}

void TCAPS_Client::to_string(char * buf, unsigned int bufsize) const
{
	snprintf(buf, bufsize, "client_mac: ");
	
	tcaps_ether_ntoa(&client_mac, buf+strlen(buf), TRUE);
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\tclient_ip: ");
	
	tcaps_ip_ntoa(&client_ip, buf+strlen(buf), TRUE);
	
	snprintf(buf+strlen(buf), bufsize-strlen(buf), "\tfs_ip: ");
	
	tcaps_ip_ntoa(&fs_ip, buf+strlen(buf), TRUE);
}
	
bool operator<(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	return (memcmp(f1.client_mac.octets, f2.client_mac.octets, NUM_MAC_OCTETS) < 0);
}

bool operator>(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	return (memcmp(f1.client_mac.octets, f2.client_mac.octets, NUM_MAC_OCTETS) > 0);
}

bool operator==(const TCAPS_Client &f1, const TCAPS_Client &f2)
{
	return (memcmp(f1.client_mac.octets, f2.client_mac.octets, NUM_MAC_OCTETS) == 0);
}
