#include "tcaps_manager.h"


TCAPS_Client_List::TCAPS_Client_List ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_List default ctor 0x%x\n", (void *) this));
}


TCAPS_Client_List::TCAPS_Client_List (const TCAPS_Client_List &copied_client_list)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_List copy ctor 0x%x\n", (void *) this));
	
	client_list = copied_client_list.get_client_list();
}

TCAPS_Client_List::~TCAPS_Client_List (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_List dtor 0x%x\n", (void *) this)); 
}
	
TCAPS_Client_List & TCAPS_Client_List::operator=(const TCAPS_Client_List &copied_client_list)
{
	if(this == &copied_client_list)
		return *this;

	client_list = copied_client_list.get_client_list();
	
	return *this;
}



std::map<struct mac_address,TCAPS_Client *,maccmp> TCAPS_Client_List::get_client_list() const
{
	return client_list;
}

void TCAPS_Client_List::add_client(TCAPS_Client * theclient)
{
	client_list_mutex.acquire();
	
	//add client to the list
	
	//client_list.push_back(*theclient);
	
	client_list.insert(client_entry(theclient->get_client_mac(), theclient));

	client_list_mutex.release();
}

bool TCAPS_Client_List::find_client(struct mac_address * clientmac)
{
	std::map<struct mac_address,TCAPS_Client *,maccmp>::iterator client_iter;
	
	client_iter = client_list.find(*clientmac);
	   				 
	if (client_iter != client_list.end())
		return true;
		
	return false;
}

TCAPS_Client * TCAPS_Client_List::get_client(struct mac_address * clientmac)
{
	std::map<struct mac_address,TCAPS_Client *,maccmp>::iterator client_iter;
	
	client_iter = client_list.find(*clientmac);
	   				 
	if (client_iter != client_list.end())
		return client_list[*clientmac];
		
	return NULL;
}

TCAPS_Client * TCAPS_Client_List::get_client(struct ip_address * clientip)
{
	std::map<struct mac_address,TCAPS_Client *,maccmp>::iterator c_itr;
	TCAPS_Client * tmpclient;
	TCAPS_Client * ret = NULL;
	bool found = FALSE;
		
	client_list_mutex.acquire();		
	
	for(c_itr = client_list.begin(); c_itr != client_list.end() && !found; c_itr++)
	{
		tmpclient = ((*c_itr).second);
		
		if(clientip->ip_version == tmpclient->get_client_ip().ip_version)
		{
			if(TCAPS_SP_IP_VERSION_4 == clientip->ip_version)
			{
				if(memcmp(clientip->octets, tmpclient->get_client_ip().octets, NUM_IPV4_OCTETS) == 0)
				{
					ret = tmpclient;
					found = true;
				}
			}
			else if(TCAPS_SP_IP_VERSION_6 == clientip->ip_version)
			{
				if(memcmp(clientip->octets, tmpclient->get_client_ip().octets, NUM_IPV6_OCTETS) == 0)
				{
					ret = tmpclient;
					found = true;
				}
			}
		}
	}
	
	client_list_mutex.release();
	
	return ret;
}

void TCAPS_Client_List::remove_client(struct mac_address * clientmac)
{
	std::map<struct mac_address,TCAPS_Client *,maccmp>::iterator c_itr;
	
	c_itr = client_list.find(*clientmac);
	   				 
	if(c_itr != client_list.end())
	{
		delete ((*c_itr).second);
		client_list.erase(c_itr);
	}
}

void TCAPS_Client_List::acquire_mutex()
{
	client_list_mutex.acquire();
}

void TCAPS_Client_List::release_mutex()
{
	client_list_mutex.release();
}



bool operator==(const TCAPS_Client_List &f1, const TCAPS_Client_List &f2)
{
//	if(//lists are the same size)
//	{
//		if(each client in the list is same)
//		{
//			return true;
//		}
//	}
	
	return false;
}
