#ifndef _TCAPS_MANAGER_CLIENT_LIST_H_
#define _TCAPS_MANAGER_CLIENT_LIST_H_


#include <map>

#include "ace/Synch.h"

typedef ACE_Thread_Mutex mutex_t;
typedef std::pair<struct mac_address,TCAPS_Client *> client_entry;


struct maccmp
{
	bool operator()(const struct mac_address &a, const struct mac_address &b) const
	{
		//ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Client_List.h in the compare function\n"));
    	return (memcmp(a.octets, b.octets, NUM_MAC_OCTETS) < 0);
	}
}; 

class TCAPS_Client_List
{
public:
	TCAPS_Client_List ();
	TCAPS_Client_List (const TCAPS_Client_List &);
	virtual ~TCAPS_Client_List (void);

	std::map<struct mac_address,TCAPS_Client *,maccmp> get_client_list() const;
	void add_client(TCAPS_Client * theclient);
	void acquire_mutex();
	void release_mutex();
	bool find_client(struct mac_address * clientmac);
	TCAPS_Client * get_client(struct mac_address * clientmac);
	TCAPS_Client * get_client(struct ip_address * clientip);
	
	void remove_client(struct mac_address * clientmac);
	
	TCAPS_Client_List & operator=(const TCAPS_Client_List &copied_client_list);
	
	friend bool operator==(const TCAPS_Client_List &f1, const TCAPS_Client_List &f2);

	
  
private:
	mutex_t client_list_mutex; //client_list_mutex.acquire(); client_list_mutex.release();
	std::map<struct mac_address,TCAPS_Client *,maccmp> client_list;
};


#endif //_TCAPS_MANAGER_CLIENT_LIST_H_
