#include "tcaps_manager.h"


TCAPS_Rule::TCAPS_Rule ()
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule default ctor 0x%x\n", (void *) this));
	
	expiry_time_secs = 0;
}

TCAPS_Rule::TCAPS_Rule (struct tcaps_sp_rmp * rule_pack)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule(struct tcaps_sp_rmp * rule_pack) ctor 0x%x\n", (void *) this));
	
	u8 * serialized = (u8 *)malloc(rule_pack->serialized_bytes);
	memcpy(serialized, rule_pack->serialized, rule_pack->serialized_bytes);
	rule = tcaps_sp_rmp_parse_packet(serialized, rule_pack->serialized_bytes);
	expiry_time_secs = (unsigned int)ACE_OS::time() + rule->timeout;
	free(serialized);
}

TCAPS_Rule::TCAPS_Rule (const TCAPS_Rule &copied_rule)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule copy ctor 0x%x\n", (void *) this));
	
	u8 * serialized = (u8 *)malloc(copied_rule.get_serialized_bytes());
	memcpy(serialized, copied_rule.get_serialized(), copied_rule.get_serialized_bytes());
	rule = tcaps_sp_rmp_parse_packet(serialized, copied_rule.get_serialized_bytes());
	expiry_time_secs = copied_rule.get_expiry_time();
	free(serialized);
}

TCAPS_Rule::~TCAPS_Rule (void)
{
	ACE_DEBUG ((LM_DEBUG, "(%P|%t) TCAPS_Rule dtor 0x%x\n", (void *) this));
	
	if(rule != NULL)
		tcaps_sp_rmp_destroy(rule);
}
	
TCAPS_Rule & TCAPS_Rule::operator=(const TCAPS_Rule &copied_rule)
{
	if(this == &copied_rule)
		return *this;
	
	u8 * serialized = (u8 *)malloc(copied_rule.get_serialized_bytes());
	memcpy(serialized, copied_rule.get_serialized(), copied_rule.get_serialized_bytes());
	rule = tcaps_sp_rmp_parse_packet(serialized, copied_rule.get_serialized_bytes());
	expiry_time_secs = copied_rule.get_expiry_time();
	free(serialized);
	
	return *this;
}

void TCAPS_Rule::set_op_type(u8 type)
{
	tcaps_sp_rmp_set_op_type(rule, type);
}

const struct tcaps_sp_rmp * TCAPS_Rule::get_rmp() const
{
	return rule;
}

unsigned int TCAPS_Rule::get_expiry_time() const
{
	return expiry_time_secs;
}

void TCAPS_Rule::set_expiry_time(unsigned int time)
{
	expiry_time_secs = time;
}

u16 TCAPS_Rule::get_rule_num() const
{
	return rule->rule_number;
}

const u8 * TCAPS_Rule::get_serialized() const
{
	return rule->serialized;
}

unsigned int TCAPS_Rule::get_serialized_bytes() const
{
	return rule->serialized_bytes;
}

void TCAPS_Rule::set_identifier(struct mac_address * mac)
{
	char buf[32];
	
	tcaps_ether_ntoa(mac, buf, TRUE);
	
	tcaps_sp_rmp_set_identifier(rule, buf);
}

void TCAPS_Rule::set_timeout(u16 timeout)
{
	tcaps_sp_rmp_set_timeout(rule, timeout);
}

void TCAPS_Rule::set_rule_num(u16 rulenum)
{
	tcaps_sp_rmp_set_rule_number(rule, rulenum);
}


void TCAPS_Rule::to_string(char * buf, unsigned int bufsize) const
{
	tcaps_sp_rmp_tostring(rule, buf, bufsize);
}

bool operator==(const TCAPS_Rule &r1, const TCAPS_Rule &r2)
{
	//if(r1.rule->rule_number == r2.rule->rule_number)
	
	if(r1.rule->serialized_bytes = r2.rule->serialized_bytes)
	{
		if(memcmp(r1.rule->serialized, r2.rule->serialized, r1.rule->serialized_bytes) == 0) 
		{
			return true;
		}
	}
	
	return false;
}
