#include "tcaps_qosss_translator.h"

#ifdef TCAPS_QOSSS_PFALTQ

int tcaps_pf_init(char * interface)
{
	int retval = 0;
	char command[128];
	int dev;
     
     //printf("We made it into tcaps_pf_init()!\n");
     
     //printf("Attempting to load pf.ko\n");
     
     if(kldload(TCAPS_PF_KERNEL_MODULE) == -1)
     {
     	if(EEXIST != errno)
     	{
     		retval = -1;
     		//printf("Failed to load kernel module. Errno info: %s\n", strerror(errno));
     	}
     }
     
     if(0 == retval)
     {
     	//printf("Attempting to initialise file: %s\n", TCAPS_PF_DEFAULT_RC_FILE);
     	retval = tcaps_pf_init_rc_file(TCAPS_PF_DEFAULT_RC_FILE, interface);
     	if(-1 == retval)
     		printf("Failed to initialise file. Errno info: %s\n", strerror(errno));
     }
     	
     if(0 == retval)
     {
     	//printf("Attempting to execute command\n");
     	
//		dev = open(TCAPS_PF_DEFAULT_PF_DEV, O_RDWR);
//        if (dev == -1)
//        	retval = -1;
//        	
//        else
//        {
//        	ioctl(dev, DIOCSTART, NULL);
//        }
     	
     	snprintf(command, sizeof(command), "%s -d > /dev/null; %s -e -f %s", TCAPS_PF_DEFAULT_PFCTL_PATH, TCAPS_PF_DEFAULT_PFCTL_PATH, TCAPS_PF_DEFAULT_RC_FILE);
     	//printf("command: %s\n", command);
     	retval = system(command);
     	
     	if(0 != retval)
     		printf("Failed to execute command. Errno info: %s\n", strerror(errno));
     }
     
     //printf("About to exit. retval: %d\n", retval);
     
     return retval;
}

int tcaps_pf_init_rc_file(char * pf_rc_file, char * interface)
{
	FILE * rc_file;
	boolean rc_file_previously_written = FALSE;
	
	rc_file = fopen(pf_rc_file, "r+");
	
	if(rc_file != NULL)
	{
		//file already exists, do a check on the existing file to see if the tcaps data is already there
		//after calling this function, the stream pointer will be set at the correct position to write to
		rc_file_previously_written = tcaps_pf_check_rc_file(rc_file);
	}
	else
	{
		//file does not exist, so let's try and make a new one
		rc_file = fopen(pf_rc_file, "w");
	}
	
	if(rc_file != NULL)
	{
		//printf("ftell() reports our position before writing Sto be: %d\n", ftell(rc_file));
		
		//if the file existed previously but didn't have tcaps data in it, let's write some explanatory information
		if(!rc_file_previously_written)
		{
			fprintf(rc_file, "\n\n# Don't edit anything below the dashed line below, as it is auto generated by tcaps_ci\n");
			fprintf(rc_file, "#---------------------------------------------------------------------------\n");
		}

		fprintf(rc_file, "%s\n", TCAPS_PF_RC_FILE_STATIC_MARKER);
		fprintf(rc_file, "altq on %s priq queue { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 }\n", interface);
		fprintf(rc_file, "queue 15 priority 0 priq (default)\n");
		fprintf(rc_file, "queue 14 priority 1\n");
		fprintf(rc_file, "queue 13 priority 2\n");
		fprintf(rc_file, "queue 12 priority 3\n");
		fprintf(rc_file, "queue 11 priority 4\n");
		fprintf(rc_file, "queue 10 priority 5\n");
		fprintf(rc_file, "queue 9 priority 6\n");
		fprintf(rc_file, "queue 8 priority 7\n");
		fprintf(rc_file, "queue 7 priority 8\n");
		fprintf(rc_file, "queue 6 priority 9\n");
		fprintf(rc_file, "queue 5 priority 10\n");
		fprintf(rc_file, "queue 4 priority 11\n");
		fprintf(rc_file, "queue 3 priority 12\n");
		fprintf(rc_file, "queue 2 priority 13\n");
		fprintf(rc_file, "queue 1 priority 14\n");
		fprintf(rc_file, "queue 0 priority 15\n");
		fprintf(rc_file, "nat on fxp0 from 192.168.1.0/24 to any -> fxp0\n");
		fprintf(rc_file, "%s\n", TCAPS_PF_RC_FILE_DYNAMIC_MARKER);
		
		fclose(rc_file);
	}
	else
	{
		//printf("Failed to open file %s. Errno info: %s\n", pf_rc_file, strerror(errno));
		return -1;
	}
	
	return 0;
}

//assumes FILE * rc_file is positioned at beginning of file
//returns TRUE if file has previously been written by TCAPS_CI
//returns false if the file has not previously been written by TCAPS_CI
boolean tcaps_pf_check_rc_file(FILE * rc_file)
{
	char buf[strlen(TCAPS_PF_RC_FILE_STATIC_MARKER)+1];
	boolean finished = FALSE;
	int lineno = 0;
	
	while(!feof(rc_file) && !finished)
	{
		fgets(buf, sizeof(buf), rc_file);
		lineno++;
		
		//printf("Comparing %s to %s\n", TCAPS_PF_RC_FILE_STATIC_MARKER, buf);
		
		if(strncmp(TCAPS_PF_RC_FILE_STATIC_MARKER, buf, strlen(TCAPS_PF_RC_FILE_STATIC_MARKER)) == 0)
		{
			//printf("We found the line in the file at line number: %d!\n", lineno);
			//printf("ftell() currently reports our position to be: %d\n", ftell(rc_file));
			//printf("We are about to set fseek to: %d\n", ftell(rc_file)-strlen(TCAPS_PF_RC_FILE_STATIC_MARKER));
			fseek(rc_file, ftell(rc_file)-strlen(TCAPS_PF_RC_FILE_STATIC_MARKER), SEEK_SET);
			//printf("ftell() now reports our position to be: %d\n", ftell(rc_file));
			finished = TRUE;
		}
		else
		{
			//this just reads to the end of the current line and makes the lineno counter reflect actual
			//line in the file
			while(buf[strlen(buf)-1] != '\n' && !feof(rc_file))
				fgets(buf, sizeof(buf), rc_file);
		}
		
	}
			
	return finished;	
}


boolean fexists(char * filename)
{
	FILE * tmp;
	
	tmp = fopen(filename, "r");
	
	if(tmp == NULL)
		return FALSE;
	
	fclose(tmp);
	return TRUE;
}

int tcaps_pf_add_rule(const char * rule_command, const char * interface)
{
	char command[strlen(TCAPS_PF_DEFAULT_PFCTL_PATH)+strlen(TCAPS_PF_DEFAULT_RC_FILE)+128];
	int retval = 0;
	FILE * rc_file;
	
	if(!fexists(TCAPS_PF_DEFAULT_RC_FILE))
	{
		retval = tcaps_pf_init_rc_file(TCAPS_PF_DEFAULT_RC_FILE, interface);
     	if(-1 == retval)
     		printf("Failed to initialise file. Errno info: %s\n", strerror(errno));
	}
	
	if(0 == retval)
	{
		//file exists, we are adding a new rule, so we just append to end of file
		rc_file = fopen(TCAPS_PF_DEFAULT_RC_FILE, "a");
		
		if(rc_file != NULL)
		{
			fprintf(rc_file, "%s\n", rule_command);
			
			fclose(rc_file);
			
			snprintf(command, sizeof(command), "%s -f %s", TCAPS_PF_DEFAULT_PFCTL_PATH, TCAPS_PF_DEFAULT_RC_FILE);
			
			retval = system(command);
		}
		else
			retval = -1;
	}
	
	return retval;
}

int tcaps_pf_delete_rule(const char * rule_command, const char * interface)
{
	int retval = 0;
	FILE * rc_file, *rc_temp_file;
	
	//if the pf rc file has been deleted, let's recreate it and exit
	if(!fexists(TCAPS_PF_DEFAULT_RC_FILE))
	{
		retval = tcaps_pf_init_rc_file(TCAPS_PF_DEFAULT_RC_FILE, interface);
     	if(-1 == retval)
     		printf("Failed to initialise file. Errno info: %s\n", strerror(errno));
	}
	
	else
	{
		char buf[strlen(rule_command)+1];
		boolean finished = FALSE;
		int lineno = 0;
		
		//file exists, we are deleting a rule, so we look through file for rule
		rc_file = fopen(TCAPS_PF_DEFAULT_RC_FILE, "r");
		rc_temp_file = fopen(TCAPS_PF_DEFAULT_RC_TEMP_FILE, "w");
		
		if(rc_file != NULL && rc_temp_file != NULL)
		{
			fgets(buf, sizeof(buf), rc_file);
			
			while(!feof(rc_file))// && !finished)
			{
				
				lineno++;
				
				//printf("Comparing %s to %s\n", rule_command, buf);
				
				if(strncmp(rule_command, buf, strlen(rule_command)) != 0)
				{
					//printf("About to write the line to temp file\n");
					fprintf(rc_temp_file, "%s", buf);
					
					//this just reads to the end of the current line and makes the lineno counter reflect actual
					//line in the file
					//printf("buf[strlen(buf)-1] = %d, \\n = %d\n", buf[strlen(buf)-1], '\n');
					while(buf[strlen(buf)-1] != '\n' && !feof(rc_file))
					{
						fgets(buf, sizeof(buf), rc_file);
						//printf("About to print this to temp file: %s\n", buf);
						if(!feof(rc_file))
							fprintf(rc_temp_file, "%s", buf);
					}

				}
				else
				{
					//printf("We have a match!\n");
				}
				
				fgets(buf, sizeof(buf), rc_file);
			}
			
			fclose(rc_file);
			fclose(rc_temp_file);
			
			rc_file = fopen(TCAPS_PF_DEFAULT_RC_FILE, "w");
			rc_temp_file = fopen(TCAPS_PF_DEFAULT_RC_TEMP_FILE, "r");
			lineno = 0;
			
			if(rc_file != NULL && rc_temp_file != NULL)
			{
				fgets(buf, sizeof(buf), rc_temp_file);
				
				while(!feof(rc_temp_file))// && !finished)
				{
					
					lineno++;
					
					//printf("About to write the line to temp file\n");
					fprintf(rc_file, "%s", buf);
					
					//this just reads to the end of the current line and makes the lineno counter reflect actual
					//line in the file
					//printf("buf[strlen(buf)-1] = %d, \\n = %d\n", buf[strlen(buf)-1], '\n');
					while(buf[strlen(buf)-1] != '\n' && !feof(rc_temp_file));
					{
						fgets(buf, sizeof(buf), rc_temp_file);
						//printf("About to print this to temp file: %s\n", buf);
						if(!feof(rc_temp_file))
							fprintf(rc_file, "%s", buf);
					}
					
					fgets(buf, sizeof(buf), rc_temp_file);
				}
				
				fclose(rc_file);
				fclose(rc_temp_file);
			}
			else
				retval = -1;
		}
		else
			retval = -1;
	}
	
	return retval;
}

void tcaps_translate_rmp2pf(const struct tcaps_sp_rmp * packet, char * interface, char * command)
{
	unsigned int priority = packet->priority_on_match;
	
	sprintf(command, "pass out on %s inet proto ", interface);
	
	translate_rule_segments(packet->rule_segments, command+strlen(command));
	
	//priority will be 0 (highest) to MAX_NUMBER_OF_PRIORITY_QUEUES-1 (lowest)
	//if our received packet has a lower priority assigned to the packet than we can accomodate for,
	//we will assume the server is assigning priority on a 0-255 scale (priority_on_match field size is 1 byte i.e. 255 combos)
	//we will therefore map the priority onto the 0 to MAX_NUMBER_OF_PRIORITY_QUEUES-1 space
	//this behaviour results in a RMP of priority equal to MAX_NUMBER_OF_PRIORITY_QUEUES to be assigned to one of the higher priority queues
	//(probably 0 or 1 depending on MAX_NUMBER_OF_PRIORITY_QUEUES).
	//the priority 255 will be mapped on to the lowest priority queue, which is named the same as the value of MAX_NUMBER_OF_PRIORITY_QUEUES-1
	if(priority > TCAPS_PF_TRANSLATOR_MAX_PRIQS-1)
	{
		//printf("%d\n", (unsigned int)(pow(2,sizeof(packet->priority_on_match)*8))-1);
		
		//if priority = 255, match rule with lowest prioirty queue TCAPS_PF_TRANSLATOR_MAX_PRIQS-1
		//otherwise, adjust the value of received priority to be between 0 and TCAPS_PF_TRANSLATOR_MAX_PRIQS-1
		priority = (priority < (unsigned int)(pow(2,sizeof(packet->priority_on_match)*8)-1) ? (unsigned int)(((float)((TCAPS_PF_TRANSLATOR_MAX_PRIQS-1)*priority))/((float)pow(2,sizeof(packet->priority_on_match)*8))) : TCAPS_PF_TRANSLATOR_MAX_PRIQS-1);
	}
	
	sprintf(command+strlen(command), "queue %d", priority);
	
	//printf("command: %s\n", command);
	
	//sprintf(command+strlen(command), "%s from %s to %s queue %d", interface, packet->priority_on_match);
	
}



void translate_rule_segments(const LINKEDLIST * rule_segs, char * command)
{
	#define TRS_BUF_SIZE 30
	LISTELEMENT tmp;
	char ** parsed_rulesegs;
	unsigned int parsed_ruleseg_types[rule_segs->size];
	struct tcaps_sp_rule_segment * rule_seg;
	unsigned int count = 0, i = 0;
	
	parsed_rulesegs = (char **)calloc(rule_segs->size, TRS_BUF_SIZE);
	
	tmp.SINGLE = rule_segs->start.SINGLE;
			
	if(rule_segs->size > 0)
	{
		while(tmp.SINGLE != NULL)
		{
			rule_seg = (struct tcaps_sp_rule_segment *)tmp.SINGLE->data;
			
			//printf("count: %d\n", count);
			
			translate_rule_segment(rule_seg, parsed_rulesegs+(count*TRS_BUF_SIZE), TRS_BUF_SIZE, &parsed_ruleseg_types[count]);
			count++;
			
			tmp.SINGLE = tmp.SINGLE->nextElement;
		}
	}
	
	for(; i < rule_segs->size; i++)
	{
		if(parsed_ruleseg_types[i] != TCAPS_PF_TRANSLATOR_RULESEG_TYPE_UNKNOWN)
		{
			//printf("rule_seg %d decoded as: %s\n", i+1, parsed_rulesegs+(i*TRS_BUF_SIZE));
		}
		else
		{
			//printf("rule_seg %d decoded as: UNKNOWN\n", i+1);
		}
	}
	
	assemble_rule(parsed_rulesegs, TRS_BUF_SIZE, parsed_ruleseg_types, rule_segs->size, command);
	
	free(parsed_rulesegs);
	
}

void assemble_rule(char ** parsed_rulesegs, unsigned int bufsize, unsigned int * parsed_ruleseg_types, unsigned int num_rulesegs, char * command)
{
	unsigned int i = 0, num_known_ruleseg_types = 0;
	
	
	for(; i < num_rulesegs; i++)
	{
		if(parsed_ruleseg_types[i] != TCAPS_PF_TRANSLATOR_RULESEG_TYPE_UNKNOWN)
		{
			//printf("Found valid ruleseg type!");
			num_known_ruleseg_types++;
		}
	}
	
	if(2 == num_known_ruleseg_types)
	{
		//handle a "SRCIP DSTIP" based rule
		//probably would never receive one of these for TCAPS as we probably
		//won't be identifying flows only by IP address
	}
	
	else if(4 == num_known_ruleseg_types)
	{
		//handle a "SRCIP/SRCPORT DSTIP/DSTPORT" based rule
		
		//use 4 for loops to ensure any order of rule segments will be constructed into a rule
		//suitable for PF, which has SRCIP/SRCPORT DSTIP/DSTPORT order from left to right
		
		sprintf(command, "udp ");
		
		for(i = 0; i < num_rulesegs; i++)
		{
			if(parsed_ruleseg_types[i] == TCAPS_PF_TRANSLATOR_RULESEG_TYPE_SRCIP)
				sprintf(command+strlen(command), "from %s ", parsed_rulesegs+(i*bufsize));
		}
		
		for(i = 0; i < num_rulesegs; i++)
		{
			if(parsed_ruleseg_types[i] == TCAPS_PF_TRANSLATOR_RULESEG_TYPE_SRCPORT)
				sprintf(command+strlen(command), "port %s to ", parsed_rulesegs+(i*bufsize));
		}
		
		for(i = 0; i < num_rulesegs; i++)
		{
			if(parsed_ruleseg_types[i] == TCAPS_PF_TRANSLATOR_RULESEG_TYPE_DSTIP)
				sprintf(command+strlen(command), "%s ", parsed_rulesegs+(i*bufsize));
		}
		
		for(i = 0; i < num_rulesegs; i++)
		{
			if(parsed_ruleseg_types[i] == TCAPS_PF_TRANSLATOR_RULESEG_TYPE_DSTPORT)
				sprintf(command+strlen(command), "port %s ", parsed_rulesegs+(i*bufsize));
		}
			
	}
}


//struct tcaps_sp_rule_segment
//{
//	unsigned int total_segment_length;
//    u8 relative_to;
//    u16 offset;
//    u16 num_bits_to_match;
//    u8 match_rule;
//    u8 conjunction;
//    u8 * ref_bit_sequence;
//};

void translate_rule_segment(const struct tcaps_sp_rule_segment * rule_seg, char * buf, unsigned int bufsize, unsigned int * ruleseg_type)
{
//	struct ip_address ip;
//	
//	if(TCAPS_TP_IP_VERSION_4 == ipversion)
//	{
//		ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
//	}
//	else if (TCAPS_TP_IP_VERSION_6 == ipversion)
//	{
//		ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
//	}
	
	
	
	//printf("rule_seg->relative_to: %d\n", rule_seg->relative_to);
	//printf("rule_seg->offset: %d\n", rule_seg->offset);
	//printf("rule_seg->num_bits_to_match: %d\n", rule_seg->num_bits_to_match);
	
	if(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET == rule_seg->relative_to)
	{
		if(TCAPS_SP_IPV4HDR_SRCIP_OFFSET == rule_seg->offset)
		{
			if(TCAPS_SP_IPV4HDR_SRCIP_BITS >= rule_seg->num_bits_to_match)
			{
				//we have a rule segment attempting to match a src IP address/0-32
				//printf("About to parse src ip\n");
				parse_ipv4_address(rule_seg->ref_bit_sequence, rule_seg->num_bits_to_match, buf, bufsize);
				*ruleseg_type = TCAPS_PF_TRANSLATOR_RULESEG_TYPE_SRCIP;
				//printf("Finished src ip\n");
			}
		}
		
		else if(TCAPS_SP_IPV4HDR_DSTIP_OFFSET == rule_seg->offset)
		{
			if(TCAPS_SP_IPV4HDR_DSTIP_BITS == rule_seg->num_bits_to_match)
			{
				//we have a rule segment attempting to match a dst IP address
				//printf("About to parse dst ip\n");
				parse_ipv4_address(rule_seg->ref_bit_sequence, rule_seg->num_bits_to_match, buf, bufsize);
				*ruleseg_type = TCAPS_PF_TRANSLATOR_RULESEG_TYPE_DSTIP;
				//printf("Finished dst ip\n");
			}
		}
	}
	
	else if(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT == rule_seg->relative_to)
	{
		if(TCAPS_SP_TCPHDR_SRCPORT_OFFSET == rule_seg->offset)
		{
			if(TCAPS_SP_TCPHDR_SRCPORT_BITS == rule_seg->num_bits_to_match)
			{
				//we have a rule segment attempting to match a src TCP/UDP port number
				//printf("About to parse src port\n");
				parse_port(rule_seg->ref_bit_sequence, rule_seg->num_bits_to_match, buf, bufsize);
				*ruleseg_type = TCAPS_PF_TRANSLATOR_RULESEG_TYPE_SRCPORT;
				//printf("Finished src port\n");
			}
		}
		
		else if(TCAPS_SP_TCPHDR_DSTPORT_OFFSET == rule_seg->offset)
		{
			if(TCAPS_SP_TCPHDR_DSTPORT_BITS == rule_seg->num_bits_to_match)
			{
				//we have a rule segment attempting to match a dst TCP/UDP port number
				//printf("About to parse dst port\n");
				parse_port(rule_seg->ref_bit_sequence, rule_seg->num_bits_to_match, buf, bufsize);
				*ruleseg_type = TCAPS_PF_TRANSLATOR_RULESEG_TYPE_DSTPORT;
				//printf("Finished dst port\n");
			}
		}
	}
	
	else
	{
		//We have a rule segment specification that is not supported by pf
		//Probably should return an error code to the tcaps manager
		*ruleseg_type = TCAPS_PF_TRANSLATOR_RULESEG_TYPE_UNKNOWN;
	}
}



void parse_ipv4_address(u8 * ref_bit_sequence, u16 num_bits_to_match, char * buf, unsigned int bufsize)
{
	struct ip_address ip;
	unsigned int i = 0;
	boolean write_zeros = FALSE;
	char test[40];
	
	ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
	
	for(; i < NUM_IPV4_OCTETS; i++)
	{
		if(write_zeros)
		{
			ip.octets[i] = 0;
		}
		else
		{
			if((num_bits_to_match/((i+1)*8.0)) > 1.0)
			{
				ip.octets[i] = *(ref_bit_sequence+i);
			}
			
			else if((num_bits_to_match/((i+1)*8.0)) == 1.0)
			{
				//They have chosen to match on an octet boundary e.g. /8 or /16 or /24 or /32
				ip.octets[i] = *(ref_bit_sequence+i);
				
				//Every octet after this one should be written as zeros irrespective of what was supplied
				write_zeros = TRUE;
			}
			
			else
			{
				//They have chosen to match not on an octet boundary
				
				unsigned int bits_in_current_octet = num_bits_to_match;
				u8 mask = 0;
				
				//if we are not in the first octet
				if(i > 0)
					bits_in_current_octet -= (i*8);
					
				//printf("bits_in_current_octet: %d\n", bits_in_current_octet);
				
				switch(bits_in_current_octet)
				{
					case 1:
						mask = 128; //bitmask: 10000000
						break;
					case 2:
						mask = 192; //bitmask: 11000000
						break;
					case 3:
						mask = 224; //bitmask: 11100000
						break;
					case 4:
						mask = 240; //bitmask: 11110000
						break;
					case 5:
						mask = 248; //bitmask: 11111000
						break;
					case 6:
						mask = 252; //bitmask: 11111100
						break;
					case 7:
						mask = 254; //bitmask: 11111110
						break;
					default:
						//Could only get to here if there is an error in the algorithm above
						break;
				}
				
				ip.octets[i] = (*(ref_bit_sequence+i))&mask;
				
				//Every octet after this one should be written as zeros irrespective of what was supplied
				write_zeros = TRUE;
			}
		}
			
			
	}
	
	ip.ip_version = TCAPS_SP_IP_VERSION_4;
	
	tcaps_ip_ntoa(&ip, test, TRUE);
	
	//printf("\nDecoded address: %s/%d\n", test, num_bits_to_match);
	snprintf(buf, bufsize, "%s/%d", test, num_bits_to_match);
	
	//printf("\nDecoded address: %s/%d\n", test, num_bits_to_match);
}


void parse_port(u8 * ref_bit_sequence, u16 num_bits_to_match, char * buf, unsigned int bufsize)
{
	u16 port = 0;
	
	if(TCAPS_SP_TCPHDR_SRCPORT_BITS >= num_bits_to_match)
	{
		if(TCAPS_SP_TCPHDR_SRCPORT_BITS == num_bits_to_match)
		{
			port = *((u16 *)ref_bit_sequence);
		}
		else
		{
			//Maybe the user is trying to save some bits and not include the full 16 bit ref sequence
			//in their rule segment... maybe every bit is important!
			//We should handle that case here... but won't for the moment
		}
	}
	
	snprintf(buf, bufsize, "%d", port);
	
	//printf("\nDecoded port: %d\n", port);
}



void tcaps_pf_shutdown()
{
	char command[128];
	
	//snprintf(command, sizeof(command), "%s -d", TCAPS_PF_DEFAULT_PFCTL_PATH);
	
	//system(command);
}



#endif //#ifdef TCAPS_QOSSS_PFALTQ

