/**
 * File: tcaps_sp.h
 * 
 * The TCAPS signalling protocol library implements the functionality outlined in the 
 * TCAPS design proposal document. This document was prepared as part of the HET550 final year
 * engineering project Swinburne University of Technology subject.
 * 
 * 
 */

#ifndef _TCAPS_SP_H_
#define _TCAPS_SP_H_


#ifdef __cplusplus
extern "C" {
#endif



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>

#include <structureslib_libs.h>


//general defines

#define TCAPS_SP_IP_VERSION_4 0x01
#define TCAPS_SP_IP_VERSION_6 0x02


//field offsets relevant to all packet types
#define TCAPS_SP_PACKET_TYPE_OFFSET 0



//generic IP/TCP header offsets

#define TCAPS_SP_IPV4HDR_PROTOCOL_OFFSET 72
#define TCAPS_SP_IPV4HDR_SRCIP_OFFSET 96
#define TCAPS_SP_IPV4HDR_DSTIP_OFFSET 128


#define TCAPS_SP_IPV6HDR_NEXTHDR_OFFSET 48
#define TCAPS_SP_IPV6HDR_SRCIP_OFFSET 64
#define TCAPS_SP_IPV6HDR_DSTIP_OFFSET 192

#define TCAPS_SP_TCPHDR_SRCPORT_OFFSET 0
#define TCAPS_SP_TCPHDR_DSTPORT_OFFSET 16
#define TCAPS_SP_TCPHDR_OFFSET_OFFSET 96

#define TCAPS_SP_UDPHDR_SRCPORT_OFFSET TCAPS_SP_TCPHDR_SRCPORT_OFFSET
#define TCAPS_SP_UDPHDR_DSTPORT_OFFSET TCAPS_SP_TCPHDR_DSTPORT_OFFSET


//generic IP/TCP header bit sizes
#define TCAPS_SP_IPV4HDR_SRCIP_BITS 32
#define TCAPS_SP_IPV4HDR_DSTIP_BITS 32

#define TCAPS_SP_IPV6HDR_SRCIP_BITS 128
#define TCAPS_SP_IPV6HDR_DSTIP_BITS 128

#define TCAPS_SP_TCPHDR_SRCPORT_BITS 16
#define TCAPS_SP_TCPHDR_DSTPORT_BITS 16





//the NULL packet type is used to signify errors and should not be used
#define TCAPS_NULL_PACKET_TYPE 0x00
#define TCAPS_RMP_PACKET_TYPE 0x01
#define TCAPS_AP_PACKET_TYPE 0x02
#define TCAPS_TP_PACKET_TYPE 0x03
#define TCAPS_PP_PACKET_TYPE 0x04
#define TCAPS_FFP_PACKET_TYPE 0x05
#define TCAPS_CRMP_PACKET_TYPE 0x06
#define TCAPS_CMP_PACKET_TYPE 0x07

//ACE_DEBUG ((LM_DEBUG, "(%P|%t) num serialized bytes: %d\n", ((struct tcaps_sp_ffp * )packet.packet)->serialized_bytes));
#define TCAPS_RMP_OP_TYPE_CREATE 0x01
#define TCAPS_RMP_OP_TYPE_DELETE 0x02
#define TCAPS_RMP_OP_TYPE_UPDATE 0x03


#define TCAPS_SP_RULESEG_RELTO_START_OF_FRAME			0x01
#define TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET		0x02
#define TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT		0x03
//#define TCAPS_SP_RULESEG_RELTO_START_OF_TCP_DATAGRAM	0x03
//#define TCAPS_SP_RULESEG_RELTO_START_OF_UDP_DATAGRAM	0x04
#define TCAPS_SP_RULESEG_RELTO_START_OF_DATA_SEGMENT	0x05


#define TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO					0x01
#define TCAPS_SP_RULESEG_MATCH_RULE_NOT_EQUAL_TO				0x02
#define TCAPS_SP_RULESEG_MATCH_RULE_LESS_THAN					0x03
#define TCAPS_SP_RULESEG_MATCH_RULE_LESS_THAN_OR_EQUAL_TO		0x04
#define TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN				0x05
#define TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN_OR_EQUAL_TO	0x06
#define TCAPS_SP_RULESEG_MATCH_RULE_LIKE						0x07


#define TCAPS_SP_RULESEG_CONJUNCTION_AND	0x01
#define TCAPS_SP_RULESEG_CONJUNCTION_OR		0x02


#define TCAPS_AP_ACK_TYPE_RMP 0x01
#define TCAPS_AP_ACK_TYPE_TP 0x02
#define TCAPS_AP_ACK_TYPE_PP 0x03

//#define TCAPS_AP_ERROR_CODES { {1, "Error code 1 desc"}, {2, "Error code2 desc"}}




#define TCAPS_PP_POLL_TYPE_REGISTER 0x01
#define TCAPS_PP_POLL_TYPE_UNREGISTER 0x02
#define TCAPS_PP_POLL_TYPE_POLL 0x03




#define TCAPS_TP_TRANSFER_TYPE_RULE 0x01
#define TCAPS_TP_TRANSFER_TYPE_SERVER 0x02
#define TCAPS_TP_TRANSFER_TYPE_CLIENT 0x03






#define TCAPS_FFP_FLOW_ID_TYPE_IPIP						0x01
#define TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT					0x02
#define TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI					0x03

#define TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT					0x01
#define TCAPS_FFP_FLOW_DST_IP_IS_CLIENT					0x02

#define TCAPS_SP_FFP_POINT_TO_FLOW_ID_SRC_OCTETS		0x01
#define TCAPS_SP_FFP_POINT_TO_START_OF_FLOW_FEATURES	0x02
#define TCAPS_SP_FFP_POINT_TO_START_OF_CLIENT_IP_OCTETS 0x03








#define TCAPS_SP_CMP_OP_TYPE_MONITOR		0x01
#define TCAPS_SP_CMP_OP_TYPE_REMOVE			0x02



#define TCAPS_SP_CMP_FC_IP_PRESENT_FALSE	0x00
#define TCAPS_SP_CMP_FC_IP_PRESENT_TRUE		0x01

#define TCAPS_SP_CMP_TM_IP_PRESENT_FALSE	0x00
#define TCAPS_SP_CMP_TM_IP_PRESENT_TRUE		0x01


#define TCAPS_SP_CMP_POINT_TO_START_OF_CLIENT_IP_OCTETS		0x01
#define TCAPS_SP_CMP_POINT_TO_START_OF_TM_IP_OCTETS			0x02
#define TCAPS_SP_CMP_POINT_TO_START_OF_FC_IP_OCTETS			0x03




#define NUM_MAC_OCTETS 6
#define MAX_MAC_CHARS 17
#define MAC_SEPARATOR_CHAR ':'

#define BYTE_SEPARATOR_CHAR ':'

#define NUM_IPV4_OCTETS 4
#define NUM_IPV6_OCTETS 16
#define IPV4_SEPARATOR_CHAR '.'
#define IPV6_SEPARATOR_CHAR ':'
#define MAX_IPV4_CHARS 15

#define LAYER2_PROTOCOL_ETHERNET	0x01


#define ETHERNET_FRAME_TYPE_IPV4	0x0800
#define ETHERNET_FRAME_TYPE_IPV6	0x86DD


#define LAYER3_PROTOCOL_IPV4	0x01
#define LAYER3_PROTOCOL_IPV6	0x02

//http://www.iana.org/assignments/protocol-numbers
#define IPV4_PACKET_PROTOCOL_TCP	6
#define IPV4_PACKET_PROTOCOL_UDP	17

#define IPV6_PACKET_NEXTHDR_TCP		IPV4_PACKET_PROTOCOL_TCP
#define IPV6_PACKET_NEXTHDR_UDP		IPV4_PACKET_PROTOCOL_UDP


#define LAYER4_PROTOCOL_TCP		0x01
#define LAYER4_PROTOCOL_UDP		0x02


#ifndef NULL
#define NULL 0
#endif

#ifndef BOOLEAN
#define BOOLEAN
typedef enum
{
    FALSE=0, TRUE=1
} boolean;
#endif




//typedef for an unsigned 8 bit type
typedef unsigned char u8;

//typedef for an unsigned 16 bit type
typedef unsigned short u16;

//typedef for an unsigned 32 bit type
typedef unsigned int u32;

//typedef for an unsigned 64 bit type
typedef long long u64;


typedef struct mac_address
{
	u8 octets[NUM_MAC_OCTETS];
} tcaps_id;

struct ip_address
{
	u8 ip_version;
	//given that IPv6 addresses are larger than IPv4 addresses,
	//we will need to dynamically figure out how large to make our byte array
	u8 * octets;
	
	#ifdef __cplusplus
	bool operator==(const struct ip_address &b) const
	{
		printf("tcaps_sp.h in the ip_address struct ==operator function\n");
		if(ip_version == b.ip_version)
		{
			if(TCAPS_SP_IP_VERSION_4 == ip_version)
    			return (memcmp(octets, b.octets, NUM_IPV4_OCTETS) == 0);
    		else if(TCAPS_SP_IP_VERSION_6 == ip_version)
    			return (memcmp(octets, b.octets, NUM_IPV6_OCTETS) == 0);
		}
		return false;
	}
	struct ip_address * operator=(const struct ip_address * copied_ip)
	{
		printf("tcaps_sp.h in the ip_address struct =operator function\n");
		
		ip_version = copied_ip->ip_version;
		
		if(TCAPS_SP_IP_VERSION_4 == ip_version)
			memcpy(octets, copied_ip->octets, NUM_IPV4_OCTETS);
		else if(TCAPS_SP_IP_VERSION_6 == ip_version)
			memcpy(octets, copied_ip->octets, NUM_IPV6_OCTETS);
			
		return this;
		
	}
	#endif
};

struct tcaps_sp_rmp
{
    u8 packet_type;
    tcaps_id identifier;
    u16 rule_number;
    u8 transfer_id;
    u8 num_retries;
    u8 op_type;
    u8 priority_on_match;
    u16 timeout;
    u8 num_rule_segments;
    
    LINKEDLIST * rule_segments;
    
    //1+6+2+1+1+1+1+2+1
    #define TCAPS_SP_RMP_STATIC_SERIALIZED_BYTES 16
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};

struct tcaps_sp_rule_segment
{
	unsigned int total_segment_length;
    u8 relative_to;
    u16 offset;
    u16 num_bits_to_match;
    u8 match_rule;
    u8 conjunction;
    u8 * ref_bit_sequence;
};

struct tcaps_sp_crmp
{
	u8 packet_type; //1 static
	struct ip_address client_ip; //1 static + {4|16} dynamic
	struct tcaps_sp_rmp * partial_rmp;
	
	//1+1
    #define TCAPS_SP_CRMP_STATIC_SERIALIZED_BYTES 2
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};

//the serialized form of this packet should always look like this in memory
//packet_type|mac|ack_type|error_code|rule_number|transfer_id
struct tcaps_sp_ap
{
    u8 packet_type;
    tcaps_id identifier;
    u8 ack_type;
    u16 error_code;
    u16 rule_number;
    u8 transfer_id;
    
	//1+6+1+2+2+1
	#define TCAPS_SP_AP_STATIC_SERIALIZED_BYTES 13

    unsigned int serialized_bytes;
    u8 * serialized;
};



struct tcaps_sp_tp
{
    u8 packet_type; //static
    //tcaps_id identifier;
    u8 transfer_type; //static
    u8 transfer_id; //static
    u16 number_of_rules; //static
    //u8 ip_version; //static
    u16 server_port; //static
    struct ip_address server_ip; //1 static + {4|16} dynamic
    
    //1+1+1+2+1+2
	#define TCAPS_SP_TP_STATIC_SERIALIZED_BYTES 8
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};



struct tcaps_sp_pp
{
    u8 packet_type;
    tcaps_id identifier;
    u8 poll_type;
    u8 num_info_segments;
    
    LINKEDLIST * info_segments;
    
    //1+6+1+1
    #define TCAPS_SP_PP_STATIC_SERIALIZED_BYTES 9
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};

struct tcaps_sp_info_segment
{
	unsigned int total_segment_length;
	u8 key_length;
    u8 * key;
    u8 * value;
    //unsigned int data_type;
};





struct tcaps_sp_flow_id_type_ipip
{
	u8 * src_octets;
	u8 * dst_octets;
};

struct tcaps_sp_flow_id_type_ipipport
{
	u8 * src_octets;
	u8 * dst_octets;
	u16 src_port;
	u16 dst_port;
};

struct tcaps_sp_flow_id_type_ipipspi
{
	u8 * src_octets;
	u8 * dst_octets;
	u32 src_spi;
	u32 dst_spi;
};

struct tcaps_sp_flow_id
{
	u8 flow_id_type;
	u8 which_is_client_ip;
	u8 clientip_version;
	unsigned int dynamic_flow_id_length;
	
	union
	{
		struct tcaps_sp_flow_id_type_ipip ipip;
		struct tcaps_sp_flow_id_type_ipipport ipipport;
		struct tcaps_sp_flow_id_type_ipipspi ipipspi;
	};
};


struct tcaps_sp_ffp
{
    u8 packet_type; //static
    tcaps_id identifier; //static
    u16 num_flow_feature_segments; //static
    struct ip_address tm_ip; //1 static + {4|16} dynamic
    struct tcaps_sp_flow_id flow_id; //3 static + some dynamic
    LINKEDLIST * feature_segments;
    
    //1+6+2+1+3
    #define TCAPS_SP_FFP_STATIC_SERIALIZED_BYTES 13
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};

struct tcaps_sp_feature_segment
{
    u16 key;
    float value;
};






struct tcaps_sp_cmp
{
    u8 packet_type; //static
    u8 op_type; //static
	u8 fc_ip_present; //static
	u8 tm_ip_present; //static
    struct ip_address client_ip; //1 static + {4|16} dynamic
    struct ip_address fc_ip; //optional
    struct ip_address tm_ip; //optional
    
    
    //1+1+1+1+1
	#define TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES 5
    
    unsigned int serialized_bytes;
    
    u8 * serialized;
};




struct tcaps_sp_packet
{
	void * packet;
	unsigned int packet_type;
};

//non specific packet functions
void tcaps_sp_parse_packet(u8 * recv_packet, unsigned int packet_len, struct tcaps_sp_packet * packet);



//rule management packet functions
struct tcaps_sp_rmp * tcaps_sp_rmp_create(char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout);
struct tcaps_sp_rmp * tcaps_sp_rmp_create_ipport2ipport(char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, char * srcip, u16 srcport, char * dstip, u16 dstport, unsigned int ipversion);
void tcaps_sp_rmp_destroy(struct tcaps_sp_rmp * rule_pack);
void tcaps_sp_rmp_destroy_rule_segment(struct tcaps_sp_rule_segment * rule_seg);
struct tcaps_sp_rule_segment * tcaps_sp_rmp_create_rule_segment(u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength);
void tcaps_sp_rmp_add_rule_segment(struct tcaps_sp_rmp * rule_pack, struct tcaps_sp_rule_segment * rule_seg);
void tcaps_sp_rmp_set_packet_type(struct tcaps_sp_rmp * rule_pack, u8 type);
void tcaps_sp_rmp_set_identifier(struct tcaps_sp_rmp * rule_pack, char * mac);
void tcaps_sp_rmp_set_rule_number(struct tcaps_sp_rmp * rule_pack, u16 rulenum);
void tcaps_sp_rmp_set_transfer_id(struct tcaps_sp_rmp * rule_pack, u8 transferid);
void tcaps_sp_rmp_set_num_retries(struct tcaps_sp_rmp * rule_pack, u8 numretries);
void tcaps_sp_rmp_set_op_type(struct tcaps_sp_rmp * rule_pack, u8 optype);
void tcaps_sp_rmp_set_priority_on_match(struct tcaps_sp_rmp * rule_pack, u8 matchpri);
void tcaps_sp_rmp_set_timeout(struct tcaps_sp_rmp * rule_pack, u16 timeout);
void tcaps_sp_rmp_set_num_rule_segments(struct tcaps_sp_rmp * rule_pack, u8 numrulesegs);
void tcaps_sp_rmp_tostring(const struct tcaps_sp_rmp * rule_pack, char * string, unsigned int stringlen);
struct tcaps_sp_rmp * tcaps_sp_rmp_parse_packet(u8 * recv_packet, unsigned int packet_len);
struct tcaps_sp_rule_segment * tcaps_sp_rmp_parse_rule_segment(u8 * recv_segment, unsigned int segment_len);
boolean tcaps_sp_rmp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_rmp_compare_packets(const struct tcaps_sp_rmp * rule_pack1, const struct tcaps_sp_rmp * rule_pack2);
boolean tcaps_sp_rmp_compare_rulesegs(const struct tcaps_sp_rule_segment * rule_seg1, const struct tcaps_sp_rule_segment * rule_seg2);
boolean tcaps_sp_rmp_compare_rules(const struct tcaps_sp_rmp * rule_pack1, const struct tcaps_sp_rmp * rule_pack2);

//classifier rule management packet functions
struct tcaps_sp_crmp * tcaps_sp_crmp_create(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout);
struct tcaps_sp_crmp * tcaps_sp_crmp_create_ip2ip(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, char * dstip);
struct tcaps_sp_crmp * tcaps_sp_crmp_create_ipport2ipport(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, u16 srcport, char * dstip, u16 dstport);
struct tcaps_sp_crmp * tcaps_sp_crmp_create_ipspi2ipspi(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, u32 srcspi, char * dstip, u32 dstspi);
void tcaps_sp_crmp_destroy(struct tcaps_sp_crmp * fcrule_pack);
struct tcaps_sp_rule_segment * tcaps_sp_crmp_create_rule_segment(u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength);
void tcaps_sp_crmp_add_rule_segment(struct tcaps_sp_crmp * fcrule_pack, struct tcaps_sp_rule_segment * rule_seg);
void tcaps_sp_crmp_set_packet_type(struct tcaps_sp_crmp * fcrule_pack, u8 type);
void tcaps_sp_crmp_set_clientip_version(struct tcaps_sp_crmp * fcrule_pack, u8 clientipversion);
void tcaps_sp_crmp_set_client_ip(struct tcaps_sp_crmp * fcrule_pack, char * clientip);
void tcaps_sp_crmp_set_partial_rmp(struct tcaps_sp_crmp * fcrule_pack, struct tcaps_sp_rmp * rmp);
void tcaps_sp_crmp_tostring(const struct tcaps_sp_crmp * fcrule_pack, char * string, unsigned int stringlen);
struct tcaps_sp_crmp * tcaps_sp_crmp_parse_packet(u8 * recv_packet, unsigned int packet_len);
boolean tcaps_sp_crmp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_crmp_compare_packets(const struct tcaps_sp_crmp * fcrule_pack1, const struct tcaps_sp_crmp * fcrule_pack2);


//acknowledgment packet functions
struct tcaps_sp_ap * tcaps_sp_ap_create(char * mac, u8 acktype, u16 errorcode, u16 rulenum, u8 transid);
void tcaps_sp_ap_destroy(struct tcaps_sp_ap * ack_pack);
void tcaps_sp_ap_set_packet_type(struct tcaps_sp_ap * ack_pack, u8 type);
void tcaps_sp_ap_set_ack_type(struct tcaps_sp_ap * ack_pack, u8 acktype);
void tcaps_sp_ap_set_identifier(struct tcaps_sp_ap * ack_pack, char * mac);
void tcaps_sp_ap_set_transfer_id(struct tcaps_sp_ap * ack_pack, u8 transid);
void tcaps_sp_ap_set_rule_number(struct tcaps_sp_ap * ack_pack, u16 rulenum);
void tcaps_sp_ap_set_error_code(struct tcaps_sp_ap * ack_pack, u16 errorcode);
void tcaps_sp_ap_tostring(const struct tcaps_sp_ap * ack_pack, char * string, unsigned int stringlen);
struct tcaps_sp_ap * tcaps_sp_ap_parse_packet(u8 * recv_packet, unsigned int packet_len);
boolean tcaps_sp_ap_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_ap_compare_packets(const struct tcaps_sp_ap * ack_pack1, const struct tcaps_sp_ap * ack_pack2);


//transfer packet functions
struct tcaps_sp_tp * tcaps_sp_tp_create(u8 transtype, u8 transid, u16 numrules, u8 ipversion, char * serverip, u16 serverport);
void tcaps_sp_tp_destroy(struct tcaps_sp_tp * trans_pack);
void tcaps_sp_tp_set_packet_type(struct tcaps_sp_tp * trans_pack, u8 type);
void tcaps_sp_tp_set_transfer_type(struct tcaps_sp_tp * trans_pack, u8 transtype);
void tcaps_sp_tp_set_transfer_id(struct tcaps_sp_tp * trans_pack, u8 transid);
void tcaps_sp_tp_set_number_of_rules(struct tcaps_sp_tp * trans_pack, u16 numrules);
void tcaps_sp_tp_set_ip_version(struct tcaps_sp_tp * trans_pack, u8 ipversion);
void tcaps_sp_tp_set_server_ip(struct tcaps_sp_tp * trans_pack, char * serverip);
void tcaps_sp_tp_set_server_port(struct tcaps_sp_tp * trans_pack, u16 serverport);
void tcaps_sp_tp_tostring(const struct tcaps_sp_tp * trans_pack, char * string, unsigned int stringlen);
struct tcaps_sp_tp * tcaps_sp_tp_parse_packet(u8 * recv_packet, unsigned int packet_len);
boolean tcaps_sp_tp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_tp_compare_packets(const struct tcaps_sp_tp * trans_pack1, const struct tcaps_sp_tp * trans_pack2);

//poll packet functions
struct tcaps_sp_pp * tcaps_sp_pp_create(char * mac, u8 polltype);
void tcaps_sp_pp_destroy(struct tcaps_sp_pp * poll_pack);
void tcaps_sp_pp_set_packet_type(struct tcaps_sp_pp * poll_pack, u8 type);
void tcaps_sp_pp_set_identifier(struct tcaps_sp_pp * poll_pack, char * mac);
void tcaps_sp_pp_set_poll_type(struct tcaps_sp_pp * poll_pack, u8 polltype);
void tcaps_sp_pp_set_num_info_segments(struct tcaps_sp_pp * poll_pack, u8 numinfosegs);
struct tcaps_sp_info_segment * tcaps_sp_pp_create_info_segment(u8 * key, u8 keylength, u8 * value, unsigned int valuelength);
void tcaps_sp_pp_destroy_info_segment(struct tcaps_sp_info_segment * info_seg);
//void tcaps_sp_pp_destroy_info_segment(void * info_seg);
void tcaps_sp_pp_add_info_segment(struct tcaps_sp_pp * poll_pack, struct tcaps_sp_info_segment * info_seg);
void tcaps_sp_pp_tostring(const struct tcaps_sp_pp * poll_pack, char * string, unsigned int stringlen);
struct tcaps_sp_pp * tcaps_sp_pp_parse_packet(u8 * recv_packet, unsigned int packet_len);
struct tcaps_sp_info_segment * tcaps_sp_pp_parse_info_segment(u8 * recv_segment, unsigned int segment_len);
boolean tcaps_sp_pp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_pp_compare_packets(const struct tcaps_sp_pp * poll_pack1, const struct tcaps_sp_pp * poll_pack2);
boolean tcaps_sp_pp_compare_infosegs(const struct tcaps_sp_info_segment * info_seg1, const struct tcaps_sp_info_segment * info_seg2);


//flow feature packet functions
struct tcaps_sp_ffp * tcaps_sp_ffp_create(char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, const u8 * flowid);
struct tcaps_sp_ffp * tcaps_sp_ffp_create_withfeatures(char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, const u8 * flowid, u8 * serialized_features, u16 num_features);
void tcaps_sp_ffp_destroy(struct tcaps_sp_ffp * feature_pack);
void tcaps_sp_ffp_set_packet_type(struct tcaps_sp_ffp * feature_pack, u8 type);
void tcaps_sp_ffp_set_identifier(struct tcaps_sp_ffp * feature_pack, char * mac);
void tcaps_sp_ffp_set_flow_id_type(struct tcaps_sp_ffp * feature_pack, u8 flowidtype);
void tcaps_sp_ffp_set_clientip_version(struct tcaps_sp_ffp * feature_pack, u8 clientipversion);
void tcaps_sp_ffp_set_which_is_client_ip(struct tcaps_sp_ffp * feature_pack, u8 whichipisclient);
void tcaps_sp_ffp_set_flow_id(struct tcaps_sp_ffp * feature_pack, u8 * flowid);
void tcaps_sp_ffp_set_num_feature_segments(struct tcaps_sp_ffp * feature_pack, u16 numfeaturesegs);
void tcaps_sp_ffp_set_tmip_version(struct tcaps_sp_ffp * feature_pack, u8 tmipversion);
void tcaps_sp_ffp_set_tm_ip(struct tcaps_sp_ffp * feature_pack, char * tmip);
struct tcaps_sp_feature_segment * tcaps_sp_ffp_create_feature_segment(u16 key, float value);
void tcaps_sp_ffp_destroy_feature_segment(struct tcaps_sp_feature_segment * feature_seg);
void tcaps_sp_ffp_add_feature_segment(struct tcaps_sp_ffp * feature_pack, struct tcaps_sp_feature_segment * feature_seg);
u8 * tcaps_sp_ffp_field_ptr(struct tcaps_sp_ffp * feature_pack, unsigned int field_to_point_to);
void tcaps_sp_ffp_tostring(const struct tcaps_sp_ffp * feature_pack, char * string, unsigned int stringlen);
struct tcaps_sp_ffp * tcaps_sp_ffp_parse_packet(u8 * recv_packet, unsigned int packet_len);
struct tcaps_sp_feature_segment * tcaps_sp_ffp_parse_feature_segment(u8 * recv_segment);
boolean tcaps_sp_ffp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_ffp_compare_packets(const struct tcaps_sp_ffp * feature_pack1, const struct tcaps_sp_ffp * feature_pack2);
boolean tcaps_sp_ffp_compare_featuresegs(const struct tcaps_sp_feature_segment * feature_seg1, const struct tcaps_sp_feature_segment * feature_seg2);


//client management packet functions
struct tcaps_sp_cmp * tcaps_sp_cmp_create(u8 optype, u8 clientipversion, char * clientip);
struct tcaps_sp_cmp * tcaps_sp_cmp_create_withfc(u8 optype, u8 clientipversion, char * clientip, u8 fcipversion, char * fcip);
struct tcaps_sp_cmp * tcaps_sp_cmp_create_withtm(u8 optype, u8 clientipversion, char * clientip, u8 tmipversion, char * tmip);
struct tcaps_sp_cmp * tcaps_sp_cmp_create_withfctm(u8 optype, u8 clientipversion, char * clientip, u8 fcipversion, char * fcip, u8 tmipversion, char * tmip);
void tcaps_sp_cmp_destroy(struct tcaps_sp_cmp * cm_pack);
void tcaps_sp_cmp_set_packet_type(struct tcaps_sp_cmp * cm_pack, u8 type);
void tcaps_sp_cmp_set_op_type(struct tcaps_sp_cmp * cm_pack, u8 optype);
void tcaps_sp_cmp_set_fc_ip_present(struct tcaps_sp_cmp * cm_pack, u8 ippresent);
void tcaps_sp_cmp_set_tm_ip_present(struct tcaps_sp_cmp * cm_pack, u8 ippresent);
void tcaps_sp_cmp_set_client_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion);
void tcaps_sp_cmp_set_client_ip(struct tcaps_sp_cmp * cm_pack, char * clientip);
void tcaps_sp_cmp_set_fc_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion);
void tcaps_sp_cmp_set_fc_ip(struct tcaps_sp_cmp * cm_pack, char * fcip);
void tcaps_sp_cmp_set_tm_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion);
void tcaps_sp_cmp_set_tm_ip(struct tcaps_sp_cmp * cm_pack, char * tmip);
u8 * tcaps_sp_cmp_field_ptr(struct tcaps_sp_cmp * cm_pack, unsigned int field_to_point_to);
void tcaps_sp_cmp_tostring(const struct tcaps_sp_cmp * cm_pack, char * string, unsigned int stringlen);
struct tcaps_sp_cmp * tcaps_sp_cmp_parse_packet(u8 * recv_packet, unsigned int packet_len);
boolean tcaps_sp_cmp_validate_serialized_packet(u8 * packet, unsigned int packet_len);
boolean tcaps_sp_cmp_compare_packets(const struct tcaps_sp_cmp * cm_pack1, const struct tcaps_sp_cmp * cm_pack2);


//utility functions
void tcaps_ether_aton(const char * string, tcaps_id * mac);
void tcaps_ether_ntoa(const tcaps_id * mac, char * string, boolean separate);
u8 hexCharToInt(const char c);
boolean isHexDigit(const char c);
char intToHexChar(const int c);
void tcaps_ip_aton(const char * string, struct ip_address * ip);
void tcaps_ip_ntoa(const struct ip_address * ip, char * string, boolean separate);
void printHexBytes(const u8 * bytes, unsigned int num_bytes, char * string, unsigned int stringlen, boolean hex_order_high_first);
void tcaps_sp_get_interface_mac(tcaps_id * mac, const char * interface);
u8 * tcaps_atohex(const char * hexstring);

//typedef tcaps_info_code enum
//{
//    INCOMPLETE_PACKET
//};
//
//
//
//struct tcaps_rmp
//{
//    packet_type;
//    identifier;
//    rule_number;
//    transfer_id;
//    num_retries;
//    op_type;
//    priority_on_match;
//    timeout;
//    num_segments;
//    
//    LINKEDLIST rule_segments;
//    
//};
//
//struct tcaps_rule_segment
//{
//    relative_to;
//    offset;
//    num_bits_to_match;
//    void * ref_bit_sequence;
//    match rule;
//    conjunction;
//}
//
//struct tcaps_ap
//{
//    packet_type;
//    identifier;
//    ack_type;
//    error_code;
//    rule_number;
//    transfer_id;
//    
//};
//
//struct tcaps_tp
//{
//    packet_type;
//    //??identifier;??
//    transfer_type;
//    transfer_id;
//    number_of_rules;
//    ip_version;
//    server_ip;
//    server_port;
//    
//};
//
//struct tcaps_pp
//{
//    packet_type;
//    identifier;
//    poll_type;
//    num_info_segments;
//    
//    LINKEDLIST info_segments;
//};
//
//struct info_segment;
//{
//    char * key;
//    char * value;
//};




////Function prototypes
//tcaps_packet *create_packet(int packet_type);
//
//
//
////rmp = rule management packet
////ap = acknowledgement packet
////tp = transfer packet
////pp = poll packet
////replicate each of these functions for each value to set within the relevant struct
//void rmp_set_x(tcaps_packet *, value);
//void ap_set_x(tcaps_packet *, value);
//void tp_set_x(tcaps_packet *, value);
//void pp_set_x(tcaps_packet *, value);
//
//
//
//<type> rmp_get_x(tcaps_packet *);
//<type> ap_get_x(tcaps_packet *);
//<type> tp_get_x(tcaps_packet *);
//<type> pp_get_x(tcaps_packet *);
//
//
//void * serialize(tcaps_packet *);
//tcaps_packet * reconstruct(void *);
//
//int validate_packet(tcaps_packet *);
//
//
//
//
//rmp_add_rule_segment(tcaps_packet *, rule_segment *);
//
//rule_segment * create_rule_segment(values for each field in struct);
//
//pp_add_info_segment(tcaps_packet *, info_segment *);
//
//info_segment * create_info_segment(char * key, char * value);

#ifdef __cplusplus
}
#endif

#endif //_TCAPS_SP_H_




  



