#include "tcaps_sp.h"


//    u8 packet_type;
//    tcaps_id identifier;
//    u8 ack_type;
//    u16 error_code;
//    u16 rule_number;
//    u8 transfer_id;

#define TCAPS_SP_AP_ACK_TYPE_OFFSET TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_AP_IDENTIFIER_OFFSET TCAPS_SP_PACKET_TYPE_OFFSET+2
#define TCAPS_SP_AP_ERROR_CODE_OFFSET TCAPS_SP_PACKET_TYPE_OFFSET+8
#define TCAPS_SP_AP_RULE_NUMBER_OFFSET TCAPS_SP_PACKET_TYPE_OFFSET+10
#define TCAPS_SP_AP_TRANSFER_ID_OFFSET TCAPS_SP_PACKET_TYPE_OFFSET+12

    
struct tcaps_sp_ap * tcaps_sp_ap_create(char * mac, u8 acktype, u16 errorcode, u16 rulenum, u8 transid)
{
	struct tcaps_sp_ap * ack_pack;
	
	ack_pack = (struct tcaps_sp_ap *) malloc(sizeof(struct tcaps_sp_ap));
	
	ack_pack->serialized_bytes = TCAPS_SP_AP_STATIC_SERIALIZED_BYTES;
	
	ack_pack->serialized = malloc(ack_pack->serialized_bytes);
	
	
	
	tcaps_sp_ap_set_packet_type(ack_pack, TCAPS_AP_PACKET_TYPE);
	tcaps_sp_ap_set_identifier(ack_pack, mac);
	tcaps_sp_ap_set_ack_type(ack_pack, acktype);
	tcaps_sp_ap_set_error_code(ack_pack, errorcode);
	tcaps_sp_ap_set_rule_number(ack_pack, rulenum);
	tcaps_sp_ap_set_transfer_id(ack_pack, transid);
	
	return ack_pack;
}

void tcaps_sp_ap_destroy(struct tcaps_sp_ap * ack_pack)
{
	free(ack_pack->serialized);
	free(ack_pack);
}
	

void tcaps_sp_ap_set_packet_type(struct tcaps_sp_ap * ack_pack, u8 type)
{
	ack_pack->packet_type = type;
	
	//set the bytes that hold the packet_type field to the tye passed in
	*((ack_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}

void tcaps_sp_ap_set_ack_type(struct tcaps_sp_ap * ack_pack, u8 acktype)
{
	ack_pack->ack_type = acktype;
	
	*((ack_pack->serialized)+TCAPS_SP_AP_ACK_TYPE_OFFSET) = acktype;
}

void tcaps_sp_ap_set_identifier(struct tcaps_sp_ap * ack_pack, char * mac)
{
	tcaps_ether_aton(mac, &ack_pack->identifier);
	
	memcpy(((ack_pack->serialized)+TCAPS_SP_AP_IDENTIFIER_OFFSET), ack_pack->identifier.octets, sizeof(ack_pack->identifier));
}

void tcaps_sp_ap_set_error_code(struct tcaps_sp_ap * ack_pack, u16 errorcode)
{
	ack_pack->error_code = errorcode;
	
	*(u16 *)((ack_pack->serialized)+TCAPS_SP_AP_ERROR_CODE_OFFSET) = errorcode;
}

void tcaps_sp_ap_set_rule_number(struct tcaps_sp_ap * ack_pack, u16 rulenum)
{
	ack_pack->rule_number = rulenum;
	
	*(u16 *)((ack_pack->serialized)+TCAPS_SP_AP_RULE_NUMBER_OFFSET) = rulenum;
}

void tcaps_sp_ap_set_transfer_id(struct tcaps_sp_ap * ack_pack, u8 transid)
{
	ack_pack->transfer_id = transid;
	
	*((ack_pack->serialized)+TCAPS_SP_AP_TRANSFER_ID_OFFSET) = transid;
}

/**
 * Max string size output: 155 chars
 */
void tcaps_sp_ap_tostring(const struct tcaps_sp_ap * ack_pack, char * string, unsigned int stringlen)
{
	char mactmp[20], bytestmp[50];
	unsigned int remainingchars;
	
    tcaps_ether_ntoa(&ack_pack->identifier, mactmp, TRUE);
    printHexBytes(ack_pack->serialized, ack_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    
    //max string size:     13 +   3 +      13 +   17 +    11 +  3 +      13 +   5 +        14 +  5 +     14 +     3 +1  + 13    + 26 + 1 = 155 chars
	snprintf(string, stringlen, "packet_type: %d\nidentifier: %s\nack_type: %d\nerror_code: %d\nrule_number: %d\ntransfer_id: %d\nserialized_bytes: %d\nserialized: %s\n",
			ack_pack->packet_type,
			mactmp,
			ack_pack->ack_type,
			ack_pack->error_code,
			ack_pack->rule_number,
			ack_pack->transfer_id,
			ack_pack->serialized_bytes,
			bytestmp
			);
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}

struct tcaps_sp_ap * tcaps_sp_ap_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_ap_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_ap * ack_pack;
		
		ack_pack = (struct tcaps_sp_ap *) malloc(sizeof(struct tcaps_sp_ap));
		
		//printf("The packet is valid!\n");
		
		ack_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		memcpy(ack_pack->identifier.octets, (recv_packet+TCAPS_SP_AP_IDENTIFIER_OFFSET), sizeof(ack_pack->identifier));
		ack_pack->ack_type = *(recv_packet+TCAPS_SP_AP_ACK_TYPE_OFFSET);
		ack_pack->error_code = *((u16 *)(recv_packet+TCAPS_SP_AP_ERROR_CODE_OFFSET));
		ack_pack->rule_number = *((u16 *)(recv_packet+TCAPS_SP_AP_RULE_NUMBER_OFFSET));
		ack_pack->transfer_id = *(recv_packet+TCAPS_SP_AP_TRANSFER_ID_OFFSET);
		ack_pack->serialized_bytes = TCAPS_SP_AP_STATIC_SERIALIZED_BYTES;
		
		ack_pack->serialized = malloc(ack_pack->serialized_bytes);
		memcpy(ack_pack->serialized, recv_packet, ack_pack->serialized_bytes);
		
		return ack_pack;
	}
	
	//printf("The packet is INvalid!\n");
	
	return NULL;
}

boolean tcaps_sp_ap_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	if(TCAPS_SP_AP_STATIC_SERIALIZED_BYTES == packet_len
		&& TCAPS_AP_PACKET_TYPE == *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
	{

		if(TCAPS_AP_ACK_TYPE_RMP == *(packet+TCAPS_SP_AP_ACK_TYPE_OFFSET)
			|| TCAPS_AP_ACK_TYPE_TP == *(packet+TCAPS_SP_AP_ACK_TYPE_OFFSET)
				|| TCAPS_AP_ACK_TYPE_PP == *(packet+TCAPS_SP_AP_ACK_TYPE_OFFSET))
		{
			return TRUE;
		}
	}
	
	return FALSE;
}


boolean tcaps_sp_ap_compare_packets(const struct tcaps_sp_ap * ack_pack1, const struct tcaps_sp_ap * ack_pack2)
{
	boolean same = TRUE;
	
	if(ack_pack1->packet_type != ack_pack2->packet_type)
		same = FALSE;
		
	if(ack_pack1->ack_type != ack_pack2->ack_type)
		same = FALSE;
	
	if(ack_pack1->error_code != ack_pack2->error_code)
		same = FALSE;
		
	if(ack_pack1->rule_number != ack_pack2->rule_number)
		same = FALSE;
		
	if(ack_pack1->transfer_id != ack_pack2->transfer_id)
		same = FALSE;

	if(memcmp(ack_pack1->identifier.octets, ack_pack2->identifier.octets, NUM_MAC_OCTETS) != 0)
		same = FALSE;
	
	if(ack_pack1->serialized_bytes != ack_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(ack_pack1->serialized, ack_pack2->serialized, ack_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}
