#include "tcaps_sp.h"

//    u8 packet_type; //static
//    u8 op_type; //static
//	u8 fc_ip_present; //static
//	u8 tm_ip_present; //static
//    struct ip_address client_ip; //1 static + {4|16} dynamic
//    struct ip_address fc_ip; //1 static + {4|16} dynamic
//    struct ip_address tm_ip; //1 static + {4|16} dynamic
    
#define TCAPS_SP_CMP_OP_TYPE_OFFSET				TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_CMP_FC_IP_PRESENT_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+2
#define TCAPS_SP_CMP_TM_IP_PRESENT_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+3
#define TCAPS_SP_CMP_CLIENT_IP_VERSION_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+4
#define TCAPS_SP_CMP_CLIENT_IP_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+5


    
struct tcaps_sp_cmp * tcaps_sp_cmp_create(u8 optype, u8 clientipversion, char * clientip)
{
	struct tcaps_sp_cmp * cm_pack;
	
	cm_pack = (struct tcaps_sp_cmp *) malloc(sizeof(struct tcaps_sp_cmp));
	
	cm_pack->fc_ip.octets = NULL;
	cm_pack->tm_ip.octets = NULL;
	
	cm_pack->serialized_bytes = TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	cm_pack->serialized = malloc(cm_pack->serialized_bytes);
	
	
	tcaps_sp_cmp_set_packet_type(cm_pack, TCAPS_CMP_PACKET_TYPE);
	tcaps_sp_cmp_set_op_type(cm_pack, optype);
	tcaps_sp_cmp_set_fc_ip_present(cm_pack, TCAPS_SP_CMP_FC_IP_PRESENT_FALSE);
	tcaps_sp_cmp_set_tm_ip_present(cm_pack, TCAPS_SP_CMP_TM_IP_PRESENT_FALSE);
	tcaps_sp_cmp_set_client_ip_version(cm_pack, clientipversion);
	tcaps_sp_cmp_set_client_ip(cm_pack, clientip);

	return cm_pack;
}

struct tcaps_sp_cmp * tcaps_sp_cmp_create_withfc(u8 optype, u8 clientipversion, char * clientip, u8 fcipversion, char * fcip)
{
	struct tcaps_sp_cmp * cm_pack;
	
	cm_pack = (struct tcaps_sp_cmp *) malloc(sizeof(struct tcaps_sp_cmp));
	
	cm_pack->tm_ip.octets = NULL;
	
	cm_pack->serialized_bytes = TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	if(TCAPS_SP_IP_VERSION_4 == fcipversion)
	{
		cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(fcipversion);
	}
	else if (TCAPS_SP_IP_VERSION_6 == fcipversion)
	{
		cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(fcipversion);
	}
	
	cm_pack->serialized = malloc(cm_pack->serialized_bytes);
	
	
	tcaps_sp_cmp_set_packet_type(cm_pack, TCAPS_CMP_PACKET_TYPE);
	tcaps_sp_cmp_set_op_type(cm_pack, optype);
	tcaps_sp_cmp_set_fc_ip_present(cm_pack, TCAPS_SP_CMP_FC_IP_PRESENT_TRUE);
	tcaps_sp_cmp_set_tm_ip_present(cm_pack, TCAPS_SP_CMP_TM_IP_PRESENT_FALSE);
	tcaps_sp_cmp_set_client_ip_version(cm_pack, clientipversion);
	tcaps_sp_cmp_set_client_ip(cm_pack, clientip);
	tcaps_sp_cmp_set_fc_ip_version(cm_pack, fcipversion);
	tcaps_sp_cmp_set_fc_ip(cm_pack, fcip);

	return cm_pack;	
}

struct tcaps_sp_cmp * tcaps_sp_cmp_create_withtm(u8 optype, u8 clientipversion, char * clientip, u8 tmipversion, char * tmip)
{
	struct tcaps_sp_cmp * cm_pack;
	
	cm_pack = (struct tcaps_sp_cmp *) malloc(sizeof(struct tcaps_sp_cmp));
	
	cm_pack->fc_ip.octets = NULL;
	
	cm_pack->serialized_bytes = TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	if(TCAPS_SP_IP_VERSION_4 == tmipversion)
	{
		cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	else if (TCAPS_SP_IP_VERSION_6 == tmipversion)
	{
		cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	
	cm_pack->serialized = malloc(cm_pack->serialized_bytes);
	
	
	tcaps_sp_cmp_set_packet_type(cm_pack, TCAPS_CMP_PACKET_TYPE);
	tcaps_sp_cmp_set_op_type(cm_pack, optype);
	tcaps_sp_cmp_set_fc_ip_present(cm_pack, TCAPS_SP_CMP_FC_IP_PRESENT_FALSE);
	tcaps_sp_cmp_set_tm_ip_present(cm_pack, TCAPS_SP_CMP_TM_IP_PRESENT_TRUE);
	tcaps_sp_cmp_set_client_ip_version(cm_pack, clientipversion);
	tcaps_sp_cmp_set_client_ip(cm_pack, clientip);
	tcaps_sp_cmp_set_tm_ip_version(cm_pack, tmipversion);
	tcaps_sp_cmp_set_tm_ip(cm_pack, tmip);

	return cm_pack;	
}

struct tcaps_sp_cmp * tcaps_sp_cmp_create_withfctm(u8 optype, u8 clientipversion, char * clientip, u8 fcipversion, char * fcip, u8 tmipversion, char * tmip)
{
	struct tcaps_sp_cmp * cm_pack;
	
	cm_pack = (struct tcaps_sp_cmp *) malloc(sizeof(struct tcaps_sp_cmp));
	
	cm_pack->fc_ip.octets = NULL;
	
	cm_pack->serialized_bytes = TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	if(TCAPS_SP_IP_VERSION_4 == fcipversion)
	{
		cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
	}
	else if (TCAPS_SP_IP_VERSION_6 == fcipversion)
	{
		cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
	}
	
	if(TCAPS_SP_IP_VERSION_4 == tmipversion)
	{
		cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	else if (TCAPS_SP_IP_VERSION_6 == tmipversion)
	{
		cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	
	cm_pack->serialized = malloc(cm_pack->serialized_bytes);
	
	
	tcaps_sp_cmp_set_packet_type(cm_pack, TCAPS_CMP_PACKET_TYPE);
	tcaps_sp_cmp_set_op_type(cm_pack, optype);
	tcaps_sp_cmp_set_fc_ip_present(cm_pack, TCAPS_SP_CMP_FC_IP_PRESENT_TRUE);
	tcaps_sp_cmp_set_tm_ip_present(cm_pack, TCAPS_SP_CMP_TM_IP_PRESENT_TRUE);
	tcaps_sp_cmp_set_client_ip_version(cm_pack, clientipversion);
	tcaps_sp_cmp_set_client_ip(cm_pack, clientip);
	tcaps_sp_cmp_set_fc_ip_version(cm_pack, fcipversion);
	tcaps_sp_cmp_set_fc_ip(cm_pack, fcip);
	tcaps_sp_cmp_set_tm_ip_version(cm_pack, tmipversion);
	tcaps_sp_cmp_set_tm_ip(cm_pack, tmip);

	return cm_pack;	
}

void tcaps_sp_cmp_destroy(struct tcaps_sp_cmp * cm_pack)
{
	if(cm_pack->fc_ip.octets != NULL)
		free(cm_pack->fc_ip.octets);
	if(cm_pack->tm_ip.octets != NULL)
		free(cm_pack->tm_ip.octets);
		
	free(cm_pack->client_ip.octets);
	free(cm_pack->serialized);
	free(cm_pack);
}

void tcaps_sp_cmp_set_packet_type(struct tcaps_sp_cmp * cm_pack, u8 type)
{
	cm_pack->packet_type = type;
	
	//set the bytes that hold the packet_type field to the type passed in
	*((cm_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}

void tcaps_sp_cmp_set_op_type(struct tcaps_sp_cmp * cm_pack, u8 optype)
{
	cm_pack->op_type = optype;
	
	//set the bytes that hold the packet_type field to the type passed in
	*((cm_pack->serialized)+TCAPS_SP_CMP_OP_TYPE_OFFSET) = optype;
}

void tcaps_sp_cmp_set_fc_ip_present(struct tcaps_sp_cmp * cm_pack, u8 ippresent)
{
	cm_pack->fc_ip_present = ippresent;
	
	*((cm_pack->serialized)+TCAPS_SP_CMP_FC_IP_PRESENT_OFFSET) = ippresent;
}

void tcaps_sp_cmp_set_tm_ip_present(struct tcaps_sp_cmp * cm_pack, u8 ippresent)
{
	cm_pack->tm_ip_present = ippresent;
	
	*((cm_pack->serialized)+TCAPS_SP_CMP_TM_IP_PRESENT_OFFSET) = ippresent;
}

void tcaps_sp_cmp_set_client_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion)
{
	cm_pack->client_ip.ip_version = ipversion;
	
	//set the bytes that hold the ip_version field to the type passed in
	*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_VERSION_OFFSET) = ipversion;
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_cmp_set_client_ip(struct tcaps_sp_cmp * cm_pack, char * clientip)
{
	tcaps_ip_aton(clientip, &cm_pack->client_ip);
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
	{
		memcpy(((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET),
			cm_pack->client_ip.octets,
			NUM_IPV4_OCTETS);
	}
	else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
	{
		memcpy(((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET),
			cm_pack->client_ip.octets,
			NUM_IPV6_OCTETS);
	}
	
}




void tcaps_sp_cmp_set_fc_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion)
{
	cm_pack->fc_ip.ip_version = ipversion;
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
	{
		*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS) = ipversion;
	}
	else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
	{
		*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS) = ipversion;
	}
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_cmp_set_fc_ip(struct tcaps_sp_cmp * cm_pack, char * fcip)
{
	tcaps_ip_aton(fcip, &cm_pack->fc_ip);
	u8 * ptr;
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
	{
		ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
	}
	else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
	{
		ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
	}

	if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
	{
		memcpy(ptr,	cm_pack->fc_ip.octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
	{
		memcpy(ptr,	cm_pack->fc_ip.octets, NUM_IPV6_OCTETS);
	}
	
}



void tcaps_sp_cmp_set_tm_ip_version(struct tcaps_sp_cmp * cm_pack, u8 ipversion)
{
	cm_pack->tm_ip.ip_version = ipversion;
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
	{
		if(cm_pack->fc_ip.octets != NULL)
		{
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
			{
				*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS) = ipversion;
			}
			else if(TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
			{
				*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS) = ipversion;
			}	
		}
		else
		{
			*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS) = ipversion;
		}
	}
	else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
	{
		if(cm_pack->fc_ip.octets != NULL)
		{
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
			{
				*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS) = ipversion;
			}
			else if(TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
			{
				*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS) = ipversion;
			}	
		}
		else
		{
			*((cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS) = ipversion;
		}
	}
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_cmp_set_tm_ip(struct tcaps_sp_cmp * cm_pack, char * tmip)
{
	tcaps_ip_aton(tmip, &cm_pack->tm_ip);
	
	u8 * ptr;
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
	{
		if(cm_pack->fc_ip.octets != NULL)
		{
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
				ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
				
			else if(TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
				ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
		}
		else
			ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
	{
		if(cm_pack->fc_ip.octets != NULL)
		{
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
				ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
				
			else if(TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
				ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
		}
		else
			ptr = (cm_pack->serialized)+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
	}
	
	
	if(TCAPS_SP_IP_VERSION_4 == cm_pack->tm_ip.ip_version)
	{
		memcpy(ptr,	cm_pack->tm_ip.octets, NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == cm_pack->tm_ip.ip_version)
	{
		memcpy(ptr,	cm_pack->tm_ip.octets, NUM_IPV6_OCTETS);
	}
}


u8 * tcaps_sp_cmp_field_ptr(struct tcaps_sp_cmp * cm_pack, unsigned int field_to_point_to)
{
	u8 * ptr = NULL;
	
	switch(field_to_point_to)
	{
		case TCAPS_SP_CMP_POINT_TO_START_OF_CLIENT_IP_OCTETS:
			ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET;
			break;
			
		case TCAPS_SP_CMP_POINT_TO_START_OF_TM_IP_OCTETS:
			if(cm_pack->tm_ip_present == TRUE)
			{
				if(cm_pack->fc_ip_present == TRUE)
				{
					if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
					{
						if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
						{
							ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS;
						}
						else
						{
							ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS;
						}
					}
					else
					{
						if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
						{
							ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV4_OCTETS;
						}
						else
						{
							ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version)+NUM_IPV6_OCTETS;
						}
					}
				}
				else
				{
					if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
					{
						ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
					}
					else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
					{
						ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
					}
				}
			}
			break;
			
		case TCAPS_SP_CMP_POINT_TO_START_OF_FC_IP_OCTETS:
			if(cm_pack->fc_ip_present == TRUE)
			{
				if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
				{
					ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
				}
				else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
				{
					ptr = cm_pack->serialized+TCAPS_SP_CMP_CLIENT_IP_OFFSET+NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
				}
			}
			break;
			
		default:
			break;
	}
	
	return ptr;
}


/**
 * Max string size output: x chars
 */
void tcaps_sp_cmp_tostring(const struct tcaps_sp_cmp * cm_pack, char * string, unsigned int stringlen)
{
	char clientiptmp[60], fciptmp[60], tmiptmp[60], bytestmp[200];
	unsigned int remainingchars;
    
    tcaps_ip_ntoa(&cm_pack->client_ip, clientiptmp, TRUE);
    printHexBytes(cm_pack->serialized, cm_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    
    //max string size:     13 +   3 +      13 +   17 +    11 +  3 +      13 +   5 +        14 +  5 +     14 +     3 +1  = 115 chars
	snprintf(string, stringlen, "packet_type: %d\nop_type: %d\nfc_ip_present: %d\ntm_ip_present: %d\nclient_ip_version: %d\nclient_ip: %s\n",
			cm_pack->packet_type,
			cm_pack->op_type,
			cm_pack->fc_ip_present,
			cm_pack->tm_ip_present,
			cm_pack->client_ip.ip_version,
			clientiptmp);
//			,
//			cm_pack->serialized_bytes,
//			bytestmp
//			);
	
	if(cm_pack->fc_ip.octets != NULL)
	{
		tcaps_ip_ntoa(&cm_pack->fc_ip, fciptmp, TRUE);
		snprintf(string+strlen(string), stringlen-strlen(string), "fc_ip_version: %d\nfc_ip: %s\n", cm_pack->fc_ip.ip_version, fciptmp);
	}
	
	if(cm_pack->tm_ip.octets != NULL)
	{
		tcaps_ip_ntoa(&cm_pack->tm_ip, tmiptmp, TRUE);
		snprintf(string+strlen(string), stringlen-strlen(string), "tm_ip_version: %d\ntm_ip: %s\n", cm_pack->tm_ip.ip_version, tmiptmp);
	}
	
	snprintf(string+strlen(string), stringlen-strlen(string), "serialized_bytes: %d\nserialized: %s\n", cm_pack->serialized_bytes, bytestmp);
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}

struct tcaps_sp_cmp * tcaps_sp_cmp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_cmp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_cmp * cm_pack;
		unsigned int num_client_ip_bytes = 0, num_fc_ip_bytes = 0;
		
		cm_pack = (struct tcaps_sp_cmp *) malloc(sizeof(struct tcaps_sp_cmp));
		
		cm_pack->fc_ip.octets = NULL;
		cm_pack->tm_ip.octets = NULL;
		
		//printf("The packet is valid!\n");
		
		cm_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		cm_pack->op_type = *(recv_packet+TCAPS_SP_CMP_OP_TYPE_OFFSET);
		cm_pack->fc_ip_present = *(recv_packet+TCAPS_SP_CMP_FC_IP_PRESENT_OFFSET);
		cm_pack->tm_ip_present = *(recv_packet+TCAPS_SP_CMP_TM_IP_PRESENT_OFFSET);
		cm_pack->client_ip.ip_version = *(recv_packet+TCAPS_SP_CMP_CLIENT_IP_VERSION_OFFSET);
		cm_pack->serialized_bytes = TCAPS_SP_CMP_STATIC_SERIALIZED_BYTES;
		
		if(TCAPS_SP_IP_VERSION_4 == cm_pack->client_ip.ip_version)
		{
			cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
			cm_pack->serialized_bytes += NUM_IPV4_OCTETS;
			num_client_ip_bytes = NUM_IPV4_OCTETS;
			memcpy(cm_pack->client_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET), NUM_IPV4_OCTETS);
		}
		else if (TCAPS_SP_IP_VERSION_6 == cm_pack->client_ip.ip_version)
		{
			cm_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
			cm_pack->serialized_bytes += NUM_IPV6_OCTETS;
			num_client_ip_bytes = NUM_IPV6_OCTETS;
			memcpy(cm_pack->client_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET), NUM_IPV6_OCTETS);
		}
		
		
		
		if(cm_pack->fc_ip_present == TCAPS_SP_CMP_FC_IP_PRESENT_TRUE)
		{
			cm_pack->fc_ip.ip_version = *(recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes);
			
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->fc_ip.ip_version)
			{
				cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
				num_fc_ip_bytes = NUM_IPV4_OCTETS;
				memcpy(cm_pack->fc_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->fc_ip.ip_version)), NUM_IPV4_OCTETS);
			}
			else if (TCAPS_SP_IP_VERSION_6 == cm_pack->fc_ip.ip_version)
			{
				cm_pack->fc_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->fc_ip.ip_version);
				num_fc_ip_bytes = NUM_IPV6_OCTETS;
				memcpy(cm_pack->fc_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->fc_ip.ip_version)), NUM_IPV6_OCTETS);
			}
			
			if(cm_pack->tm_ip_present == TCAPS_SP_CMP_TM_IP_PRESENT_TRUE)
			{
				cm_pack->tm_ip.ip_version = *(recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->fc_ip.ip_version)+num_fc_ip_bytes);
			
				if(TCAPS_SP_IP_VERSION_4 == cm_pack->tm_ip.ip_version)
				{
					cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
					memcpy(cm_pack->tm_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->fc_ip.ip_version)+num_fc_ip_bytes+sizeof(cm_pack->tm_ip.ip_version)), NUM_IPV4_OCTETS);
				}
				else if (TCAPS_SP_IP_VERSION_6 == cm_pack->tm_ip.ip_version)
				{
					cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
					memcpy(cm_pack->tm_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->fc_ip.ip_version)+num_fc_ip_bytes+sizeof(cm_pack->tm_ip.ip_version)), NUM_IPV6_OCTETS);
				}
			}
		}
		
		else if(cm_pack->tm_ip_present == TCAPS_SP_CMP_TM_IP_PRESENT_TRUE)
		{
			cm_pack->tm_ip.ip_version = *(recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes);
			
			if(TCAPS_SP_IP_VERSION_4 == cm_pack->tm_ip.ip_version)
			{
				cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				cm_pack->serialized_bytes += NUM_IPV4_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
				memcpy(cm_pack->tm_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->tm_ip.ip_version)), NUM_IPV4_OCTETS);
			}
			else if (TCAPS_SP_IP_VERSION_6 == cm_pack->tm_ip.ip_version)
			{
				cm_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				cm_pack->serialized_bytes += NUM_IPV6_OCTETS+sizeof(cm_pack->tm_ip.ip_version);
				memcpy(cm_pack->tm_ip.octets, (recv_packet+TCAPS_SP_CMP_CLIENT_IP_OFFSET+num_client_ip_bytes+sizeof(cm_pack->tm_ip.ip_version)), NUM_IPV6_OCTETS);
			}
		}
				
		
		cm_pack->serialized = malloc(cm_pack->serialized_bytes);
		memcpy(cm_pack->serialized, recv_packet, cm_pack->serialized_bytes);
		
		return cm_pack;
	}
	
	//printf("The packet is INvalid!\n");
	
	return NULL;
}

boolean tcaps_sp_cmp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	boolean ret = TRUE;
	
	if(TCAPS_CMP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
		ret = FALSE;
		
	if(TCAPS_SP_CMP_OP_TYPE_MONITOR != *(packet+TCAPS_SP_CMP_OP_TYPE_OFFSET)
		&& TCAPS_SP_CMP_OP_TYPE_REMOVE != *(packet+TCAPS_SP_CMP_OP_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	if(TCAPS_SP_CMP_FC_IP_PRESENT_FALSE != *(packet+TCAPS_SP_CMP_FC_IP_PRESENT_OFFSET)
		&& TCAPS_SP_CMP_FC_IP_PRESENT_TRUE != *(packet+TCAPS_SP_CMP_FC_IP_PRESENT_OFFSET))
	{
		ret = FALSE;
	}
	
	if(TCAPS_SP_CMP_TM_IP_PRESENT_FALSE != *(packet+TCAPS_SP_CMP_TM_IP_PRESENT_OFFSET)
		&& TCAPS_SP_CMP_TM_IP_PRESENT_TRUE != *(packet+TCAPS_SP_CMP_TM_IP_PRESENT_OFFSET))
	{
		ret = FALSE;
	}
	
	return ret;
}

    
boolean tcaps_sp_cmp_compare_packets(const struct tcaps_sp_cmp * cm_pack1, const struct tcaps_sp_cmp * cm_pack2)
{
	boolean same = TRUE;
	
	if(cm_pack1->packet_type != cm_pack2->packet_type)
		same = FALSE;
		
	if(cm_pack1->op_type != cm_pack2->op_type)
		same = FALSE;
		
	if(cm_pack1->fc_ip_present != cm_pack2->fc_ip_present)
		same = FALSE;
		
	if(cm_pack1->tm_ip_present != cm_pack2->tm_ip_present)
		same = FALSE;
		

	if(cm_pack1->client_ip.ip_version != cm_pack2->client_ip.ip_version)
		same = FALSE;
	else
	{
		if(cm_pack1->client_ip.ip_version == TCAPS_SP_IP_VERSION_4)
		{
			if(memcmp(cm_pack1->client_ip.octets, cm_pack2->client_ip.octets, NUM_IPV4_OCTETS) != 0)
				same = FALSE;
		}
		else
		{
			if(memcmp(cm_pack1->client_ip.octets, cm_pack2->client_ip.octets, NUM_IPV6_OCTETS) != 0)
				same = FALSE;
		}
	}
	
	if(cm_pack1->tm_ip_present == TRUE && same)
	{
		if(cm_pack1->tm_ip.ip_version != cm_pack2->tm_ip.ip_version)
			same = FALSE;
		else
		{
			if(cm_pack1->tm_ip.ip_version == TCAPS_SP_IP_VERSION_4)
			{
				if(memcmp(cm_pack1->tm_ip.octets, cm_pack2->tm_ip.octets, NUM_IPV4_OCTETS) != 0)
					same = FALSE;
			}
			else
			{
				if(memcmp(cm_pack1->tm_ip.octets, cm_pack2->tm_ip.octets, NUM_IPV6_OCTETS) != 0)
					same = FALSE;
			}
		}
	}
	
	if(cm_pack1->fc_ip_present == TRUE && same)
	{
		if(cm_pack1->fc_ip.ip_version != cm_pack2->fc_ip.ip_version)
			same = FALSE;
		else
		{
			if(cm_pack1->fc_ip.ip_version == TCAPS_SP_IP_VERSION_4)
			{
				if(memcmp(cm_pack1->fc_ip.octets, cm_pack2->fc_ip.octets, NUM_IPV4_OCTETS) != 0)
					same = FALSE;
			}
			else
			{
				if(memcmp(cm_pack1->fc_ip.octets, cm_pack2->fc_ip.octets, NUM_IPV6_OCTETS) != 0)
					same = FALSE;
			}
		}
	}
	
	if(cm_pack1->serialized_bytes != cm_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(cm_pack1->serialized, cm_pack2->serialized, cm_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}
