#include "tcaps_sp.h"


//	u8 packet_type; //1 static
//	struct ip_address client_ip; //1 static + {4|16} dynamic
//	struct tcaps_sp_rmp partial_rmp;

#define TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+2



    
struct tcaps_sp_crmp * tcaps_sp_crmp_create(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout)
{
	struct tcaps_sp_crmp * fcrule_pack;
	
	fcrule_pack = (struct tcaps_sp_crmp *) malloc(sizeof(struct tcaps_sp_crmp));
	
	fcrule_pack->serialized_bytes = TCAPS_SP_CRMP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		fcrule_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		fcrule_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if(TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		fcrule_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		fcrule_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	struct tcaps_sp_rmp * tmp_rmp = tcaps_sp_rmp_create(mac, rulenum, transferid, numretries, optype, matchpri, timeout);

	fcrule_pack->serialized_bytes += tmp_rmp->serialized_bytes;
	
	fcrule_pack->serialized = malloc(fcrule_pack->serialized_bytes);
	
	tcaps_sp_crmp_set_packet_type(fcrule_pack, TCAPS_CRMP_PACKET_TYPE);
	tcaps_sp_crmp_set_clientip_version(fcrule_pack, clientipversion);
	tcaps_sp_crmp_set_client_ip(fcrule_pack, clientip);
	tcaps_sp_crmp_set_partial_rmp(fcrule_pack, tmp_rmp);
	
	return fcrule_pack;
}

struct tcaps_sp_crmp * tcaps_sp_crmp_create_ip2ip(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, char * dstip)
{
	struct ip_address ip;
	
	struct tcaps_sp_crmp * fcrule_pack = tcaps_sp_crmp_create(clientipversion, clientip, mac, rulenum, transferid, numretries, optype, matchpri, timeout);
	
	if(TCAPS_SP_IP_VERSION_4 == flowipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_4;
		ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		tcaps_ip_aton(srcip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_SRCIP_OFFSET, TCAPS_SP_IPV4HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
		tcaps_ip_aton(dstip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_DSTIP_OFFSET, TCAPS_SP_IPV4HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
	}
	else if (TCAPS_SP_IP_VERSION_6 == flowipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_6;
		ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		tcaps_ip_aton(srcip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_SRCIP_OFFSET, TCAPS_SP_IPV6HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
		tcaps_ip_aton(dstip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_DSTIP_OFFSET, TCAPS_SP_IPV6HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
	}

	free(ip.octets);
		
	return fcrule_pack;
}

struct tcaps_sp_crmp * tcaps_sp_crmp_create_ipport2ipport(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, u16 srcport, char * dstip, u16 dstport)
{
	struct ip_address ip;
	
	struct tcaps_sp_crmp * fcrule_pack = tcaps_sp_crmp_create(clientipversion, clientip, mac, rulenum, transferid, numretries, optype, matchpri, timeout);
	
	
	if(TCAPS_SP_IP_VERSION_4 == flowipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_4;
		ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		tcaps_ip_aton(srcip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_SRCIP_OFFSET, TCAPS_SP_IPV4HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_SRCPORT_OFFSET, TCAPS_SP_TCPHDR_SRCPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&srcport, sizeof(srcport)));
		tcaps_ip_aton(dstip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_DSTIP_OFFSET, TCAPS_SP_IPV4HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_DSTPORT_OFFSET, TCAPS_SP_TCPHDR_DSTPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&dstport, sizeof(dstport)));
	}
	else if (TCAPS_SP_IP_VERSION_6 == flowipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_6;
		ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		tcaps_ip_aton(srcip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_SRCIP_OFFSET, TCAPS_SP_IPV6HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_SRCPORT_OFFSET, TCAPS_SP_TCPHDR_SRCPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&srcport, sizeof(srcport)));
		tcaps_ip_aton(dstip, &ip);
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_DSTIP_OFFSET, TCAPS_SP_IPV6HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_DSTPORT_OFFSET, TCAPS_SP_TCPHDR_DSTPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&dstport, sizeof(dstport)));
	}
	

	free(ip.octets);
		
	return fcrule_pack;
}

struct tcaps_sp_crmp * tcaps_sp_crmp_create_ipspi2ipspi(u8 clientipversion, char * clientip, char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, u8 flowipversion, char * srcip, u32 srcspi, char * dstip, u32 dstspi)
{
	return tcaps_sp_crmp_create_ip2ip(clientipversion, clientip, mac, rulenum, transferid, numretries, optype, matchpri, timeout, flowipversion, srcip, dstip);
	
	//the below code needs a bit of work - #defines for SPI fields and the like
//	struct ip_address ip;
//	
//	struct tcaps_sp_crmp * fcrule_pack = tcaps_sp_crmp_create(clientipversion, clientip, mac, rulenum, transferid, numretries, optype, matchpri, timeout);
//	
//	tcaps_ip_aton(srcip, &ip);
//	
//	if(TCAPS_SP_IP_VERSION_4 == flowipversion)
//	{
//		ip.ip_version = TCAPS_SP_IP_VERSION_4;
//		ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_SRCIP_OFFSET, TCAPS_SP_IPV4HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IPSEC_HDR, TCAPS_SP_IPSECHDR_SRCSPI_OFFSET, TCAPS_SP_IPSECHDR_SRCSPI_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&srcspi, sizeof(srcspi)));
//		tcaps_ip_aton(dstip, &ip);
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_DSTIP_OFFSET, TCAPS_SP_IPV4HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IPSEC_HDR, TCAPS_SP_IPSECHDR_DSTSPI_OFFSET, TCAPS_SP_IPSECHDR_DSTSPI_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&dstspi, sizeof(dstspi)));
//	}
//	else if (TCAPS_SP_IP_VERSION_6 == flowipversion)
//	{
//		ip.ip_version = TCAPS_SP_IP_VERSION_6;
//		ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_SRCIP_OFFSET, TCAPS_SP_IPV6HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IPSEC_HDR, TCAPS_SP_IPSECHDR_SRCSPI_OFFSET, TCAPS_SP_IPSECHDR_SRCSPI_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&srcspi, sizeof(srcspi)));
//		tcaps_ip_aton(dstip, &ip);
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV6HDR_DSTIP_OFFSET, TCAPS_SP_IPV6HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV6_OCTETS));
//		tcaps_sp_crmp_add_rule_segment(fcrule_pack, tcaps_sp_crmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IPSEC_HDR, TCAPS_SP_IPSECHDR_DSTSPI_OFFSET, TCAPS_SP_IPSECHDR_DSTSPI_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&dstspi, sizeof(dstspi)));
//	}
//	
//
//	free(ip.octets);
//		
//	return fcrule_pack;
}


void tcaps_sp_crmp_destroy(struct tcaps_sp_crmp * fcrule_pack)
{
	tcaps_sp_rmp_destroy(fcrule_pack->partial_rmp);
	
	free(fcrule_pack->client_ip.octets);
	
	free(fcrule_pack->serialized);
	
	//after we have cleaned up all dynamic memory within the packet struct,
	//free the struct itself
	free(fcrule_pack);
}

	
	


struct tcaps_sp_rule_segment * tcaps_sp_crmp_create_rule_segment(u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength)
{
	return tcaps_sp_rmp_create_rule_segment(relto, offset, numbits2match, matchrule, conjunction, refbits, refbitslength);
}

void tcaps_sp_crmp_add_rule_segment(struct tcaps_sp_crmp * fcrule_pack, struct tcaps_sp_rule_segment * rule_seg)
{
	tcaps_sp_rmp_add_rule_segment(fcrule_pack->partial_rmp, rule_seg);

	u8 * tmp = (u8 *)malloc(fcrule_pack->serialized_bytes+rule_seg->total_segment_length);
	
	if(TCAPS_SP_IP_VERSION_4 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(tmp, fcrule_pack->serialized, (2*sizeof(u8))+NUM_IPV4_OCTETS);
		memcpy(tmp+(TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV4_OCTETS), fcrule_pack->partial_rmp->serialized, fcrule_pack->partial_rmp->serialized_bytes);
	}
	else if(TCAPS_SP_IP_VERSION_6 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(tmp, fcrule_pack->serialized, (2*sizeof(u8))+NUM_IPV6_OCTETS);
		memcpy(tmp+(TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV6_OCTETS), fcrule_pack->partial_rmp->serialized, fcrule_pack->partial_rmp->serialized_bytes);
	}
	
	free(fcrule_pack->serialized);
	
	fcrule_pack->serialized = tmp;
	
	fcrule_pack->serialized_bytes += rule_seg->total_segment_length;
	
}
	

void tcaps_sp_crmp_set_packet_type(struct tcaps_sp_crmp * fcrule_pack, u8 type)
{
	fcrule_pack->packet_type = type;
	
	*((fcrule_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}


void tcaps_sp_crmp_set_clientip_version(struct tcaps_sp_crmp * fcrule_pack, u8 clientipversion)
{
	fcrule_pack->client_ip.ip_version = clientipversion;
	
	//set the bytes that hold the ip_version field to the type passed in
	*((fcrule_pack->serialized)+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET) = clientipversion;
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_crmp_set_client_ip(struct tcaps_sp_crmp * fcrule_pack, char * clientip)
{
	tcaps_ip_aton(clientip, &fcrule_pack->client_ip);
	
	if(TCAPS_SP_IP_VERSION_4 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(((fcrule_pack->serialized)+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET),
			fcrule_pack->client_ip.octets,
			NUM_IPV4_OCTETS);
	}
	else if(TCAPS_SP_IP_VERSION_6 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(((fcrule_pack->serialized)+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET),
			fcrule_pack->client_ip.octets,
			NUM_IPV6_OCTETS);
	}
	
}

void tcaps_sp_crmp_set_partial_rmp(struct tcaps_sp_crmp * fcrule_pack, struct tcaps_sp_rmp * rmp)
{
	fcrule_pack->partial_rmp = rmp;
	
	if(TCAPS_SP_IP_VERSION_4 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(fcrule_pack->serialized+(TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV4_OCTETS), fcrule_pack->partial_rmp->serialized, fcrule_pack->partial_rmp->serialized_bytes);
	}
	else if(TCAPS_SP_IP_VERSION_6 == fcrule_pack->client_ip.ip_version)
	{
		memcpy(fcrule_pack->serialized+(TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV6_OCTETS), fcrule_pack->partial_rmp->serialized, fcrule_pack->partial_rmp->serialized_bytes);
	}
}




/**
 * Max string size output: 155 chars
 */
void tcaps_sp_crmp_tostring(const struct tcaps_sp_crmp * fcrule_pack, char * string, unsigned int stringlen)
{
	char iptmp[64], bytestmp[1024];
	unsigned int remainingchars;//, i = 0;
	
    tcaps_ip_ntoa(&fcrule_pack->client_ip, iptmp, TRUE);
    printHexBytes(fcrule_pack->serialized, fcrule_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    //sprintf(bytestmp, "EMPTY");
    

	snprintf(string, stringlen, "packet_type: %d\nclientip_version: %d\nclientip: %s\nserialized_bytes: %d\nserialized: %s\nENCAPSULATED_RMP\n",
			fcrule_pack->packet_type,
			fcrule_pack->client_ip.ip_version,
			iptmp,
			fcrule_pack->serialized_bytes,
			bytestmp
			);

	
	remainingchars = stringlen-strlen(string);
	
	tcaps_sp_rmp_tostring(fcrule_pack->partial_rmp, string+strlen(string), remainingchars);
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}





struct tcaps_sp_crmp * tcaps_sp_crmp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_crmp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_crmp * fcrule_pack;
		//unsigned int i = 0, curoffset = 0;
		
		fcrule_pack = (struct tcaps_sp_crmp *) malloc(sizeof(struct tcaps_sp_crmp));
		
		//printf("The rmp packet is valid!\n");
		
		fcrule_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		fcrule_pack->client_ip.ip_version = *(recv_packet+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET);
		
		if(TCAPS_SP_IP_VERSION_4 == fcrule_pack->client_ip.ip_version)
		{
			fcrule_pack->client_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
			memcpy(fcrule_pack->client_ip.octets, recv_packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET, NUM_IPV4_OCTETS);
			fcrule_pack->partial_rmp = tcaps_sp_rmp_parse_packet(recv_packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV4_OCTETS, packet_len-sizeof(u8)-sizeof(u8)-NUM_IPV4_OCTETS);
		}
		else if (TCAPS_SP_IP_VERSION_6 == fcrule_pack->client_ip.ip_version)
		{
			fcrule_pack->client_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
			memcpy(fcrule_pack->client_ip.octets, recv_packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET, NUM_IPV6_OCTETS);
			fcrule_pack->partial_rmp = tcaps_sp_rmp_parse_packet(recv_packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV6_OCTETS, packet_len-sizeof(u8)-sizeof(u8)-NUM_IPV6_OCTETS);
		}
		
		fcrule_pack->serialized_bytes = packet_len;
		

		fcrule_pack->serialized = malloc(fcrule_pack->serialized_bytes);
		memcpy(fcrule_pack->serialized, recv_packet, fcrule_pack->serialized_bytes);
		
		return fcrule_pack;
	}
	
	//printf("The rmp packet is INvalid!\n");
	
	return NULL;
}






boolean tcaps_sp_crmp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	//unsigned int numrulesegs;
	boolean ret = TRUE;//, warning = FALSE;
	
	if(TCAPS_CRMP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	if(TCAPS_SP_IP_VERSION_4 != *(packet+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET)
		&& TCAPS_SP_IP_VERSION_6 != *(packet+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET))
	{
		ret = FALSE;
	}

	if(TCAPS_SP_IP_VERSION_4 == *(packet+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET))
	{
		if(!tcaps_sp_rmp_validate_serialized_packet(packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV4_OCTETS, packet_len-sizeof(u8)-sizeof(u8)-NUM_IPV4_OCTETS))
		{
			ret = FALSE;
		}
	}
	else if(TCAPS_SP_IP_VERSION_6 == *(packet+TCAPS_SP_CRMP_CLIENTIP_VERSION_OFFSET))
	{
		if(!tcaps_sp_rmp_validate_serialized_packet(packet+TCAPS_SP_CRMP_CLIENTIP_OCTETS_OFFSET+NUM_IPV6_OCTETS, packet_len-sizeof(u8)-sizeof(u8)-NUM_IPV6_OCTETS))
		{
			ret = FALSE;
		}
	}
	
	//do something with the "warning" variable... perhaps return it via an optional parameter?
	return ret;
}

//true if packets are same
boolean tcaps_sp_crmp_compare_packets(const struct tcaps_sp_crmp * fcrule_pack1, const struct tcaps_sp_crmp * fcrule_pack2)
{
	boolean same = TRUE;
	
	if(fcrule_pack1->packet_type != fcrule_pack2->packet_type)
		same = FALSE;
		
	if(fcrule_pack1->client_ip.ip_version != fcrule_pack2->client_ip.ip_version)
		same = FALSE;
	else
	{
		if(fcrule_pack1->client_ip.ip_version == TCAPS_SP_IP_VERSION_4)
		{
			if(memcmp(fcrule_pack1->client_ip.octets, fcrule_pack2->client_ip.octets, NUM_IPV4_OCTETS) != 0)
				same = FALSE;
		}
		else
		{
			if(memcmp(fcrule_pack1->client_ip.octets, fcrule_pack2->client_ip.octets, NUM_IPV6_OCTETS) != 0)
				same = FALSE;
		}
	}
	
	if(tcaps_sp_rmp_compare_packets(fcrule_pack1->partial_rmp, fcrule_pack2->partial_rmp) == FALSE)
		same = FALSE;
	
	if(fcrule_pack1->serialized_bytes != fcrule_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(fcrule_pack1->serialized, fcrule_pack2->serialized, fcrule_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}

