#include "tcaps_sp.h"

//    u8 packet_type; //static
//    tcaps_id identifier; //static
//    u8 num_flow_feature_segments; //static
//    u8 ip_version; //static

#define TCAPS_SP_FFP_IDENTIFIER_OFFSET					(TCAPS_SP_PACKET_TYPE_OFFSET+1)
#define TCAPS_SP_FFP_NUM_FLOW_FEATURE_SEGMENTS_OFFSET	(TCAPS_SP_PACKET_TYPE_OFFSET+7)
#define TCAPS_SP_FFP_TMIP_VERSION_OFFSET				(TCAPS_SP_PACKET_TYPE_OFFSET+9)
#define TCAPS_SP_FFP_FLOW_ID_TYPE_OFFSET				(TCAPS_SP_PACKET_TYPE_OFFSET+10)
#define TCAPS_SP_FFP_CLIENTIP_VERSION_OFFSET			(TCAPS_SP_PACKET_TYPE_OFFSET+11)
#define TCAPS_SP_FFP_WHICH_IS_CLIENT_IP_OFFSET			(TCAPS_SP_PACKET_TYPE_OFFSET+12)
#define TCAPS_SP_FFP_FLOW_ID_OFFSET						(TCAPS_SP_PACKET_TYPE_OFFSET+13)



struct tcaps_sp_ffp * tcaps_sp_ffp_create(char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, const u8 * flowid)
{
	struct tcaps_sp_ffp * feature_pack;
	struct tcaps_sp_feature_segment * feature_seg;
	
	feature_pack = (struct tcaps_sp_ffp *) malloc(sizeof(struct tcaps_sp_ffp));
	
	feature_pack->serialized_bytes = TCAPS_SP_FFP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == tmipversion)
	{
		feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		feature_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == tmipversion)
	{
		feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		feature_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		switch(flowidtype)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->serialized_bytes += 2*NUM_IPV4_OCTETS;
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->serialized_bytes += (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->serialized_bytes += (2*NUM_IPV4_OCTETS)+(2*sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
		
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		switch(flowidtype)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->serialized_bytes += 2*NUM_IPV6_OCTETS;
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->serialized_bytes += (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->serialized_bytes += (2*NUM_IPV6_OCTETS)+(2*sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
	}
	
	feature_pack->serialized = malloc(feature_pack->serialized_bytes);
	
	tcaps_sp_ffp_set_packet_type(feature_pack, TCAPS_FFP_PACKET_TYPE);
	tcaps_sp_ffp_set_identifier(feature_pack, mac);
	tcaps_sp_ffp_set_flow_id_type(feature_pack, flowidtype);
	tcaps_sp_ffp_set_clientip_version(feature_pack, clientipversion);
	tcaps_sp_ffp_set_which_is_client_ip(feature_pack, whichipisclient);
	tcaps_sp_ffp_set_flow_id(feature_pack, flowid);
	tcaps_sp_ffp_set_num_feature_segments(feature_pack, 0);
	tcaps_sp_ffp_set_tmip_version(feature_pack, tmipversion);
	tcaps_sp_ffp_set_tm_ip(feature_pack, tmip);
	feature_pack->feature_segments = listlib_createLinkedList(LISTLIB_SINGLE);
	
	return feature_pack;
}



struct tcaps_sp_ffp * tcaps_sp_ffp_create_withfeatures(char * mac, u8 tmipversion, char * tmip, u8 flowidtype, u8 whichipisclient, u8 clientipversion, const u8 * flowid, u8 * serialized_features, u16 num_features)
{
	struct tcaps_sp_ffp * feature_pack;
	struct tcaps_sp_feature_segment * feature_seg;
	unsigned int curoffset = 0, i = 0;
	
	feature_pack = (struct tcaps_sp_ffp *) malloc(sizeof(struct tcaps_sp_ffp));
	
	feature_pack->serialized_bytes = TCAPS_SP_FFP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == tmipversion)
	{
		feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		feature_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == tmipversion)
	{
		feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		feature_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	
	if(TCAPS_SP_IP_VERSION_4 == clientipversion)
	{
		switch(flowidtype)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->serialized_bytes += 2*NUM_IPV4_OCTETS;
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->serialized_bytes += (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->serialized_bytes += (2*NUM_IPV4_OCTETS)+(2*sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
		
	}
	else if (TCAPS_SP_IP_VERSION_6 == clientipversion)
	{
		switch(flowidtype)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->serialized_bytes += 2*NUM_IPV6_OCTETS;
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->serialized_bytes += (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->serialized_bytes += (2*NUM_IPV6_OCTETS)+(2*sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
	}
	
	feature_pack->serialized = malloc(feature_pack->serialized_bytes);
	
	tcaps_sp_ffp_set_packet_type(feature_pack, TCAPS_FFP_PACKET_TYPE);
	tcaps_sp_ffp_set_identifier(feature_pack, mac);
	tcaps_sp_ffp_set_flow_id_type(feature_pack, flowidtype);
	tcaps_sp_ffp_set_clientip_version(feature_pack, clientipversion);
	tcaps_sp_ffp_set_which_is_client_ip(feature_pack, whichipisclient);
	tcaps_sp_ffp_set_flow_id(feature_pack, flowid);
	tcaps_sp_ffp_set_num_feature_segments(feature_pack, 0);
	tcaps_sp_ffp_set_tmip_version(feature_pack, tmipversion);
	tcaps_sp_ffp_set_tm_ip(feature_pack, tmip);
	feature_pack->feature_segments = listlib_createLinkedList(LISTLIB_SINGLE);
	
	for(; i < num_features; i++)
	{
		feature_seg = tcaps_sp_ffp_parse_feature_segment(serialized_features+curoffset);
		
		tcaps_sp_ffp_add_feature_segment(feature_pack, feature_seg);

		curoffset += sizeof(u16)+sizeof(float);
	}
	
	return feature_pack;
}

void tcaps_sp_ffp_destroy(struct tcaps_sp_ffp * feature_pack)
{
	//listlib supplies a function to clean up the linkedlist struct
	listlib_destroyLinkedList(feature_pack->feature_segments, &tcaps_sp_ffp_destroy_feature_segment);
	
	free(feature_pack->serialized);
	
	free(feature_pack->tm_ip.octets);
	
	switch(feature_pack->flow_id.flow_id_type)
	{
		case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
			free(feature_pack->flow_id.ipip.src_octets);
			free(feature_pack->flow_id.ipip.dst_octets);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
			free(feature_pack->flow_id.ipipport.src_octets);
			free(feature_pack->flow_id.ipipport.dst_octets);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
			free(feature_pack->flow_id.ipipspi.src_octets);
			free(feature_pack->flow_id.ipipspi.dst_octets);
			break;
	}
	
	//after we have cleaned up all dynamic memory within the packet struct,
	//free the struct itself
	free(feature_pack);
}

void tcaps_sp_ffp_set_packet_type(struct tcaps_sp_ffp * feature_pack, u8 type)
{
	feature_pack->packet_type = type;
	
	//set the bytes that hold the packet_type field to the tye passed in
	*((feature_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}

void tcaps_sp_ffp_set_identifier(struct tcaps_sp_ffp * feature_pack, char * mac)
{
	tcaps_ether_aton(mac, &feature_pack->identifier);
	
	memcpy(((feature_pack->serialized)+TCAPS_SP_FFP_IDENTIFIER_OFFSET), feature_pack->identifier.octets, sizeof(feature_pack->identifier));
}


void tcaps_sp_ffp_set_flow_id_type(struct tcaps_sp_ffp * feature_pack, u8 flowidtype)
{
	feature_pack->flow_id.flow_id_type = flowidtype;
	
	*((feature_pack->serialized)+TCAPS_SP_FFP_FLOW_ID_TYPE_OFFSET) = flowidtype;
}

void tcaps_sp_ffp_set_which_is_client_ip(struct tcaps_sp_ffp * feature_pack, u8 whichipisclient)
{
	feature_pack->flow_id.which_is_client_ip = whichipisclient;
	
	*((feature_pack->serialized)+TCAPS_SP_FFP_WHICH_IS_CLIENT_IP_OFFSET) = whichipisclient;
}

void tcaps_sp_ffp_set_flow_id(struct tcaps_sp_ffp * feature_pack, u8 * flowid)
{
	if(TCAPS_SP_IP_VERSION_4 == feature_pack->flow_id.clientip_version)
	{
		switch(feature_pack->flow_id.flow_id_type)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS);
				feature_pack->flow_id.ipip.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				feature_pack->flow_id.ipip.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				
				memcpy(feature_pack->flow_id.ipip.src_octets, flowid, NUM_IPV4_OCTETS);
				memcpy(feature_pack->flow_id.ipip.dst_octets, flowid+NUM_IPV4_OCTETS, NUM_IPV4_OCTETS);
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipip.src_octets, NUM_IPV4_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS, feature_pack->flow_id.ipip.dst_octets, NUM_IPV4_OCTETS);
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
				feature_pack->flow_id.ipipport.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				feature_pack->flow_id.ipipport.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				
				memcpy(feature_pack->flow_id.ipipport.src_octets, flowid, NUM_IPV4_OCTETS);
				memcpy(feature_pack->flow_id.ipipport.dst_octets, flowid+NUM_IPV4_OCTETS, NUM_IPV4_OCTETS);
				memcpy(&(feature_pack->flow_id.ipipport.src_port), flowid+(2*NUM_IPV4_OCTETS), sizeof(u16));
				memcpy(&(feature_pack->flow_id.ipipport.dst_port), flowid+(2*NUM_IPV4_OCTETS)+sizeof(u16), sizeof(u16));
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipipport.src_octets, NUM_IPV4_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS, feature_pack->flow_id.ipipport.dst_octets, NUM_IPV4_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS), &(feature_pack->flow_id.ipipport.src_port), sizeof(u16));
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)+sizeof(u16), &(feature_pack->flow_id.ipipport.dst_port), sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS)+(2*sizeof(u32));
				feature_pack->flow_id.ipipspi.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				feature_pack->flow_id.ipipspi.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
				
				memcpy(feature_pack->flow_id.ipipspi.src_octets, flowid, NUM_IPV4_OCTETS);
				memcpy(feature_pack->flow_id.ipipspi.dst_octets, flowid+NUM_IPV4_OCTETS, NUM_IPV4_OCTETS);
				memcpy(&(feature_pack->flow_id.ipipspi.src_spi), flowid+(2*NUM_IPV4_OCTETS), sizeof(u32));
				memcpy(&(feature_pack->flow_id.ipipspi.dst_spi), flowid+(2*NUM_IPV4_OCTETS)+sizeof(u32), sizeof(u32));
				
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipipspi.src_octets, NUM_IPV4_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS, feature_pack->flow_id.ipipspi.dst_octets, NUM_IPV4_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS), &(feature_pack->flow_id.ipipspi.src_spi), sizeof(u32));
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)+sizeof(u32), &(feature_pack->flow_id.ipipspi.dst_spi), sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
		
	}
	else if (TCAPS_SP_IP_VERSION_6 == feature_pack->flow_id.clientip_version)
	{
		switch(feature_pack->flow_id.flow_id_type)
		{
			case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS);
				feature_pack->flow_id.ipip.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				feature_pack->flow_id.ipip.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				
				memcpy(feature_pack->flow_id.ipip.src_octets, flowid, NUM_IPV6_OCTETS);
				memcpy(feature_pack->flow_id.ipip.dst_octets, flowid+NUM_IPV6_OCTETS, NUM_IPV6_OCTETS);
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipip.src_octets, NUM_IPV6_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS, feature_pack->flow_id.ipip.dst_octets, NUM_IPV6_OCTETS);
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
				feature_pack->flow_id.ipipport.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				feature_pack->flow_id.ipipport.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				
				memcpy(feature_pack->flow_id.ipipport.src_octets, flowid, NUM_IPV6_OCTETS);
				memcpy(feature_pack->flow_id.ipipport.dst_octets, flowid+NUM_IPV6_OCTETS, NUM_IPV6_OCTETS);
				memcpy(&(feature_pack->flow_id.ipipport.src_port), flowid+(2*NUM_IPV6_OCTETS), sizeof(u16));
				memcpy(&(feature_pack->flow_id.ipipport.dst_port), flowid+(2*NUM_IPV6_OCTETS)+sizeof(u16), sizeof(u16));
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipipport.src_octets, NUM_IPV6_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS, feature_pack->flow_id.ipipport.dst_octets, NUM_IPV6_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS), &(feature_pack->flow_id.ipipport.src_port), sizeof(u16));
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)+sizeof(u16), &(feature_pack->flow_id.ipipport.dst_port), sizeof(u16));
				break;
			case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
				feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS)+(2*sizeof(u32));
				feature_pack->flow_id.ipipspi.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				feature_pack->flow_id.ipipspi.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
				
				memcpy(feature_pack->flow_id.ipipspi.src_octets, flowid, NUM_IPV6_OCTETS);
				memcpy(feature_pack->flow_id.ipipspi.dst_octets, flowid+NUM_IPV6_OCTETS, NUM_IPV6_OCTETS);
				memcpy(&(feature_pack->flow_id.ipipspi.src_spi), flowid+(2*NUM_IPV6_OCTETS), sizeof(u32));
				memcpy(&(feature_pack->flow_id.ipipspi.dst_spi), flowid+(2*NUM_IPV6_OCTETS)+sizeof(u32), sizeof(u32));
				
				
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET, feature_pack->flow_id.ipipspi.src_octets, NUM_IPV6_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS, feature_pack->flow_id.ipipspi.dst_octets, NUM_IPV6_OCTETS);
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS), &(feature_pack->flow_id.ipipspi.src_spi), sizeof(u32));
				memcpy(feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)+sizeof(u32), &(feature_pack->flow_id.ipipspi.dst_spi), sizeof(u32));
				break;
			default:
				//We were passed a bogus flow id type
				break;
		}
	}
}

void tcaps_sp_ffp_set_clientip_version(struct tcaps_sp_ffp * feature_pack, u8 clientipversion)
{
	feature_pack->flow_id.clientip_version = clientipversion;
	
	*((feature_pack->serialized)+TCAPS_SP_FFP_CLIENTIP_VERSION_OFFSET) = clientipversion;
}


void tcaps_sp_ffp_set_num_feature_segments(struct tcaps_sp_ffp * feature_pack, u16 numfeaturesegs)
{
	feature_pack->num_flow_feature_segments = numfeaturesegs;
	
	*((u16 *)((feature_pack->serialized)+TCAPS_SP_FFP_NUM_FLOW_FEATURE_SEGMENTS_OFFSET)) = numfeaturesegs;
}

void tcaps_sp_ffp_set_tmip_version(struct tcaps_sp_ffp * feature_pack, u8 tmipversion)
{
	feature_pack->tm_ip.ip_version = tmipversion;
	
	//set the bytes that hold the ip_version field to the type passed in
	*((feature_pack->serialized)+TCAPS_SP_FFP_TMIP_VERSION_OFFSET) = tmipversion;
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_ffp_set_tm_ip(struct tcaps_sp_ffp * feature_pack, char * tmip)
{
	tcaps_ip_aton(tmip, &feature_pack->tm_ip);
	
	if(TCAPS_SP_IP_VERSION_4 == feature_pack->tm_ip.ip_version)
	{
		memcpy(((feature_pack->serialized)+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length),
			feature_pack->tm_ip.octets,
			NUM_IPV4_OCTETS);
	}
	else if (TCAPS_SP_IP_VERSION_6 == feature_pack->tm_ip.ip_version)
	{
		memcpy(((feature_pack->serialized)+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length),
			feature_pack->tm_ip.octets,
			NUM_IPV6_OCTETS);
	}
	
}

struct tcaps_sp_feature_segment * tcaps_sp_ffp_create_feature_segment(u16 key, float value)
{
	struct tcaps_sp_feature_segment * feature_seg;
	
	feature_seg = (struct tcaps_sp_feature_segment *) malloc(sizeof(struct tcaps_sp_feature_segment));
	
	feature_seg->key = key;
	feature_seg->value = value;
	
	return feature_seg;
}

void tcaps_sp_ffp_destroy_feature_segment(struct tcaps_sp_feature_segment * feature_seg)
{
	free(feature_seg);
}

void tcaps_sp_ffp_add_feature_segment(struct tcaps_sp_ffp * feature_pack, struct tcaps_sp_feature_segment * feature_seg)
{
	void * tmp, * offsetptr;
	
	listlib_appendNoCopyData(feature_pack->feature_segments, feature_seg);
	
	tmp = malloc((feature_pack->serialized_bytes)+(sizeof(u16)+sizeof(float)));
	
	memcpy(tmp, feature_pack->serialized, feature_pack->serialized_bytes);
	
	offsetptr = tmp+(feature_pack->serialized_bytes);
	
	*((u16 *)offsetptr) = feature_seg->key;
	
	offsetptr += sizeof(feature_seg->key);
	
	*((float *)offsetptr) = feature_seg->value;

	free(feature_pack->serialized);
	
	feature_pack->serialized = tmp;
	feature_pack->serialized_bytes += (sizeof(u16)+sizeof(float));
	tcaps_sp_ffp_set_num_feature_segments(feature_pack, feature_pack->num_flow_feature_segments+1);
}

u8 * tcaps_sp_ffp_field_ptr(struct tcaps_sp_ffp * feature_pack, unsigned int field_to_point_to)
{
	u8 * ptr = NULL;
	
	switch(field_to_point_to)
	{
		case TCAPS_SP_FFP_POINT_TO_FLOW_ID_SRC_OCTETS:
			ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET;
			break;
			
		case TCAPS_SP_FFP_POINT_TO_START_OF_FLOW_FEATURES:
			if(TCAPS_SP_IP_VERSION_4 == feature_pack->tm_ip.ip_version)
				ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length+NUM_IPV4_OCTETS;
			else if(TCAPS_SP_IP_VERSION_6 == feature_pack->tm_ip.ip_version)
				ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length+NUM_IPV6_OCTETS;
			break;
			
		case TCAPS_SP_FFP_POINT_TO_START_OF_CLIENT_IP_OCTETS:
			if(TCAPS_FFP_FLOW_SRC_IP_IS_CLIENT == feature_pack->flow_id.which_is_client_ip)
				ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET;
			else if(TCAPS_FFP_FLOW_DST_IP_IS_CLIENT == feature_pack->flow_id.which_is_client_ip)
			{
				if(TCAPS_SP_IP_VERSION_4 == feature_pack->flow_id.clientip_version)
					ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS;
				else if(TCAPS_SP_IP_VERSION_6 == feature_pack->flow_id.clientip_version)
					ptr = feature_pack->serialized+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS;
			}
			break;
			
		default:
			break;
	}
	
	return ptr;
}


/**
 * Max string size output: 155 chars
 */
void tcaps_sp_ffp_tostring(const struct tcaps_sp_ffp * feature_pack, char * string, unsigned int stringlen)
{
	char mactmp[20], bytestmp[512], iptmp1[60], iptmp2[60];
	unsigned int remainingchars, i = 0;
	LISTELEMENT tmp;
	struct tcaps_sp_feature_segment * feature_seg;
	
    tcaps_ether_ntoa(&feature_pack->identifier, mactmp, TRUE);
    printHexBytes(feature_pack->serialized, feature_pack->serialized_bytes, bytestmp, sizeof(bytestmp), FALSE);
    //sprintf(bytestmp, "EMPTY");
    	
    
    
	snprintf(string, stringlen, "packet_type: %d\nidentifier: %s\nnum_flow_feature_segments: %d\ntmip_version: %d\nflow_id_type: %d\nclient_ip_version: %d\nwhich_is_client_ip: %d\n",
			feature_pack->packet_type,
			mactmp,
			feature_pack->num_flow_feature_segments,
			feature_pack->tm_ip.ip_version,
			feature_pack->flow_id.flow_id_type,
			feature_pack->flow_id.clientip_version,
			feature_pack->flow_id.which_is_client_ip
			);

			
	switch(feature_pack->flow_id.flow_id_type)
	{
		struct ip_address tmpaddress;
		
		case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
			tmpaddress.ip_version = feature_pack->flow_id.clientip_version;
			tmpaddress.octets = feature_pack->flow_id.ipip.src_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp1, TRUE);
			
			tmpaddress.octets = feature_pack->flow_id.ipip.dst_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp2, TRUE);
			
			snprintf(string+strlen(string), stringlen, "flow_src_ip: %s\nflow_dst_ip: %s\n", iptmp1, iptmp2);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
			tmpaddress.ip_version = feature_pack->flow_id.clientip_version;
			tmpaddress.octets = feature_pack->flow_id.ipipport.src_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp1, TRUE);
			
			tmpaddress.octets = feature_pack->flow_id.ipipport.dst_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp2, TRUE);
			
			snprintf(string+strlen(string), stringlen, "flow_src_ip: %s\nflow_dst_ip: %s\nflow_src_port: %d\nflow_dst_port: %d\n", iptmp1, iptmp2, feature_pack->flow_id.ipipport.src_port, feature_pack->flow_id.ipipport.dst_port);
			break;
		case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
			tmpaddress.ip_version = feature_pack->flow_id.clientip_version;
			tmpaddress.octets = feature_pack->flow_id.ipipspi.src_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp1, TRUE);
			
			tmpaddress.octets = feature_pack->flow_id.ipipspi.dst_octets;
			tcaps_ip_ntoa(&tmpaddress, iptmp2, TRUE);
			
			snprintf(string+strlen(string), stringlen, "flow_src_ip: %s\nflow_dst_ip: %s\nflow_src_spi: %u\nflow_dst_spi: %u\n", iptmp1, iptmp2, feature_pack->flow_id.ipipspi.src_spi, feature_pack->flow_id.ipipspi.dst_spi);
			break;
		default:
			//We were passed a bogus flow id type
			break;
	}
	
	tcaps_ip_ntoa(&feature_pack->tm_ip, iptmp1, TRUE);
	
	snprintf(string+strlen(string), stringlen, "tm_ip: %s\nserialized_bytes: %d\nserialized: %s\n",
			iptmp1,
			feature_pack->serialized_bytes,
			bytestmp
			);

			
	remainingchars = stringlen-strlen(string);
	tmp.SINGLE = feature_pack->feature_segments->start.SINGLE;
			
	if(feature_pack->feature_segments->size > 0)
	{
		while(tmp.SINGLE != NULL)
		{
			feature_seg = (struct tcaps_sp_feature_segment *)tmp.SINGLE->data;
			snprintf(string+strlen(string),
					remainingchars,
					"FEATURE_SEGMENT%d\ntotal_segment_length: %d\nkey: %d\nvalue: %0.2f\n",
					++i,
					sizeof(u16)+sizeof(float),
					feature_seg->key,
					feature_seg->value);
			remainingchars = stringlen-strlen(string);
			tmp.SINGLE = tmp.SINGLE->nextElement;
		}
	}
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}

struct tcaps_sp_ffp * tcaps_sp_ffp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_ffp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_ffp * feature_pack;
		struct tcaps_sp_feature_segment * feature_seg;
		unsigned int i = 0, curoffset = 0;
//		char buf[1024];
		
		feature_pack = (struct tcaps_sp_ffp *) malloc(sizeof(struct tcaps_sp_ffp));
		
//		printf("###The packet is valid!###\n");
//		printHexBytes(recv_packet, packet_len, buf, sizeof(buf), FALSE);
//		printf("Contents of ffp in hex: %s\n", buf);
		
		feature_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		memcpy(feature_pack->identifier.octets, (recv_packet+TCAPS_SP_FFP_IDENTIFIER_OFFSET), sizeof(feature_pack->identifier));
		feature_pack->num_flow_feature_segments = *((u16 *)(recv_packet+TCAPS_SP_FFP_NUM_FLOW_FEATURE_SEGMENTS_OFFSET));
		feature_pack->tm_ip.ip_version = *(recv_packet+TCAPS_SP_FFP_TMIP_VERSION_OFFSET);
		feature_pack->flow_id.flow_id_type = *(recv_packet+TCAPS_SP_FFP_FLOW_ID_TYPE_OFFSET);
		feature_pack->flow_id.clientip_version = *(recv_packet+TCAPS_SP_FFP_CLIENTIP_VERSION_OFFSET);
		feature_pack->flow_id.which_is_client_ip = *(recv_packet+TCAPS_SP_FFP_WHICH_IS_CLIENT_IP_OFFSET);
		
		feature_pack->serialized_bytes = packet_len;
		
		if(TCAPS_SP_IP_VERSION_4 == feature_pack->flow_id.clientip_version)
		{
			switch(feature_pack->flow_id.flow_id_type)
			{
				case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS);
					feature_pack->flow_id.ipip.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					feature_pack->flow_id.ipip.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipip.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipip.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS), NUM_IPV4_OCTETS);
					break;
				case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS)+(2*sizeof(u16));
					feature_pack->flow_id.ipipport.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					feature_pack->flow_id.ipipport.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipipport.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipipport.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS), NUM_IPV4_OCTETS);
					memcpy(&(feature_pack->flow_id.ipipport.src_port), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)), sizeof(u16));
					memcpy(&(feature_pack->flow_id.ipipport.dst_port), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)+sizeof(u16)), sizeof(u16));
					break;
				case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV4_OCTETS)+(2*sizeof(u32));
					feature_pack->flow_id.ipipspi.src_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					feature_pack->flow_id.ipipspi.dst_octets = (u8 *)malloc(NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipipspi.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV4_OCTETS);
					memcpy(feature_pack->flow_id.ipipspi.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV4_OCTETS), NUM_IPV4_OCTETS);
					memcpy(&(feature_pack->flow_id.ipipspi.src_spi), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)), sizeof(u32));
					memcpy(&(feature_pack->flow_id.ipipspi.dst_spi), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV4_OCTETS)+sizeof(u32)), sizeof(u32));
					break;
				default:
					//We were passed a bogus flow id type
					break;
			}
			
		}
		else if (TCAPS_SP_IP_VERSION_6 == feature_pack->flow_id.clientip_version)
		{
			switch(feature_pack->flow_id.flow_id_type)
			{
				case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS);
					feature_pack->flow_id.ipip.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					feature_pack->flow_id.ipip.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipip.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipip.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS), NUM_IPV6_OCTETS);
					break;
				case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS)+(2*sizeof(u16));
					feature_pack->flow_id.ipipport.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					feature_pack->flow_id.ipipport.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipipport.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipipport.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS), NUM_IPV6_OCTETS);
					memcpy(&(feature_pack->flow_id.ipipport.src_port), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)), sizeof(u16));
					memcpy(&(feature_pack->flow_id.ipipport.dst_port), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)+sizeof(u16)), sizeof(u16));
					break;
				case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
					feature_pack->flow_id.dynamic_flow_id_length = (2*NUM_IPV6_OCTETS)+(2*sizeof(u32));
					feature_pack->flow_id.ipipspi.src_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					feature_pack->flow_id.ipipspi.dst_octets = (u8 *)malloc(NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipipspi.src_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET), NUM_IPV6_OCTETS);
					memcpy(feature_pack->flow_id.ipipspi.dst_octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+NUM_IPV6_OCTETS), NUM_IPV6_OCTETS);
					memcpy(&(feature_pack->flow_id.ipipspi.src_spi), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)), sizeof(u32));
					memcpy(&(feature_pack->flow_id.ipipspi.dst_spi), (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+(2*NUM_IPV6_OCTETS)+sizeof(u32)), sizeof(u32));
					break;
				default:
					//We were passed a bogus flow id type
					break;
			}
		}
		
//		printf("feature_pack->num_flow_feature_segments: %d\n", feature_pack->num_flow_feature_segments);
//		printf("feature_pack->tm_ip.ip_version: %d\n", feature_pack->tm_ip.ip_version);
//		printf("feature_pack->flow_id.flow_id_type: %d\n", feature_pack->flow_id.flow_id_type);
//		printf("feature_pack->flow_id.which_is_client_ip: %d\n", feature_pack->flow_id.which_is_client_ip);
		
		if(TCAPS_SP_IP_VERSION_4 == feature_pack->tm_ip.ip_version)
		{
			feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
			curoffset = TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length + NUM_IPV4_OCTETS;
			memcpy(feature_pack->tm_ip.octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length), NUM_IPV4_OCTETS);
		}
		else if (TCAPS_SP_IP_VERSION_6 == feature_pack->tm_ip.ip_version)
		{
			feature_pack->tm_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
			curoffset = TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length + NUM_IPV6_OCTETS;
			memcpy(feature_pack->tm_ip.octets, (recv_packet+TCAPS_SP_FFP_FLOW_ID_OFFSET+feature_pack->flow_id.dynamic_flow_id_length), NUM_IPV6_OCTETS);
		}
		
		//poll_pack->serialized_bytes += packet_len-TCAPS_SP_PP_STATIC_SERIALIZED_BYTES;
		
//		printf("curoffset: %d\n", curoffset);
		
		feature_pack->feature_segments = listlib_createLinkedList(LISTLIB_SINGLE);
		
		for(; i < feature_pack->num_flow_feature_segments; i++)
		{
			feature_seg = tcaps_sp_ffp_parse_feature_segment(recv_packet+curoffset);
			
			//add each feature segment to the linked list
			listlib_appendNoCopyData(feature_pack->feature_segments, feature_seg);

			curoffset += (sizeof(u16)+sizeof(float));
		}
		
		feature_pack->serialized = malloc(feature_pack->serialized_bytes);
		memcpy(feature_pack->serialized, recv_packet, feature_pack->serialized_bytes);
		
		return feature_pack;
	}
	
	//printf("###The packet is INvalid!###\n");
	
	return NULL;
}

struct tcaps_sp_feature_segment * tcaps_sp_ffp_parse_feature_segment(u8 * recv_segment)
{
	struct tcaps_sp_feature_segment * feature_seg;
	unsigned int offset = 0;
	
	feature_seg = (struct tcaps_sp_feature_segment *) malloc(sizeof(struct tcaps_sp_feature_segment));
	
	
	feature_seg->key = *((u16 *)(recv_segment+offset));
	
	offset += sizeof(u16);
	
	feature_seg->value = *((float *)(recv_segment+offset));
	
//	printf("feature_seg->key: %d\n", feature_seg->key);
//	printf("feature_seg->value: %0.2f\n", feature_seg->value);
	
	return feature_seg;
	
}

boolean tcaps_sp_ffp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	unsigned int numfeaturesegs;
	boolean ret = TRUE;
	
	if(TCAPS_FFP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
		ret = FALSE;

	//printf("There are %d feature segs!\n", *(packet+TCAPS_SP_FFP_NUM_FLOW_FEATURE_SEGMENTS_OFFSET));
	
	if(0 == *(packet+TCAPS_SP_FFP_NUM_FLOW_FEATURE_SEGMENTS_OFFSET))
	{
		ret = FALSE;
		
	}
	
	return ret;
}



    u8 packet_type; //static
    tcaps_id identifier; //static
    u16 num_flow_feature_segments; //static
    struct ip_address tm_ip; //1 static + {4|16} dynamic
    struct tcaps_sp_flow_id flow_id; //3 static + some dynamic
    LINKEDLIST * feature_segments;
    
boolean tcaps_sp_ffp_compare_packets(const struct tcaps_sp_ffp * feature_pack1, const struct tcaps_sp_ffp * feature_pack2)
{
	boolean same = TRUE;
	LISTELEMENT tmp1, tmp2;
	struct tcaps_sp_feature_segment * feature_seg1, * feature_seg2;
	
	if(feature_pack1->packet_type != feature_pack2->packet_type)
		same = FALSE;
		
	if(feature_pack1->num_flow_feature_segments != feature_pack2->num_flow_feature_segments)
		same = FALSE;

	if(memcmp(feature_pack1->identifier.octets, feature_pack2->identifier.octets, NUM_MAC_OCTETS) != 0)
		same = FALSE;
		
		
	
	if(feature_pack1->tm_ip.ip_version != feature_pack2->tm_ip.ip_version)
		same = FALSE;
	else
	{
		if(feature_pack1->tm_ip.ip_version == TCAPS_SP_IP_VERSION_4)
		{
			if(memcmp(feature_pack1->tm_ip.octets, feature_pack2->tm_ip.octets, NUM_IPV4_OCTETS) != 0)
				same = FALSE;
		}
		else
		{
			if(memcmp(feature_pack1->tm_ip.octets, feature_pack2->tm_ip.octets, NUM_IPV6_OCTETS) != 0)
				same = FALSE;
		}
	}
		
	
	
	if(feature_pack1->flow_id.flow_id_type != feature_pack2->flow_id.flow_id_type)
		same = FALSE;
	else
	{
		if(feature_pack1->flow_id.which_is_client_ip != feature_pack2->flow_id.which_is_client_ip)
			same = FALSE;
			
		if(feature_pack1->flow_id.dynamic_flow_id_length != feature_pack2->flow_id.dynamic_flow_id_length)
			same = FALSE;
			
		if(feature_pack1->flow_id.clientip_version != feature_pack2->flow_id.clientip_version)
			same = FALSE;
		else
		{
			if(feature_pack1->flow_id.clientip_version == TCAPS_SP_IP_VERSION_4)
			{
				switch(feature_pack1->flow_id.flow_id_type)
				{
					case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
						if(memcmp(feature_pack1->flow_id.ipip.src_octets, feature_pack2->flow_id.ipip.src_octets, NUM_IPV4_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipip.dst_octets, feature_pack2->flow_id.ipip.dst_octets, NUM_IPV4_OCTETS) != 0)
								same = FALSE;
						break;
					case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
						if(memcmp(feature_pack1->flow_id.ipipport.src_octets, feature_pack2->flow_id.ipipport.src_octets, NUM_IPV4_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipipport.dst_octets, feature_pack2->flow_id.ipipport.dst_octets, NUM_IPV4_OCTETS) != 0
								|| feature_pack1->flow_id.ipipport.src_port != feature_pack2->flow_id.ipipport.src_port
									|| feature_pack1->flow_id.ipipport.dst_port != feature_pack2->flow_id.ipipport.dst_port)
										same = FALSE;
						break;
					case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
						if(memcmp(feature_pack1->flow_id.ipipspi.src_octets, feature_pack2->flow_id.ipipspi.src_octets, NUM_IPV4_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipipspi.dst_octets, feature_pack2->flow_id.ipipspi.dst_octets, NUM_IPV4_OCTETS) != 0
								|| feature_pack1->flow_id.ipipspi.src_spi != feature_pack2->flow_id.ipipspi.src_spi
									|| feature_pack1->flow_id.ipipspi.dst_spi != feature_pack2->flow_id.ipipspi.dst_spi)
										same = FALSE;
						break;
				}
				
			}
			else
			{
				switch(feature_pack1->flow_id.flow_id_type)
				{
					case TCAPS_FFP_FLOW_ID_TYPE_IPIP:
						if(memcmp(feature_pack1->flow_id.ipip.src_octets, feature_pack2->flow_id.ipip.src_octets, NUM_IPV6_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipip.dst_octets, feature_pack2->flow_id.ipip.dst_octets, NUM_IPV6_OCTETS) != 0)
								same = FALSE;
						break;
					case TCAPS_FFP_FLOW_ID_TYPE_IPIPPORT:
						if(memcmp(feature_pack1->flow_id.ipipport.src_octets, feature_pack2->flow_id.ipipport.src_octets, NUM_IPV6_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipipport.dst_octets, feature_pack2->flow_id.ipipport.dst_octets, NUM_IPV6_OCTETS) != 0
								|| feature_pack1->flow_id.ipipport.src_port != feature_pack2->flow_id.ipipport.src_port
									|| feature_pack1->flow_id.ipipport.dst_port != feature_pack2->flow_id.ipipport.dst_port)
										same = FALSE;
						break;
					case TCAPS_FFP_FLOW_ID_TYPE_IPIPSPI:
						if(memcmp(feature_pack1->flow_id.ipipspi.src_octets, feature_pack2->flow_id.ipipspi.src_octets, NUM_IPV6_OCTETS) != 0
							|| memcmp(feature_pack1->flow_id.ipipspi.dst_octets, feature_pack2->flow_id.ipipspi.dst_octets, NUM_IPV6_OCTETS) != 0
								|| feature_pack1->flow_id.ipipspi.src_spi != feature_pack2->flow_id.ipipspi.src_spi
									|| feature_pack1->flow_id.ipipspi.dst_spi != feature_pack2->flow_id.ipipspi.dst_spi)
										same = FALSE;
						break;
				}
			}
		}
		
	}
	
	
	
	if(same && feature_pack1->num_flow_feature_segments > 0)
	{
		boolean feature_same = TRUE;
		
		tmp1.SINGLE = feature_pack1->feature_segments->start.SINGLE;
		tmp2.SINGLE = feature_pack2->feature_segments->start.SINGLE;
		
		while(tmp1.SINGLE != NULL && same)
		{
			feature_seg1 = (struct tcaps_sp_feature_segment *)tmp1.SINGLE->data;
			feature_seg2 = (struct tcaps_sp_feature_segment *)tmp2.SINGLE->data;
			
			feature_same = tcaps_sp_ffp_compare_featuresegs(feature_seg1, feature_seg2);
			
			if(!feature_same)
				same = FALSE;
			
			tmp1.SINGLE = tmp1.SINGLE->nextElement;
			tmp2.SINGLE = tmp2.SINGLE->nextElement;
		}
	}
	
	if(feature_pack1->serialized_bytes != feature_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(feature_pack1->serialized, feature_pack2->serialized, feature_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}



boolean tcaps_sp_ffp_compare_featuresegs(const struct tcaps_sp_feature_segment * feature_seg1, const struct tcaps_sp_feature_segment * feature_seg2)
{
	boolean same = TRUE;
	
	if(feature_seg1->key != feature_seg2->key)
		same = FALSE;
		
	if(feature_seg1->value != feature_seg2->value)
		same = FALSE;

	return same;
}

