#include "tcaps_sp.h"

//    u8 packet_type;
//    tcaps_id identifier;
//    u8 poll_type;
//    u8 num_info_segments;
//    
//    LINKEDLIST * info_segments;

#define TCAPS_SP_PP_IDENTIFIER_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_PP_POLL_TYPE_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+7
#define TCAPS_SP_PP_NUM_INFO_SEGMENTS_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+8
#define TCAPS_SP_PP_INFO_SEGMENTS_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+9

#define TCAPS_SP_INFOSEG_TOTAL_SEGMENT_LENGTH_OFFSET	0
#define TCAPS_SP_INFOSEG_KEY_LENGTH_OFFSET				4


struct tcaps_sp_pp * tcaps_sp_pp_create(char * mac, u8 polltype)
{
	struct tcaps_sp_pp * poll_pack;
	
	poll_pack = (struct tcaps_sp_pp *) malloc(sizeof(struct tcaps_sp_pp));
	
	poll_pack->serialized_bytes = TCAPS_SP_PP_STATIC_SERIALIZED_BYTES;
	
	poll_pack->serialized = malloc(poll_pack->serialized_bytes);
	
	tcaps_sp_pp_set_packet_type(poll_pack, TCAPS_PP_PACKET_TYPE);
	tcaps_sp_pp_set_identifier(poll_pack, mac);
	tcaps_sp_pp_set_poll_type(poll_pack, polltype);
	tcaps_sp_pp_set_num_info_segments(poll_pack, 0);
	poll_pack->info_segments = listlib_createLinkedList(LISTLIB_SINGLE);
	
	return poll_pack;
}

void tcaps_sp_pp_destroy(struct tcaps_sp_pp * poll_pack)
{
	//listlib supplies a function to clean up the linkedlist struct
	listlib_destroyLinkedList(poll_pack->info_segments, &tcaps_sp_pp_destroy_info_segment);
	
	free(poll_pack->serialized);
	
	//after we have cleaned up all dynamic memory within the packet struct,
	//free the struct itself
	free(poll_pack);
}

void tcaps_sp_pp_set_packet_type(struct tcaps_sp_pp * poll_pack, u8 type)
{
	poll_pack->packet_type = type;
	
	//set the bytes that hold the packet_type field to the tye passed in
	*((poll_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}

void tcaps_sp_pp_set_identifier(struct tcaps_sp_pp * poll_pack, char * mac)
{
	tcaps_ether_aton(mac, &poll_pack->identifier);
	
	memcpy(((poll_pack->serialized)+TCAPS_SP_PP_IDENTIFIER_OFFSET), poll_pack->identifier.octets, sizeof(poll_pack->identifier));
}

void tcaps_sp_pp_set_poll_type(struct tcaps_sp_pp * poll_pack, u8 polltype)
{
	poll_pack->poll_type = polltype;
	
	//set the bytes that hold the poll_type field to the type passed in
	*((poll_pack->serialized)+TCAPS_SP_PP_POLL_TYPE_OFFSET) = polltype;
}

void tcaps_sp_pp_set_num_info_segments(struct tcaps_sp_pp * poll_pack, u8 numinfosegs)
{
	poll_pack->num_info_segments = numinfosegs;
	
	*((poll_pack->serialized)+TCAPS_SP_PP_NUM_INFO_SEGMENTS_OFFSET) = numinfosegs;
}

struct tcaps_sp_info_segment * tcaps_sp_pp_create_info_segment(u8 * key, u8 keylength, u8 * value, unsigned int valuelength)
{
	struct tcaps_sp_info_segment * info_seg;
	
	info_seg = (struct tcaps_sp_info_segment *) malloc(sizeof(struct tcaps_sp_info_segment));
	
	info_seg->key = (u8 *) malloc(keylength);
	info_seg->value = (u8 *) malloc(valuelength);
	
	memcpy(info_seg->key, key, keylength);
	memcpy(info_seg->value, value, valuelength);
	
	info_seg->key_length = keylength;
	info_seg->total_segment_length = sizeof(info_seg->total_segment_length) + sizeof(info_seg->key_length) + keylength + valuelength;
	
	return info_seg;
}

//void tcaps_sp_pp_destroy_info_segment(void * test)
void tcaps_sp_pp_destroy_info_segment(struct tcaps_sp_info_segment * info_seg)
{
	//struct tcaps_sp_info_segment * info_seg = (struct tcaps_sp_info_segment *) test;
	
	free(info_seg->key);
	free(info_seg->value);
	free(info_seg);
}

void tcaps_sp_pp_add_info_segment(struct tcaps_sp_pp * poll_pack, struct tcaps_sp_info_segment * info_seg)
{
	void * tmp, * offsetptr;
	unsigned int valuelength;
	
	listlib_appendNoCopyData(poll_pack->info_segments, info_seg);
	
	tmp = malloc((poll_pack->serialized_bytes)+(info_seg->total_segment_length));
	
	memcpy(tmp, poll_pack->serialized, poll_pack->serialized_bytes);
	
	offsetptr = tmp+(poll_pack->serialized_bytes);
	
	*((int *)offsetptr) = info_seg->total_segment_length;
	
	offsetptr += sizeof(info_seg->total_segment_length);
	
	*((u8 *)offsetptr) = info_seg->key_length;
	
	offsetptr += sizeof(info_seg->key_length);
	
	memcpy(offsetptr, info_seg->key, info_seg->key_length);
	
	offsetptr += info_seg->key_length;
	
	valuelength = info_seg->total_segment_length-(info_seg->key_length)-sizeof(info_seg->total_segment_length)-sizeof(info_seg->key_length);
	
	memcpy(offsetptr, info_seg->value, valuelength);

	free(poll_pack->serialized);
	
	poll_pack->serialized = tmp;
	poll_pack->serialized_bytes += info_seg->total_segment_length;
	tcaps_sp_pp_set_num_info_segments(poll_pack, poll_pack->num_info_segments+1);
}







/**
 * Max string size output: 155 chars
 */
void tcaps_sp_pp_tostring(const struct tcaps_sp_pp * poll_pack, char * string, unsigned int stringlen)
{
	char mactmp[20], bytestmp[512];
	unsigned int remainingchars, i = 0;
	LISTELEMENT tmp;
	struct tcaps_sp_info_segment * info_seg;
	
    tcaps_ether_ntoa(&poll_pack->identifier, mactmp, TRUE);
    printHexBytes(poll_pack->serialized, poll_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    //sprintf(bytestmp, "EMPTY");
    
    //max string size:     13 +   3 +      13 +   17 +    11 +  3 +      13 +   5 +        14 +  5 +     14 +     3 +1  + 13    + 26 + 1 = 155 chars
	snprintf(string, stringlen, "packet_type: %d\nidentifier: %s\npoll_type: %d\nnum_info_segments: %d\nserialized_bytes: %d\nserialized: %s\n",
			poll_pack->packet_type,
			mactmp,
			poll_pack->poll_type,
			poll_pack->num_info_segments,
			poll_pack->serialized_bytes,
			bytestmp
			);
			
	remainingchars = stringlen-strlen(string);
	tmp.SINGLE = poll_pack->info_segments->start.SINGLE;
			
	if(poll_pack->info_segments->size > 0)
	{
		while(tmp.SINGLE != NULL)
		{
			info_seg = (struct tcaps_sp_info_segment *)tmp.SINGLE->data;
			snprintf(string+strlen(string),
					remainingchars,
					"INFO_SEGMENT%d\ntotal_segment_length: %d\nkey_length: %d\nkey: %s\nvalue: %s\n",
					++i,
					info_seg->total_segment_length,
					info_seg->key_length,
					info_seg->key,
					info_seg->value);
			remainingchars = stringlen-strlen(string);
			tmp.SINGLE = tmp.SINGLE->nextElement;
		}
	}
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}

struct tcaps_sp_pp * tcaps_sp_pp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_pp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_pp * poll_pack;
		struct tcaps_sp_info_segment * info_seg;
		unsigned int i = 0, curoffset = 0;
		
		poll_pack = (struct tcaps_sp_pp *) malloc(sizeof(struct tcaps_sp_pp));
		
		//printf("The packet is valid!\n");
		
		poll_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		memcpy(poll_pack->identifier.octets, (recv_packet+TCAPS_SP_PP_IDENTIFIER_OFFSET), sizeof(poll_pack->identifier));
		poll_pack->poll_type = *(recv_packet+TCAPS_SP_PP_POLL_TYPE_OFFSET);
		poll_pack->num_info_segments = *(recv_packet+TCAPS_SP_PP_NUM_INFO_SEGMENTS_OFFSET);
		poll_pack->serialized_bytes = packet_len;
		
		//poll_pack->serialized_bytes += packet_len-TCAPS_SP_PP_STATIC_SERIALIZED_BYTES;
		
		poll_pack->info_segments = listlib_createLinkedList(LISTLIB_SINGLE);
		
		for(; i < poll_pack->num_info_segments; i++)
		{
			info_seg = tcaps_sp_pp_parse_info_segment(recv_packet+TCAPS_SP_PP_INFO_SEGMENTS_OFFSET+curoffset, *((unsigned int *)recv_packet+TCAPS_SP_PP_INFO_SEGMENTS_OFFSET+curoffset));
			
			//add each info segment to the linked list
			listlib_appendNoCopyData(poll_pack->info_segments, info_seg);
			
			//printf("info_seg->total_segment_length: %d\n", info_seg->total_segment_length);
			//printf("info_seg->key_length: %d\n", info_seg->key_length);
			//printf("info_seg->value: %s\n", info_seg->value);
			
			//poll_pack->serialized_bytes += info_seg->total_segment_length;
			curoffset += info_seg->total_segment_length;
		}
		
		poll_pack->serialized = malloc(poll_pack->serialized_bytes);
		memcpy(poll_pack->serialized, recv_packet, poll_pack->serialized_bytes);
		
		return poll_pack;
	}
	
	//printf("The packet is INvalid!\n");
	
	return NULL;
}

struct tcaps_sp_info_segment * tcaps_sp_pp_parse_info_segment(u8 * recv_segment, unsigned int segment_len)
{
	struct tcaps_sp_info_segment * info_seg;
	unsigned int offset = 0, valuelength;
	
	info_seg = (struct tcaps_sp_info_segment *) malloc(sizeof(struct tcaps_sp_info_segment));
	
	info_seg->total_segment_length = *((unsigned int *)recv_segment);
	
	//printf("info_seg->total_segment_length: %d\n", info_seg->total_segment_length);
	
	offset += sizeof(info_seg->total_segment_length);
	
	info_seg->key_length = *(recv_segment+offset);
	
	//printf("info_seg->key_length: %d\n", info_seg->key_length);
	
	offset += sizeof(info_seg->key_length);
		
	info_seg->key = (u8 *)malloc(info_seg->key_length);
	
	memcpy(info_seg->key, (recv_segment+offset), info_seg->key_length);
	
	//printf("info_seg->key: %s\n", info_seg->key);
	
	offset += info_seg->key_length;
	
	valuelength = info_seg->total_segment_length-(info_seg->key_length)-sizeof(info_seg->total_segment_length)-sizeof(info_seg->key_length);
	
	//printf("valuelength: %d\n", valuelength);
	
	info_seg->value = (u8 *)malloc(valuelength);
	
	memcpy(info_seg->value, (recv_segment+offset), valuelength);
	
	//printf("info_seg->value: %s\n", info_seg->value);
	
	return info_seg;
	
}

boolean tcaps_sp_pp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	unsigned int numinfosegs;
	boolean ret = TRUE;
	
	if(TCAPS_PP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
		ret = FALSE;
		
	if(TCAPS_PP_POLL_TYPE_REGISTER != *(packet+TCAPS_SP_PP_POLL_TYPE_OFFSET)
		&& TCAPS_PP_POLL_TYPE_UNREGISTER != *(packet+TCAPS_SP_PP_POLL_TYPE_OFFSET)
			&& TCAPS_PP_POLL_TYPE_POLL != *(packet+TCAPS_SP_PP_POLL_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	if((numinfosegs = *(packet+TCAPS_SP_PP_NUM_INFO_SEGMENTS_OFFSET)) > 0)
	{
		//make sure there are the specified number of segments
		//and validate each one
		
		unsigned int i = 0, cursegoffset = 0, cursegsize = 0;
		
		for(; i < numinfosegs && ret; i++)
		{
			cursegsize = *(packet+TCAPS_SP_PP_INFO_SEGMENTS_OFFSET+cursegoffset);
			
			//if the calculated end of the current segment is past the actual end of the packet
			//and we aren't processing the last segment, the numbers defining the size of the
			//segment must be incorrect
			if(i < numinfosegs-1
				&& TCAPS_SP_PP_STATIC_SERIALIZED_BYTES+cursegoffset+cursegsize > packet_len)
				ret = FALSE;
			
			//if we are processing the last segment, the sum of static bytes + current offset into the segment
			//list + current segment size should add up to the end of the packet. If it doesn't, some of the numbers
			//defining the size of the segment must be incorrect
			if(i == numinfosegs-1 && TCAPS_SP_PP_STATIC_SERIALIZED_BYTES+cursegoffset+cursegsize != packet_len)
				ret = FALSE;
			
			//the beginning of the next segment is located at the end of the current segment
			cursegoffset += cursegsize;
		}
	}
	
	return ret;
}


boolean tcaps_sp_pp_compare_packets(const struct tcaps_sp_pp * poll_pack1, const struct tcaps_sp_pp * poll_pack2)
{
	boolean same = TRUE;
	unsigned int i = 0;
	LISTELEMENT tmp1, tmp2;
	struct tcaps_sp_info_segment * info_seg1, * info_seg2;
	
	if(poll_pack1->packet_type != poll_pack2->packet_type)
		same = FALSE;
		
	if(poll_pack1->poll_type != poll_pack2->poll_type)
		same = FALSE;
	
	if(poll_pack1->num_info_segments != poll_pack2->num_info_segments)
		same = FALSE;


	if(memcmp(poll_pack1->identifier.octets, poll_pack2->identifier.octets, NUM_MAC_OCTETS) != 0)
		same = FALSE;
	
	if(same && poll_pack1->num_info_segments > 0)
	{
		boolean info_same = TRUE;
		
		tmp1.SINGLE = poll_pack1->info_segments->start.SINGLE;
		tmp2.SINGLE = poll_pack2->info_segments->start.SINGLE;
		
		while(tmp1.SINGLE != NULL && same)
		{
			info_seg1 = (struct tcaps_sp_info_segment *)tmp1.SINGLE->data;
			info_seg2 = (struct tcaps_sp_info_segment *)tmp2.SINGLE->data;
			
			info_same = tcaps_sp_pp_compare_infosegs(info_seg1, info_seg2);
			
			if(!info_same)
				same = FALSE;
			
			tmp1.SINGLE = tmp1.SINGLE->nextElement;
			tmp2.SINGLE = tmp2.SINGLE->nextElement;
		}
	}
	
	if(poll_pack1->serialized_bytes != poll_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(poll_pack1->serialized, poll_pack2->serialized, poll_pack1->serialized_bytes) != 0)
		{
			//printf("Failed test: if(memcmp(poll_pack1->serialized, poll_pack2->serialized, poll_pack1->serialized_bytes) != 0)\n");
			same = FALSE;
		}
	}
	
	return same;
}


boolean tcaps_sp_pp_compare_infosegs(const struct tcaps_sp_info_segment * info_seg1, const struct tcaps_sp_info_segment * info_seg2)
{
	boolean same = TRUE;
	
	if(info_seg1->total_segment_length != info_seg2->total_segment_length)
		same = FALSE;
		
	if(info_seg1->key_length != info_seg2->key_length)
		same = FALSE;
	
	if(same)
	{
		if(memcmp(info_seg1->key, info_seg2->key, info_seg1->key_length) != 0)
			same = FALSE;
			
		if(memcmp(info_seg1->value, info_seg2->value, (info_seg1->total_segment_length)-(info_seg1->key_length)-sizeof(info_seg1->total_segment_length)) != 0)
			same = FALSE;
	}
	
	return same;
}
