#include "tcaps_sp.h"


//    u8 packet_type;
//    tcaps_id identifier;
//    u16 rule_number;
//    u8 transfer_id;
//    u8 num_retries;
//    u8 op_type;
//    u8 priority_on_match;
//    u16 timeout;
//    u8 num_rule_segments;
//    
//    LINKEDLIST * rule_segments;

#define TCAPS_SP_RMP_IDENTIFIER_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_RMP_RULE_NUMBER_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+7
#define TCAPS_SP_RMP_TRANSFER_ID_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+9
#define TCAPS_SP_RMP_NUM_RETRIES_OFFSET			TCAPS_SP_PACKET_TYPE_OFFSET+10
#define TCAPS_SP_RMP_OP_TYPE_OFFSET				TCAPS_SP_PACKET_TYPE_OFFSET+11
#define TCAPS_SP_RMP_PRIORITY_ON_MATCH_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+12
#define TCAPS_SP_RMP_TIMEOUT_OFFSET				TCAPS_SP_PACKET_TYPE_OFFSET+13
#define TCAPS_SP_RMP_NUM_RULE_SEGMENTS_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+15
#define TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+16


#define TCAPS_SP_RULESEG_TOTAL_SEGMENT_LENGTH_OFFSET	0
#define TCAPS_SP_RULESEG_RELATIVE_TO_OFFSET				4
#define TCAPS_SP_RULESEG_OFFSET_OFFSET					5
#define TCAPS_SP_RULESEG_NUM_BITS_TO_MATCH_OFFSET		7
#define TCAPS_SP_RULESEG_MATCH_RULE_OFFSET				9
#define TCAPS_SP_RULESEG_CONJUNCTION_OFFSET				10
#define TCAPS_SP_RULESEG_REF_BIT_SEQUENCE_OFFSET		11

    
struct tcaps_sp_rmp * tcaps_sp_rmp_create(char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout)
{
	struct tcaps_sp_rmp * rule_pack;
	
	rule_pack = (struct tcaps_sp_rmp *) malloc(sizeof(struct tcaps_sp_rmp));
	
	rule_pack->serialized_bytes = TCAPS_SP_RMP_STATIC_SERIALIZED_BYTES;
	
	rule_pack->serialized = malloc(rule_pack->serialized_bytes);
	
	tcaps_sp_rmp_set_packet_type(rule_pack, TCAPS_RMP_PACKET_TYPE);
	tcaps_sp_rmp_set_identifier(rule_pack, mac);
	tcaps_sp_rmp_set_rule_number(rule_pack, rulenum);
	tcaps_sp_rmp_set_transfer_id(rule_pack, transferid);
	tcaps_sp_rmp_set_num_retries(rule_pack, numretries);
	tcaps_sp_rmp_set_op_type(rule_pack, optype);
	tcaps_sp_rmp_set_priority_on_match(rule_pack, matchpri);
	tcaps_sp_rmp_set_timeout(rule_pack, timeout);
	tcaps_sp_rmp_set_num_rule_segments(rule_pack, 0);
	
	rule_pack->rule_segments = listlib_createLinkedList(LISTLIB_SINGLE);
	
	return rule_pack;
}


//u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength
struct tcaps_sp_rmp * tcaps_sp_rmp_create_ipport2ipport(char * mac, u16 rulenum, u8 transferid, u8 numretries, u8 optype, u8 matchpri, u16 timeout, char * srcip, u16 srcport, char * dstip, u16 dstport, unsigned int ipversion)
{
	struct ip_address ip;
	
	if(TCAPS_SP_IP_VERSION_4 == ipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_4;
		ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
	}
	else if (TCAPS_SP_IP_VERSION_6 == ipversion)
	{
		ip.ip_version = TCAPS_SP_IP_VERSION_6;
		ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
	}
	
	struct tcaps_sp_rmp * rule_pack = tcaps_sp_rmp_create(mac, rulenum, transferid, numretries, optype, matchpri, timeout);
	
	tcaps_ip_aton(srcip, &ip);

	tcaps_sp_rmp_add_rule_segment(rule_pack, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_SRCIP_OFFSET, TCAPS_SP_IPV4HDR_SRCIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
	tcaps_sp_rmp_add_rule_segment(rule_pack, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_SRCPORT_OFFSET, TCAPS_SP_TCPHDR_SRCPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&srcport, sizeof(srcport)));
	
	tcaps_ip_aton(dstip, &ip);
	
	tcaps_sp_rmp_add_rule_segment(rule_pack, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET, TCAPS_SP_IPV4HDR_DSTIP_OFFSET, TCAPS_SP_IPV4HDR_DSTIP_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, ip.octets, NUM_IPV4_OCTETS));
	tcaps_sp_rmp_add_rule_segment(rule_pack, tcaps_sp_rmp_create_rule_segment(TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT, TCAPS_SP_TCPHDR_DSTPORT_OFFSET, TCAPS_SP_TCPHDR_DSTPORT_BITS, TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO, TCAPS_SP_RULESEG_CONJUNCTION_AND, (u8 *)&dstport, sizeof(dstport)));
	
	free(ip.octets);
		
	return rule_pack;
}

void tcaps_sp_rmp_destroy(struct tcaps_sp_rmp * rule_pack)
{
	//listlib supplies a function to clean up the linkedlist struct
	listlib_destroyLinkedList(rule_pack->rule_segments, (void *)&tcaps_sp_rmp_destroy_rule_segment);
	
	free(rule_pack->serialized);
	
	//after we have cleaned up all dynamic memory within the packet struct,
	//free the struct itself
	free(rule_pack);
}


void tcaps_sp_rmp_destroy_rule_segment(struct tcaps_sp_rule_segment * rule_seg)
{
	//struct tcaps_sp_rule_segment * rule_seg = (struct tcaps_sp_rule_segment *) test;

	free(rule_seg->ref_bit_sequence);
	
	free(rule_seg);
}
	
	


struct tcaps_sp_rule_segment * tcaps_sp_rmp_create_rule_segment(u8 relto, u16 offset, u16 numbits2match, u8 matchrule, u8 conjunction, u8 * refbits, unsigned int refbitslength)
{
	struct tcaps_sp_rule_segment * rule_seg;
	
	rule_seg = (struct tcaps_sp_rule_segment *) malloc(sizeof(struct tcaps_sp_rule_segment));
	
	rule_seg->relative_to = relto;
	rule_seg->offset = offset;
	rule_seg->num_bits_to_match = numbits2match;
	rule_seg->match_rule = matchrule;
	rule_seg->conjunction = conjunction;
	
	//this perhaps should be based on numbits2match rather than refbits...
	//i.e. only copy numbits2match bits of refbits and discard remainder...
	rule_seg->ref_bit_sequence = (u8 *)malloc(refbitslength);
	
	memcpy(rule_seg->ref_bit_sequence, refbits, refbitslength);
	
	rule_seg->total_segment_length =  refbitslength
									+ sizeof(rule_seg->total_segment_length)
									+ sizeof(rule_seg->relative_to)
									+ sizeof(rule_seg->offset)
									+ sizeof(rule_seg->num_bits_to_match)
									+ sizeof(rule_seg->match_rule)
									+ sizeof(rule_seg->conjunction);
	
	return rule_seg;
}

void tcaps_sp_rmp_add_rule_segment(struct tcaps_sp_rmp * rule_pack, struct tcaps_sp_rule_segment * rule_seg)
{
	void * tmp, * offsetptr;
	unsigned int refbitslength;
	
	listlib_appendNoCopyData(rule_pack->rule_segments, rule_seg);
	
	
	tmp = malloc((rule_pack->serialized_bytes)+(rule_seg->total_segment_length));
	
	memcpy(tmp, rule_pack->serialized, rule_pack->serialized_bytes);
	
	//start appending the rule segment data after the previous packet contents
	offsetptr = tmp+rule_pack->serialized_bytes;
	
	*((unsigned int *)(offsetptr+TCAPS_SP_RULESEG_TOTAL_SEGMENT_LENGTH_OFFSET)) = rule_seg->total_segment_length;
	
	*((u8 *)(offsetptr+TCAPS_SP_RULESEG_RELATIVE_TO_OFFSET)) = rule_seg->relative_to;
	
	*((u16 *)(offsetptr+TCAPS_SP_RULESEG_OFFSET_OFFSET)) = rule_seg->offset;
	
	*((u16 *)(offsetptr+TCAPS_SP_RULESEG_NUM_BITS_TO_MATCH_OFFSET)) = rule_seg->num_bits_to_match;
	
	*((u8 *)(offsetptr+TCAPS_SP_RULESEG_MATCH_RULE_OFFSET)) = rule_seg->match_rule;
	
	*((u8 *)(offsetptr+TCAPS_SP_RULESEG_CONJUNCTION_OFFSET)) = rule_seg->conjunction;
	
	
	refbitslength = rule_seg->total_segment_length
					- sizeof(rule_seg->total_segment_length)
					- sizeof(rule_seg->relative_to)
					- sizeof(rule_seg->offset)
					- sizeof(rule_seg->num_bits_to_match)
					- sizeof(rule_seg->match_rule)
					- sizeof(rule_seg->conjunction);
	
	
	memcpy(offsetptr+TCAPS_SP_RULESEG_REF_BIT_SEQUENCE_OFFSET, rule_seg->ref_bit_sequence, refbitslength);

	free(rule_pack->serialized);
	
	rule_pack->serialized = tmp;
	rule_pack->serialized_bytes += rule_seg->total_segment_length;
	tcaps_sp_rmp_set_num_rule_segments(rule_pack, rule_pack->num_rule_segments+1);

	
}
	

void tcaps_sp_rmp_set_packet_type(struct tcaps_sp_rmp * rule_pack, u8 type)
{
	rule_pack->packet_type = type;
	
	*((rule_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}


void tcaps_sp_rmp_set_identifier(struct tcaps_sp_rmp * rule_pack, char * mac)
{
	tcaps_ether_aton(mac, &rule_pack->identifier);
	
	memcpy(((rule_pack->serialized)+TCAPS_SP_RMP_IDENTIFIER_OFFSET), rule_pack->identifier.octets, sizeof(rule_pack->identifier));
}


void tcaps_sp_rmp_set_rule_number(struct tcaps_sp_rmp * rule_pack, u16 rulenum)
{
	rule_pack->rule_number = rulenum;
	
	*(u16 *)((rule_pack->serialized)+TCAPS_SP_RMP_RULE_NUMBER_OFFSET) = rulenum;
}


void tcaps_sp_rmp_set_transfer_id(struct tcaps_sp_rmp * rule_pack, u8 transferid)
{
	rule_pack->transfer_id = transferid;
	
	*((rule_pack->serialized)+TCAPS_SP_RMP_TRANSFER_ID_OFFSET) = transferid;
}


void tcaps_sp_rmp_set_num_retries(struct tcaps_sp_rmp * rule_pack, u8 numretries)
{
	rule_pack->num_retries = numretries;
	
	*((rule_pack->serialized)+TCAPS_SP_RMP_NUM_RETRIES_OFFSET) = numretries;
}


void tcaps_sp_rmp_set_op_type(struct tcaps_sp_rmp * rule_pack, u8 optype)
{
	rule_pack->op_type = optype;
	
	*((rule_pack->serialized)+TCAPS_SP_RMP_OP_TYPE_OFFSET) = optype;
}


void tcaps_sp_rmp_set_priority_on_match(struct tcaps_sp_rmp * rule_pack, u8 matchpri)
{
	rule_pack->priority_on_match = matchpri;
	
	*((rule_pack->serialized)+TCAPS_SP_RMP_PRIORITY_ON_MATCH_OFFSET) = matchpri;
}


void tcaps_sp_rmp_set_timeout(struct tcaps_sp_rmp * rule_pack, u16 timeout)
{
	rule_pack->timeout = timeout;
	
	*(u16 *)((rule_pack->serialized)+TCAPS_SP_RMP_TIMEOUT_OFFSET) = timeout;
}


void tcaps_sp_rmp_set_num_rule_segments(struct tcaps_sp_rmp * rule_pack, u8 numrulesegs)
{
	rule_pack->num_rule_segments = numrulesegs;
	
	*((rule_pack->serialized)+TCAPS_SP_RMP_NUM_RULE_SEGMENTS_OFFSET) = numrulesegs;
}




/**
 * Max string size output: 155 chars
 */
void tcaps_sp_rmp_tostring(const struct tcaps_sp_rmp * rule_pack, char * string, unsigned int stringlen)
{
	char mactmp[20], bytestmp[1024];
	unsigned int remainingchars, i = 0;
	LISTELEMENT tmp;
	struct tcaps_sp_rule_segment * rule_seg;
	
    tcaps_ether_ntoa(&rule_pack->identifier, mactmp, TRUE);
    printHexBytes(rule_pack->serialized, rule_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    //sprintf(bytestmp, "EMPTY");
    
    //max string size:     13 +   3 +      13 +   17 +    11 +  3 +      13 +   5 +        14 +  5 +     14 +     3 +1  + 13    + 26 + 1 = 155 chars
	snprintf(string, stringlen, "packet_type: %d\nidentifier: %s\nrule_number: %d\ntransfer_id: %d\nnum_retries: %d\nop_type: %d\npriority_on_match: %d\ntimeout: %d\nnum_rule_segments: %d\nserialized_bytes: %d\nserialized: %s\n",
			rule_pack->packet_type,
			mactmp,
			rule_pack->rule_number,
			rule_pack->transfer_id,
			rule_pack->num_retries,
			rule_pack->op_type,
			rule_pack->priority_on_match,
			rule_pack->timeout,
			rule_pack->num_rule_segments,
			rule_pack->serialized_bytes,
			bytestmp
			);
			
	remainingchars = stringlen-strlen(string);
	tmp.SINGLE = rule_pack->rule_segments->start.SINGLE;
			
	if(rule_pack->rule_segments->size > 0)
	{
		while(tmp.SINGLE != NULL)
		{
			rule_seg = (struct tcaps_sp_rule_segment *)tmp.SINGLE->data;
			
			printHexBytes(rule_seg->ref_bit_sequence, ceil(rule_seg->num_bits_to_match/8.0), bytestmp, sizeof(bytestmp), TRUE);
			
			snprintf(string+strlen(string),
					remainingchars,
					"RULE_SEGMENT%d\ntotal_segment_length: %d\nrelative_to: %d\noffset: %d\nnum_bits_to_match: %d\nmatch_rule: %d\nconjunction: %d\nref_bit_sequence: %s\n",
					++i,
					rule_seg->total_segment_length,
					rule_seg->relative_to,
					rule_seg->offset,
					rule_seg->num_bits_to_match,
					rule_seg->match_rule,
					rule_seg->conjunction,
					bytestmp);
					
			remainingchars = stringlen-strlen(string);
			tmp.SINGLE = tmp.SINGLE->nextElement;
		}
	}
	
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}




struct tcaps_sp_rmp * tcaps_sp_rmp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_rmp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_rmp * rule_pack;
		struct tcaps_sp_rule_segment * rule_seg;
		unsigned int i = 0, curoffset = 0;
		
		rule_pack = (struct tcaps_sp_rmp *) malloc(sizeof(struct tcaps_sp_rmp));
		
		//printf("The rmp packet is valid!\n");
		
		rule_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		memcpy(rule_pack->identifier.octets, (recv_packet+TCAPS_SP_RMP_IDENTIFIER_OFFSET), sizeof(rule_pack->identifier));
		rule_pack->rule_number = *((u16 *)(recv_packet+TCAPS_SP_RMP_RULE_NUMBER_OFFSET));
		rule_pack->transfer_id = *(recv_packet+TCAPS_SP_RMP_TRANSFER_ID_OFFSET);
		rule_pack->num_retries = *(recv_packet+TCAPS_SP_RMP_NUM_RETRIES_OFFSET);
		rule_pack->op_type = *(recv_packet+TCAPS_SP_RMP_OP_TYPE_OFFSET);
		rule_pack->priority_on_match = *(recv_packet+TCAPS_SP_RMP_PRIORITY_ON_MATCH_OFFSET);
		rule_pack->timeout = *((u16 *)(recv_packet+TCAPS_SP_RMP_TIMEOUT_OFFSET));
		rule_pack->num_rule_segments = *(recv_packet+TCAPS_SP_RMP_NUM_RULE_SEGMENTS_OFFSET);
		rule_pack->serialized_bytes = packet_len;
		
		rule_pack->rule_segments = listlib_createLinkedList(LISTLIB_SINGLE);

		for(; i < rule_pack->num_rule_segments; i++)
		{
			rule_seg = tcaps_sp_rmp_parse_rule_segment(recv_packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+curoffset, *((unsigned int *)recv_packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+curoffset));
			
			//add each info segment to the linked list
			listlib_appendNoCopyData(rule_pack->rule_segments, rule_seg);
			
			//printf("rule_seg->total_segment_length: %d\n", rule_seg->total_segment_length);

			curoffset += rule_seg->total_segment_length;
		}
		
		rule_pack->serialized = malloc(rule_pack->serialized_bytes);
		memcpy(rule_pack->serialized, recv_packet, rule_pack->serialized_bytes);
		
		return rule_pack;
	}
	
	//printf("The rmp packet is INvalid!\n");
	
	return NULL;
}



struct tcaps_sp_rule_segment * tcaps_sp_rmp_parse_rule_segment(u8 * recv_segment, unsigned int segment_len)
{
	struct tcaps_sp_rule_segment * rule_seg;
	unsigned int refbitslength;
	
	rule_seg = (struct tcaps_sp_rule_segment *) malloc(sizeof(struct tcaps_sp_rule_segment));
	
	rule_seg->total_segment_length = *((unsigned int *)(recv_segment+TCAPS_SP_RULESEG_TOTAL_SEGMENT_LENGTH_OFFSET));
	
	//printf("rule_seg->total_segment_length: %d\n", rule_seg->total_segment_length);
	
	
	rule_seg->relative_to = *(recv_segment+TCAPS_SP_RULESEG_RELATIVE_TO_OFFSET);
	
	rule_seg->offset = *((u16 *)(recv_segment+TCAPS_SP_RULESEG_OFFSET_OFFSET));
	
	rule_seg->num_bits_to_match = *((u16 *)(recv_segment+TCAPS_SP_RULESEG_NUM_BITS_TO_MATCH_OFFSET));
	
	rule_seg->match_rule = *(recv_segment+TCAPS_SP_RULESEG_MATCH_RULE_OFFSET);
	
	rule_seg->conjunction = *(recv_segment+TCAPS_SP_RULESEG_CONJUNCTION_OFFSET);

	
	refbitslength = rule_seg->total_segment_length
					- sizeof(rule_seg->total_segment_length)
					- sizeof(rule_seg->relative_to)
					- sizeof(rule_seg->offset)
					- sizeof(rule_seg->num_bits_to_match)
					- sizeof(rule_seg->match_rule)
					- sizeof(rule_seg->conjunction);
	
	//printf("refbitslength: %d\n", refbitslength);
	rule_seg->ref_bit_sequence = (u8 *)malloc(refbitslength);
	
	memcpy(rule_seg->ref_bit_sequence, (recv_segment+TCAPS_SP_RULESEG_REF_BIT_SEQUENCE_OFFSET), refbitslength);
	

	
	return rule_seg;
	
}





boolean tcaps_sp_rmp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	unsigned int numrulesegs;
	boolean ret = TRUE, warning = FALSE;
	
	if(TCAPS_RMP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	if(TCAPS_RMP_OP_TYPE_CREATE != *(packet+TCAPS_SP_RMP_OP_TYPE_OFFSET)
		&& TCAPS_RMP_OP_TYPE_DELETE != *(packet+TCAPS_SP_RMP_OP_TYPE_OFFSET)
			&& TCAPS_RMP_OP_TYPE_UPDATE != *(packet+TCAPS_SP_RMP_OP_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	if((numrulesegs = *(packet+TCAPS_SP_RMP_NUM_RULE_SEGMENTS_OFFSET)) > 0)
	{
		//make sure there are the specified number of segments
		//and validate each one
		
		
		unsigned int i = 0, cursegoffset = 0, cursegsize = 0, refbitsbytelength = 0, num_bits_to_match_in_bytes = 0;
		struct tcaps_sp_rule_segment rule_seg;
		u8 * tmpoffset;
		
		for(; i < numrulesegs && ret; i++)
		{
			//there should be an if() check to make sure TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset
			//isn't past the end of the packet... e.g. if numrulesegs > 0 but packet_len == TCAPS_SP_RMP_STATIC_SERIALIZED_BYTES
			//then we know that numrulesegs is foobed
			cursegsize = *(packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset);
			
			//if the calculated end of the current segment is past the actual end of the packet
			//and we aren't processing the last segment, the numbers defining the size of the
			//segment must be incorrect
			if(i < numrulesegs-1
				&& TCAPS_SP_RMP_STATIC_SERIALIZED_BYTES+cursegoffset+cursegsize > packet_len)
				ret = FALSE;
			
			//if we are processing the last segment, the sum of static bytes + current offset into the segment
			//list + current segment size should add up to the end of the packet. If it doesn't, some of the numbers
			//defining the size of the segment must be incorrect
			if(i == numrulesegs-1 && TCAPS_SP_RMP_STATIC_SERIALIZED_BYTES+cursegoffset+cursegsize != packet_len)
				ret = FALSE;

				
			if(ret)
			{
				//let's do some more detailed checking of this segment
				
				tmpoffset = (packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset+TCAPS_SP_RULESEG_RELATIVE_TO_OFFSET);
				
				if(TCAPS_SP_RULESEG_RELTO_START_OF_FRAME != *tmpoffset
					&& TCAPS_SP_RULESEG_RELTO_START_OF_IP_PACKET != *tmpoffset
						&& TCAPS_SP_RULESEG_RELTO_START_OF_TCP_SEGMENT != *tmpoffset
							&& TCAPS_SP_RULESEG_RELTO_START_OF_DATA_SEGMENT != *tmpoffset)
				{
					//printf("RELATIVE TO error\n");
					ret = FALSE;
				}
				
				tmpoffset = (packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset+TCAPS_SP_RULESEG_MATCH_RULE_OFFSET);
				
				if(TCAPS_SP_RULESEG_MATCH_RULE_EQUAL_TO != *tmpoffset
					&& TCAPS_SP_RULESEG_MATCH_RULE_NOT_EQUAL_TO != *tmpoffset
						&& TCAPS_SP_RULESEG_MATCH_RULE_LESS_THAN != *tmpoffset
							&& TCAPS_SP_RULESEG_MATCH_RULE_LESS_THAN_OR_EQUAL_TO != *tmpoffset
								&& TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN != *tmpoffset
									&& TCAPS_SP_RULESEG_MATCH_RULE_GREATER_THAN_OR_EQUAL_TO != *tmpoffset
										&& TCAPS_SP_RULESEG_MATCH_RULE_LIKE != *tmpoffset)
				{
					//printf("MATCH RULE error\n");
					ret = FALSE;
				}
				
				
				tmpoffset = (packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset+TCAPS_SP_RULESEG_CONJUNCTION_OFFSET);
				
				//we don't need a conjunction for the last segment, so let's not require the caller to set a valid one!
				if((i < numrulesegs-1)
					&& (TCAPS_SP_RULESEG_CONJUNCTION_AND != *tmpoffset && TCAPS_SP_RULESEG_CONJUNCTION_OR != *tmpoffset))
				{
					//printf("CONJUNCTION error\n");
					ret = FALSE;
				}
				
				refbitsbytelength = cursegsize
									- sizeof(rule_seg.total_segment_length)
									- sizeof(rule_seg.relative_to)
									- sizeof(rule_seg.offset)
									- sizeof(rule_seg.num_bits_to_match)
									- sizeof(rule_seg.match_rule)
									- sizeof(rule_seg.conjunction);
								
				tmpoffset = (packet+TCAPS_SP_RMP_RULE_SEGMENTS_OFFSET+cursegoffset+TCAPS_SP_RULESEG_NUM_BITS_TO_MATCH_OFFSET);
				
				//calculate the minimum number of bytes required to match the requested number of bits
				num_bits_to_match_in_bytes = ceil((*tmpoffset)/8.0);
				
				//printf("num_bits_to_match_in_bytes: %d\n", num_bits_to_match_in_bytes);
				//printf("refbitsbytelength: %d\n", refbitsbytelength);
				
				//if we have a packet that has less reference bytes than the minimum required to match
				//the requested number of bits, something has gone astray!
				//If we have more reference bytes, we are being inefficient, but I've decided not to flag
				//that as an outright error, but rather as a warning
				if(refbitsbytelength < num_bits_to_match_in_bytes)
				{
					//printf("REF BITS error... refbitsbytelength: %d, num_bits_to_match_in_bytes: %d\n", refbitsbytelength, num_bits_to_match_in_bytes);
					ret = FALSE;
				}
				else if(refbitsbytelength > num_bits_to_match_in_bytes)
				{
					warning = TRUE;
				}
				
			}
			
			//the beginning of the next segment is located at the end of the current segment
			cursegoffset += cursegsize;
		}
	}
	
	//do something with the "warning" variable... perhaps return it via an optional parameter?
	return ret;
}

    
    
boolean tcaps_sp_rmp_compare_packets(const struct tcaps_sp_rmp * rule_pack1, const struct tcaps_sp_rmp * rule_pack2)
{
	boolean same = TRUE;
	unsigned int i = 0;
	LISTELEMENT tmp1, tmp2;
	struct tcaps_sp_rule_segment * rule_seg1, * rule_seg2;
	
	if(rule_pack1->packet_type != rule_pack2->packet_type)
		same = FALSE;
		
	if(rule_pack1->rule_number != rule_pack2->rule_number)
		same = FALSE;
	
	if(rule_pack1->transfer_id != rule_pack2->transfer_id)
		same = FALSE;
		
	if(rule_pack1->num_retries != rule_pack2->num_retries)
		same = FALSE;
		
	if(rule_pack1->op_type != rule_pack2->op_type)
		same = FALSE;
		
	if(rule_pack1->priority_on_match != rule_pack2->priority_on_match)
		same = FALSE;
		
	if(rule_pack1->timeout != rule_pack2->timeout)
		same = FALSE;
		
	if(rule_pack1->num_rule_segments != rule_pack2->num_rule_segments)
		same = FALSE;


	if(memcmp(rule_pack1->identifier.octets, rule_pack2->identifier.octets, NUM_MAC_OCTETS) != 0)
		same = FALSE;
	
	if(same && rule_pack1->num_rule_segments > 0)
	{
		boolean rule_same = TRUE;
		
		tmp1.SINGLE = rule_pack1->rule_segments->start.SINGLE;
		tmp2.SINGLE = rule_pack2->rule_segments->start.SINGLE;
		
		while(tmp1.SINGLE != NULL && same)
		{
			rule_seg1 = (struct tcaps_sp_rule_segment *)tmp1.SINGLE->data;
			rule_seg2 = (struct tcaps_sp_rule_segment *)tmp2.SINGLE->data;
			
			rule_same = tcaps_sp_rmp_compare_rulesegs(rule_seg1, rule_seg2);
			
			if(!rule_same)
				same = FALSE;
			
			tmp1.SINGLE = tmp1.SINGLE->nextElement;
			tmp2.SINGLE = tmp2.SINGLE->nextElement;
		}
	}
	
	if(rule_pack1->serialized_bytes != rule_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(rule_pack1->serialized, rule_pack2->serialized, rule_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}

boolean tcaps_sp_rmp_compare_rules(const struct tcaps_sp_rmp * rule_pack1, const struct tcaps_sp_rmp * rule_pack2)
{
	boolean same = TRUE, rule_same = TRUE;
	
	if(rule_pack1->num_rule_segments > 0 && (rule_pack1->num_rule_segments == rule_pack2->num_rule_segments))
	{
		LISTELEMENT tmp1, tmp2;
		struct tcaps_sp_rule_segment * rule_seg1, * rule_seg2;
		
		tmp1.SINGLE = rule_pack1->rule_segments->start.SINGLE;
		tmp2.SINGLE = rule_pack2->rule_segments->start.SINGLE;
		
		while(tmp1.SINGLE != NULL && same)
		{
			rule_seg1 = (struct tcaps_sp_rule_segment *)tmp1.SINGLE->data;
			rule_seg2 = (struct tcaps_sp_rule_segment *)tmp2.SINGLE->data;
			
			rule_same = tcaps_sp_rmp_compare_rulesegs(rule_seg1, rule_seg2);
			
			if(!rule_same)
				same = FALSE;
			
			tmp1.SINGLE = tmp1.SINGLE->nextElement;
			tmp2.SINGLE = tmp2.SINGLE->nextElement;
		}
	}
	else
		same = FALSE;
	
	return same;
}



boolean tcaps_sp_rmp_compare_rulesegs(const struct tcaps_sp_rule_segment * rule_seg1, const struct tcaps_sp_rule_segment * rule_seg2)
{
	boolean same = TRUE;
	
	//printf("tcaps_sp_rmp_compare_rulesegs()\n");
	
	if(rule_seg1->total_segment_length != rule_seg2->total_segment_length)
		same = FALSE;
		
	if(rule_seg1->relative_to != rule_seg2->relative_to)
		same = FALSE;
		
	if(rule_seg1->offset != rule_seg2->offset)
		same = FALSE;
		
	if(rule_seg1->num_bits_to_match != rule_seg2->num_bits_to_match)
		same = FALSE;
	
	if(rule_seg1->match_rule != rule_seg2->match_rule)
		same = FALSE;
		
	if(rule_seg1->conjunction != rule_seg2->conjunction)
		same = FALSE;

	
	if(same)
	{
		if(memcmp(rule_seg1->ref_bit_sequence, rule_seg2->ref_bit_sequence, (int)ceilf((rule_seg1->num_bits_to_match)/8.0)) != 0)
			same = FALSE;

	}
	
	return same;
}
