#include "tcaps_sp.h"

//    u8 packet_type;
//    //tcaps_id identifier;
//    u8 transfer_type;
//    u8 transfer_id;
//    u16 number_of_rules;
//    u8 ip_version;
//    struct ip_address server_ip;
//    u16 server_port;
    
#define TCAPS_SP_TP_TRANSFER_TYPE_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+1
#define TCAPS_SP_TP_TRANSFER_ID_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+2
#define TCAPS_SP_TP_NUMBER_OF_RULES_OFFSET	TCAPS_SP_PACKET_TYPE_OFFSET+3
#define TCAPS_SP_TP_SERVER_PORT_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+5
#define TCAPS_SP_TP_IP_VERSION_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+7
#define TCAPS_SP_TP_SERVER_IP_OFFSET		TCAPS_SP_PACKET_TYPE_OFFSET+8


    
struct tcaps_sp_tp * tcaps_sp_tp_create(u8 transtype, u8 transid, u16 numrules, u8 ipversion, char * serverip, u16 serverport)
{
	struct tcaps_sp_tp * trans_pack;
	
	trans_pack = (struct tcaps_sp_tp *) malloc(sizeof(struct tcaps_sp_tp));
	
	trans_pack->serialized_bytes = TCAPS_SP_TP_STATIC_SERIALIZED_BYTES;
	
	if(TCAPS_SP_IP_VERSION_4 == ipversion)
	{
		trans_pack->server_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
		trans_pack->serialized_bytes += NUM_IPV4_OCTETS;
	}
	else if (TCAPS_SP_IP_VERSION_6 == ipversion)
	{
		trans_pack->server_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
		trans_pack->serialized_bytes += NUM_IPV6_OCTETS;
	}
	
	trans_pack->serialized = malloc(trans_pack->serialized_bytes);
	
	
	tcaps_sp_tp_set_packet_type(trans_pack, TCAPS_TP_PACKET_TYPE);
	tcaps_sp_tp_set_transfer_type(trans_pack, transtype);
	tcaps_sp_tp_set_transfer_id(trans_pack, transid);
	tcaps_sp_tp_set_number_of_rules(trans_pack, numrules);
	tcaps_sp_tp_set_server_port(trans_pack, serverport);
	tcaps_sp_tp_set_ip_version(trans_pack, ipversion);
	tcaps_sp_tp_set_server_ip(trans_pack, serverip);
	
	
	//tcaps_ether_aton(mac, &trans_pack->identifier);

	return trans_pack;
}

void tcaps_sp_tp_destroy(struct tcaps_sp_tp * trans_pack)
{
	free(trans_pack->server_ip.octets);
	free(trans_pack->serialized);
	free(trans_pack);
}

void tcaps_sp_tp_set_packet_type(struct tcaps_sp_tp * trans_pack, u8 type)
{
	trans_pack->packet_type = type;
	
	//set the bytes that hold the packet_type field to the type passed in
	*((trans_pack->serialized)+TCAPS_SP_PACKET_TYPE_OFFSET) = type;
}

void tcaps_sp_tp_set_transfer_type(struct tcaps_sp_tp * trans_pack, u8 transtype)
{
	trans_pack->transfer_type = transtype;
	
	//set the bytes that hold the transfer_type field to the type passed in
	*((trans_pack->serialized)+TCAPS_SP_TP_TRANSFER_TYPE_OFFSET) = transtype;
}



void tcaps_sp_tp_set_transfer_id(struct tcaps_sp_tp * trans_pack, u8 transid)
{
	trans_pack->transfer_id = transid;
	
	//set the bytes that hold the transfer_id field to the type passed in
	*((trans_pack->serialized)+TCAPS_SP_TP_TRANSFER_ID_OFFSET) = transid;
}

void tcaps_sp_tp_set_number_of_rules(struct tcaps_sp_tp * trans_pack, u16 numrules)
{
	trans_pack->number_of_rules = numrules;
	
	//set the bytes that hold the number_of_rules field to the type passed in
	*(u16 *)((trans_pack->serialized)+TCAPS_SP_TP_NUMBER_OF_RULES_OFFSET) = numrules;
}

void tcaps_sp_tp_set_ip_version(struct tcaps_sp_tp * trans_pack, u8 ipversion)
{
	trans_pack->server_ip.ip_version = ipversion;
	
	//set the bytes that hold the ip_version field to the type passed in
	*((trans_pack->serialized)+TCAPS_SP_TP_IP_VERSION_OFFSET) = ipversion;
}

//assumes that the "ip_version" field has been set prior to calling
void tcaps_sp_tp_set_server_ip(struct tcaps_sp_tp * trans_pack, char * serverip)
{
	tcaps_ip_aton(serverip, &trans_pack->server_ip);
	
	if(TCAPS_SP_IP_VERSION_4 == trans_pack->server_ip.ip_version)
	{
		memcpy(((trans_pack->serialized)+TCAPS_SP_TP_SERVER_IP_OFFSET),
			trans_pack->server_ip.octets,
			NUM_IPV4_OCTETS);
	}
	else if (TCAPS_SP_IP_VERSION_6 == trans_pack->server_ip.ip_version)
	{
		memcpy(((trans_pack->serialized)+TCAPS_SP_TP_SERVER_IP_OFFSET),
			trans_pack->server_ip.octets,
			NUM_IPV6_OCTETS);
	}
	
}

void tcaps_sp_tp_set_server_port(struct tcaps_sp_tp * trans_pack, u16 serverport)
{
	trans_pack->server_port = serverport;
	
	//set the bytes that hold the transfer_type field to the type passed in
	*((u16 *)((trans_pack->serialized)+TCAPS_SP_TP_SERVER_PORT_OFFSET)) = serverport;
}

/**
 * Max string size output: x chars
 */
void tcaps_sp_tp_tostring(const struct tcaps_sp_tp * trans_pack, char * string, unsigned int stringlen)
{
	char iptmp[60], bytestmp[100];
	unsigned int remainingchars;
    
    tcaps_ip_ntoa(&trans_pack->server_ip, iptmp, TRUE);
    printHexBytes(trans_pack->serialized, trans_pack->serialized_bytes, bytestmp, sizeof(bytestmp), TRUE);
    
    //max string size:     13 +   3 +      13 +   17 +    11 +  3 +      13 +   5 +        14 +  5 +     14 +     3 +1  = 115 chars
	snprintf(string, stringlen, "packet_type: %d\ntransfer_type: %d\ntransfer_id: %d\nnumber_of_rules: %d\nserver_port: %d\nip_version: %d\nserver_ip: %s\nserialized_bytes: %d\nserialized: %s\n",
			trans_pack->packet_type,
			trans_pack->transfer_type,
			trans_pack->transfer_id,
			trans_pack->number_of_rules,
			trans_pack->server_port,
			trans_pack->server_ip.ip_version,
			iptmp,
			trans_pack->serialized_bytes,
			bytestmp
			);
		
	remainingchars = stringlen-strlen(string);
	
	//printf("BEFORE\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
	
	if(remainingchars == 1)
	{
		sprintf(string+stringlen-4, "...");
	}
	
	//printf("AFTER\nstringlen: %d\nstrlen(string): %d\n\n", stringlen, strlen(string));
}

struct tcaps_sp_tp * tcaps_sp_tp_parse_packet(u8 * recv_packet, unsigned int packet_len)
{
	if(tcaps_sp_tp_validate_serialized_packet(recv_packet, packet_len))
	{
		struct tcaps_sp_tp * trans_pack;
		
		trans_pack = (struct tcaps_sp_tp *) malloc(sizeof(struct tcaps_sp_tp));
		
		//printf("The packet is valid!\n");
		
		trans_pack->packet_type = *(recv_packet+TCAPS_SP_PACKET_TYPE_OFFSET);
		trans_pack->transfer_type = *(recv_packet+TCAPS_SP_TP_TRANSFER_TYPE_OFFSET);
		trans_pack->transfer_id = *(recv_packet+TCAPS_SP_TP_TRANSFER_ID_OFFSET);
		trans_pack->number_of_rules = *((u16 *)(recv_packet+TCAPS_SP_TP_NUMBER_OF_RULES_OFFSET));
		trans_pack->server_port = *((u16 *)(recv_packet+TCAPS_SP_TP_SERVER_PORT_OFFSET));
		trans_pack->server_ip.ip_version = *(recv_packet+TCAPS_SP_TP_IP_VERSION_OFFSET);
		trans_pack->serialized_bytes = TCAPS_SP_TP_STATIC_SERIALIZED_BYTES;
		
		if(TCAPS_SP_IP_VERSION_4 == trans_pack->server_ip.ip_version)
		{
			trans_pack->server_ip.octets = (u8 *)malloc(NUM_IPV4_OCTETS);
			trans_pack->serialized_bytes += NUM_IPV4_OCTETS;
			memcpy(trans_pack->server_ip.octets, (recv_packet+TCAPS_SP_TP_SERVER_IP_OFFSET), NUM_IPV4_OCTETS);
		}
		else if (TCAPS_SP_IP_VERSION_6 == trans_pack->server_ip.ip_version)
		{
			trans_pack->server_ip.octets = (u8 *)malloc(NUM_IPV6_OCTETS);
			trans_pack->serialized_bytes += NUM_IPV6_OCTETS;
			memcpy(trans_pack->server_ip.octets, (recv_packet+TCAPS_SP_TP_SERVER_IP_OFFSET), NUM_IPV6_OCTETS);
		}
				
		
		trans_pack->serialized = malloc(trans_pack->serialized_bytes);
		memcpy(trans_pack->serialized, recv_packet, trans_pack->serialized_bytes);
		
		return trans_pack;
	}
	
	//printf("The packet is INvalid!\n");
	
	return NULL;
}

boolean tcaps_sp_tp_validate_serialized_packet(u8 * packet, unsigned int packet_len)
{
	boolean ret = TRUE;
	
	if(TCAPS_TP_PACKET_TYPE != *(packet+TCAPS_SP_PACKET_TYPE_OFFSET))
		ret = FALSE;
		
	if(TCAPS_SP_IP_VERSION_4 == *(packet+TCAPS_SP_TP_IP_VERSION_OFFSET))
	{
		if((TCAPS_SP_TP_STATIC_SERIALIZED_BYTES+NUM_IPV4_OCTETS) != packet_len)
			ret = FALSE;
	}
	else if(TCAPS_SP_IP_VERSION_6 == *(packet+TCAPS_SP_TP_IP_VERSION_OFFSET))
	{
		if((TCAPS_SP_TP_STATIC_SERIALIZED_BYTES+NUM_IPV6_OCTETS) != packet_len)
			ret = FALSE;
	}
	else
		ret = FALSE;
	

	if(TCAPS_TP_TRANSFER_TYPE_RULE != *(packet+TCAPS_SP_TP_TRANSFER_TYPE_OFFSET)
		&& TCAPS_TP_TRANSFER_TYPE_SERVER != *(packet+TCAPS_SP_TP_TRANSFER_TYPE_OFFSET)
			&& TCAPS_TP_TRANSFER_TYPE_CLIENT != *(packet+TCAPS_SP_TP_TRANSFER_TYPE_OFFSET))
	{
		ret = FALSE;
	}
	
	return ret;
}


    u8 packet_type; //static
    //tcaps_id identifier;
    u8 transfer_type; //static
    u8 transfer_id; //static
    u16 number_of_rules; //static
    //u8 ip_version; //static
    u16 server_port; //static
    struct ip_address server_ip; //1 static + {4|16} dynamic
    
    
boolean tcaps_sp_tp_compare_packets(const struct tcaps_sp_tp * trans_pack1, const struct tcaps_sp_tp * trans_pack2)
{
	boolean same = TRUE;
	
	if(trans_pack1->packet_type != trans_pack2->packet_type)
		same = FALSE;
		
	if(trans_pack1->transfer_type != trans_pack2->transfer_type)
		same = FALSE;
		
	if(trans_pack1->transfer_id != trans_pack2->transfer_id)
		same = FALSE;
		
	if(trans_pack1->number_of_rules != trans_pack2->number_of_rules)
		same = FALSE;
		
	if(trans_pack1->server_port != trans_pack2->server_port)
		same = FALSE;
	

	if(trans_pack1->server_ip.ip_version != trans_pack2->server_ip.ip_version)
		same = FALSE;
	else
	{
		if(trans_pack1->server_ip.ip_version == TCAPS_SP_IP_VERSION_4)
		{
			if(memcmp(trans_pack1->server_ip.octets, trans_pack2->server_ip.octets, NUM_IPV4_OCTETS) != 0)
				same = FALSE;
		}
		else
		{
			if(memcmp(trans_pack1->server_ip.octets, trans_pack2->server_ip.octets, NUM_IPV6_OCTETS) != 0)
				same = FALSE;
		}
	}
	
	if(trans_pack1->serialized_bytes != trans_pack2->serialized_bytes)
		same = FALSE;
	else
	{	
		if(memcmp(trans_pack1->serialized, trans_pack2->serialized, trans_pack1->serialized_bytes) != 0)
			same = FALSE;
	}
	
	return same;
}
