#include "tcaps_sp.h"

#include <sys/param.h>
#include <sys/linker.h>
#include <errno.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>




#define __FREEBSD__

//xdtoi(c)
//        register int c;
//{
//        if (isdigit(c))
//                return c - '0';
//        else if (islower(c))
//                return c - 'a' + 10;
//        else
//                return c - 'A' + 10;
//}

void tcaps_sp_get_interface_mac(tcaps_id * mac, const char * interface)
{
	//tcaps_id * mac;
	char macbuf[64];
	
	//mac = (tcaps_id *)malloc(sizeof(tcaps_id));
	
	#ifdef __FREEBSD__
	
		#include <sys/types.h>
		#include <sys/socket.h>
		#include <sys/ioctl.h>
		#include <sys/sysctl.h>
		#include <net/if.h>
		#include <net/if_dl.h>
		#include <netinet/in.h>
		#include <arpa/inet.h>
		#include <errno.h>
		#include <stdio.h>
		#include <stdlib.h>
		
		int mib[6], len;
		char *buf;
		struct if_msghdr *ifm;
		struct sockaddr_dl *sdl;
	
	
		mib[0] = CTL_NET;
		mib[1] = AF_ROUTE;
		mib[2] = 0;
		mib[3] = AF_LINK;
		mib[4] = NET_RT_IFLIST;
		
		if ((mib[5] = if_nametoindex(interface)) == 0) {
			perror("if_nametoindex error");
			exit(2);
		}
	
		if (sysctl(mib, 6, NULL, &len, NULL, 0) < 0) {
			perror("sysctl 1 error");
			exit(3);
		}
	
		if ((buf = malloc(len)) == NULL) {
			perror("malloc error");
			exit(4);
		}
	
		if (sysctl(mib, 6, buf, &len, NULL, 0) < 0) {
			perror("sysctl 2 error");
			exit(5);
		}
	
		ifm = (struct if_msghdr *)buf;
		sdl = (struct sockaddr_dl *)(ifm + 1);
		
		//tcaps_ether_ntoa(const tcaps_id * mac, char * string, boolean separate);
		memcpy(mac->octets, LLADDR(sdl), sizeof(mac->octets));
		//tcaps_ether_ntoa(mac, macbuf, TRUE);
		//printHexBytes(mac->octets, 6, macbuf, sizeof(macbuf), TRUE);
		//printf("mac address: %s\n", macbuf);
		
		free(buf);
		
	#endif
}

void tcaps_ip_aton(const char * string, struct ip_address * ip)
{
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		char temp[4];
		unsigned int i = 0, cur_octet = 0;
	
		memset(temp, 0, sizeof(temp));
		//zero the byte array
		memset((ip->octets), 0, sizeof(u8)*NUM_IPV4_OCTETS);
		
		//should be this while(*string && pos_counter < NUM_MAC_OCTETS)
		while(*string)
		{
			if(IPV4_SEPARATOR_CHAR != *string && isdigit(*string) && i < 3)
			{
				temp[i++] = *string;
			}
			
			else if(i > 0)
			{
				ip->octets[cur_octet++] = atoi(temp);
				i = 0;
				memset(temp, 0, sizeof(temp));
			}
			
			string++;
		}
		
		//catch the final octet, as the loop exits when *string is null,
		//which occurs after processing the final character of the final octet
		ip->octets[cur_octet++] = atoi(temp);
		
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		boolean high = FALSE;
		u8 temp = 0;
		unsigned int pos_counter = 0;
		
		//zero the byte array
		memset((ip->octets), 0, sizeof(u8)*NUM_IPV6_OCTETS);
		
		while(*string && pos_counter < NUM_IPV6_OCTETS)
		{
			if(*string == IPV6_SEPARATOR_CHAR)
			{
				string += 1;
			}
			
			if(isHexDigit(*string))
			{
				temp = hexCharToInt(*string);
				
				if(high)
				{
					//if high = true, we are storing the second
					//hex digit of the current set, so stuff it into
					//the top 4 bits of our byte
					*((ip->octets)+pos_counter) |= temp<<4;
					high = FALSE;
				}
				else
				{
					//if high = false, we are storing the first
					//hex digit of the current set, so stuff it into
					//the bottom 4 bits of our byte
					*((ip->octets)+pos_counter) |= temp;
					high = TRUE;
				}
				
				//*((mac->octets)+pos_counter) |= hexCharToInt(*string);
			}
			
			//if we are here and high is FALSE, we have processed 2 hex chars
			//and stuffed them into 1 byte chunk. Time to move onto the next array index
			if(!high)
				pos_counter += 1;
			
			//increment out string pointer index
			string++;
		}
	}
	
}

void tcaps_ip_ntoa(const struct ip_address * ip, char * string, boolean separate)
{
	if(TCAPS_SP_IP_VERSION_4 == ip->ip_version)
	{
		if(separate)
		{
			sprintf(string, "%d%c%d%c%d%c%d",
					ip->octets[0], IPV4_SEPARATOR_CHAR,
					ip->octets[1], IPV4_SEPARATOR_CHAR,
					ip->octets[2], IPV4_SEPARATOR_CHAR,
					ip->octets[3]);
		}
		else
		{
			sprintf(string, "%d%d%d%d",
					ip->octets[0],
					ip->octets[1],
					ip->octets[2],
					ip->octets[3]);
		}
			
	}
	else if(TCAPS_SP_IP_VERSION_6 == ip->ip_version)
	{
		int pos_counter = 0, tmp;
		boolean high = FALSE, seen4chars = FALSE;
		
		//loop as many times as there are bytes in the MAC array
		for(; pos_counter < NUM_IPV6_OCTETS; string++)
		{
			if(high)
			{
				//if high = true, we are converting the second
				//hex digit of the current set, so read the top
				//4 bits of the byte and then shift them down 4 positions
				//to bring them back into range 0-15 and convert
				tmp = ((u8)*((ip->octets)+pos_counter)) & 0xF0;
				tmp >>= 4;
				*string = intToHexChar(tmp);
				
				high = FALSE;
			}
			else
			{
				//if high = false, we are converting the first
				//hex digit of the current set, so read the bottom
				//4 bits of the byte and convert
				tmp = ((u8)*((ip->octets)+pos_counter)) & 0x0F;
				*string = intToHexChar(tmp);
				high = TRUE;
			}
			
			//if we are here and high is FALSE, we have processed 2 hex chars
			//Time to move onto the next array index
			if(!high)
			{
				//if the caller wants each set of 4 hex digits separated, and we have processed 4 hex chars,
				//move 1 char further along and write the separator char
				//(unless we are processing the final set of 4 chars i.e. don't put a trailing separator)
				if(separate && pos_counter < (NUM_IPV6_OCTETS-1) && seen4chars)
				{
					*++string = IPV6_SEPARATOR_CHAR;
					seen4chars = FALSE;
				}
				else
					seen4chars = TRUE;
				
				pos_counter++;			
			}
		}
		
		//append a trailing NULL terminator to the string
		*string = '\0';
	}
	
}


/**
 * Function to convert an ASCII string representation of a MAC
 * address to a binary "tcaps_id" struct form. The ASCII string
 * must be of the format "xx:xx:xx:xx:xx:xx" or "xxxxxxxxxxxx",
 * where the x's represent hexadecimal digits 0-9, a-f, A-F.
 * The function zeros the tcaps_id MAC struct before beginning to
 * ensure the ORing process used in the function will work.
 * The digits are stored in an array of 1 byte memory blocks.
 * Each 1 byte block stores 2 hex digits. Digits are read from beginning
 * of the input string to the end, and the first digit of each 2 digit block is
 * stored in the lower 4 bits of the 1 byte memory block. The second digit of
 * each 2 digit block is stored in the upper 4 bits of the 1 byte memory block.
 * 
 * @param string - contains the formatted ASCII MAC address
 * @param mac - the previously allocated struct to store the MAC in
 * 
 */
void tcaps_ether_aton(const char * string, tcaps_id * mac)
{
	unsigned int i = 0, pos_counter = 0;
	u8 temp = 0;
	boolean high = FALSE;
	
	//memset(temp, 0, sizeof(temp));
	
	//zero the byte array
	memset((mac->octets), 0, sizeof(u8)*NUM_MAC_OCTETS);
	
	//should be this while(*string && pos_counter < NUM_MAC_OCTETS)
	while(*string && i < MAX_MAC_CHARS)
	{
		if(*string == MAC_SEPARATOR_CHAR)
		{
			string += 1;
		}
		
		if(isHexDigit(*string))
		{
			temp = hexCharToInt(*string);
			
			if(high)
			{
				//if high = true, we are storing the second
				//hex digit of the current set, so stuff it into
				//the top 4 bits of our byte
				*((mac->octets)+pos_counter) |= temp<<4;
				high = FALSE;
			}
			else
			{
				//if high = false, we are storing the first
				//hex digit of the current set, so stuff it into
				//the bottom 4 bits of our byte
				*((mac->octets)+pos_counter) |= temp;
				high = TRUE;
			}
			
			//*((mac->octets)+pos_counter) |= hexCharToInt(*string);
		}
		
		//if we are here and high is FALSE, we have processed 2 hex chars
		//and stuffed them into 1 byte chunk. Time to move onto the next array index
		if(!high)
			pos_counter += 1;
		
		//increment our safety counter
		i++;
		
		//increment out string pointer index
		string++;
	}
}

/**
 * Function to convert a tcaps_id struct into an ASCII string representation.
 * If the "separate" paramter is TRUE, each pair of hex digits will be separated
 * by the separator char defined by MAC_SEPARATOR_CHAR in tcaps_sp.h
 * The maximum number of chars this function will require is MAX_MAC_CHARS+1 (for '\0').
 * The caller is required to be passing in a string that has the required space
 * already allocated to it.
 * 
 * @param mac - tcaps_id struct containing the MAC address to convert
 * @param string - the string to write the ASCII MAC address to
 * @param separate - is true, separate pairs of hex digits by MAC_SEPARATOR_CHAR 
 */
void tcaps_ether_ntoa(const tcaps_id * mac, char * string, boolean separate)
{
	int pos_counter = 0, tmp;
	boolean high = FALSE;
	
	//loop as many times as there are bytes in the MAC array
	for(; pos_counter < NUM_MAC_OCTETS; string++)
	{
		if(high)
		{
			//if high = true, we are converting the second
			//hex digit of the current set, so read the top
			//4 bits of the byte and then shift them down 4 positions
			//to bring them back into range 0-15 and convert
			tmp = ((u8)*((mac->octets)+pos_counter)) & 0xF0;
			tmp >>= 4;
			*string = intToHexChar(tmp);
			
			//if the caller wants pairs separated, move 1 char further
			//along and write the separator char
			//(unless we are processing the final pair of chars i.e. don't put a trailing separator)
			if(separate && pos_counter < (NUM_MAC_OCTETS-1))
				*++string = MAC_SEPARATOR_CHAR;
			
			high = FALSE;
		}
		else
		{
			//if high = false, we are converting the first
			//hex digit of the current set, so read the bottom
			//4 bits of the byte and convert
			tmp = ((u8)*((mac->octets)+pos_counter)) & 0x0F;
			*string = intToHexChar(tmp);
			high = TRUE;
		}
		
		//if we are here and high is FALSE, we have processed 2 hex chars
		//Time to move onto the next array index
		if(!high)
			pos_counter++;
	}
	
	//append a trailing NULL terminator to the string
	*string = '\0';
}

/**
 * Simple function to convert a character that represents a hex number
 * into a number. The characters 0-9, a-f and A-F will be converted to
 * their respective numeric equivalents. Any other character passed as an input
 * to this function will be returned as the number 16 (an illegal hex number).
 * i.e. you can test if this function successfully converted a hext digit by
 * testing to see if the return value is less than 16.
 * 
 * @param c - char to convert to a hex number
 * @return numeric equivalent of the hex input character
 */
u8 hexCharToInt(const char c)
{
	switch(c)
	{
		case '0':
			return 0;
			break;
		case '1':
			return 1;
			break;
		case '2':
			return 2;
			break;
		case '3':
			return 3;
			break;
		case '4':
			return 4;
			break;
		case '5':
			return 5;
			break;
		case '6':
			return 6;
			break;
		case '7':
			return 7;
			break;
		case '8':
			return 8;
			break;
		case '9':
			return 9;
			break;
		case 'A':
		case 'a':
			return 10;
			break;
		case 'B':
		case 'b':
			return 11;
			break;
		case 'C':
		case 'c':
			return 12;
			break;
		case 'D':
		case 'd':
			return 13;
			break;
		case 'E':
		case 'e':
			return 14;
			break;
		case 'F':
		case 'f':
			return 15;
			break;
		default:
			return 16;
			break;
	}
}

char intToHexChar(const int c)
{
	switch(c)
	{
		case 0:
			return '0';
			break;
		case 1:
			return '1';
			break;
		case 2:
			return '2';
			break;
		case 3:
			return '3';
			break;
		case 4:
			return '4';
			break;
		case 5:
			return '5';
			break;
		case 6:
			return '6';
			break;
		case 7:
			return '7';
			break;
		case 8:
			return '8';
			break;
		case 9:
			return '9';
			break;
		case 10:
			return 'A';
			break;
		case 11:
			return 'B';
			break;
		case 12:
			return 'C';
			break;
		case 13:
			return 'D';
			break;
		case 14:
			return 'E';
			break;
		case 15:
			return 'F';
			break;
		default:
			return '\0';
			break;
	}
}
	
/**
 * Simple function to determine if a character represents a valid hex digit.
 * The characters 0-9, a-f and A-F are considered valid, and will cause the function
 * to return true. All other characters will cause the function to return false.
 * 
 * @param c - character to check
 * @return true if character repesents a valid hex digit, false otherwise
 */
boolean isHexDigit(const char c)
{
	switch(c)
	{
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		case 'A':
		case 'a':
		case 'B':
		case 'b':
		case 'C':
		case 'c':
		case 'D':
		case 'd':
		case 'E':
		case 'e':
		case 'F':
		case 'f':
			return TRUE;
			break;
		default:
			return FALSE;
			break;
	}
}

void printHexBytes(const u8 * bytes, unsigned int num_bytes, char * string, unsigned int stringlen, boolean hex_order_high_first)
{
	int pos_counter = 0, tmp;
	boolean high = hex_order_high_first;
	
	//loop as many times as there are bytes in the MAC array or until we've filled
	//up the string we were given, leaving room to append these 4 chars "...\0"
	//don't forget that each increment of pos_counter represents having written 2 hex
	//digits to the string i.e. why we multiply pos_counter by 2 before comparing to string length
	for(; pos_counter < num_bytes && pos_counter*2 < stringlen-4; string++)
	{
		if(high)
		{
			//if high = true, we are converting the second
			//hex digit of the current set, so read the top
			//4 bits of the byte and then shift them down 4 positions
			//to bring them back into range 0-15 and convert
			tmp = ((u8)*(bytes+pos_counter)) & 0xF0;
			tmp >>= 4;
			*string = intToHexChar(tmp);
			
			high = FALSE;
		}
		else
		{
			//if high = false, we are converting the first
			//hex digit of the current set, so read the bottom
			//4 bits of the byte and convert
			tmp = ((u8)*(bytes+pos_counter)) & 0x0F;
			*string = intToHexChar(tmp);
			high = TRUE;
		}
		
		//if we are here and high is FALSE, we have processed 2 hex chars
		//Time to move onto the next array index
		if((hex_order_high_first && high)
			|| (!hex_order_high_first && !high))
			pos_counter++;
			
	}
	
	if(pos_counter*2 == stringlen-4)
	{
		*string = '.';
		*++string = '.';
		*++string = '.';
		string++;
	}
	
	//append a trailing NULL terminator to the string
	*string = '\0';
}




u8 * tcaps_atohex(const char * hexstring)
{
	unsigned int i = 0, pos_counter = 0, num_hex_bytes;
	u8 temp = 0;
	boolean high = FALSE;
	
	num_hex_bytes = (int)(ceilf(strlen(hexstring)/2.0));
	
	//printf("num_hex_bytes: %d\n", num_hex_bytes);
	
	u8 * hexbytes = (u8 *)malloc(num_hex_bytes);
	
	//memset(temp, 0, sizeof(temp));
	
	//zero the byte array
	memset(hexbytes, 0, num_hex_bytes);
	
	//should be this while(*string && pos_counter < NUM_MAC_OCTETS)
	while(*hexstring)
	{
		if(*hexstring == BYTE_SEPARATOR_CHAR)
		{
			hexstring += 1;
		}
		
		if(isHexDigit(*hexstring))
		{
			temp = hexCharToInt(*hexstring);
			//printf("tcaps_atohex()\t temp: %d\n", temp);
			
			if(high)
			{
				//if high = true, we are storing the second
				//hex digit of the current set, so stuff it into
				//the top 4 bits of our byte
				*(hexbytes+pos_counter) |= temp<<4;
				high = FALSE;
			}
			else
			{
				//if high = false, we are storing the first
				//hex digit of the current set, so stuff it into
				//the bottom 4 bits of our byte
				*(hexbytes+pos_counter) |= temp;
				high = TRUE;
			}
			
			//*((mac->octets)+pos_counter) |= hexCharToInt(*string);
		}
		
		//if we are here and high is FALSE, we have processed 2 hex chars
		//and stuffed them into 1 byte chunk. Time to move onto the next array index
		if(!high)
			pos_counter += 1;
		
		//increment our safety counter
		i++;
		
		//increment out string pointer index
		hexstring++;
	}
	
	return hexbytes;
}
		
		
		
		
